/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.allseen;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;

import org.opendof.tools.repository.interfaces.core.AuthenticatedUser;
import org.opendof.tools.repository.interfaces.core.CoreController;
import org.opendof.tools.repository.interfaces.core.PersistenceController;
import org.opendof.tools.repository.interfaces.core.RawXMLTranslator;
import org.opendof.tools.repository.interfaces.core.XMLTranslator;
import org.opendof.tools.repository.interfaces.core.XSLTTranslator;
import org.opendof.tools.repository.interfaces.da.AlreadyExistsException;
import org.opendof.tools.repository.interfaces.da.DataAccessor;
import org.opendof.tools.repository.interfaces.da.DataAccessor.InterfaceIterable;
import org.opendof.tools.repository.interfaces.da.ImmutableException;
import org.opendof.tools.repository.interfaces.da.InsufficentRightsException;
import org.opendof.tools.repository.interfaces.da.InterfaceData;
import org.opendof.tools.repository.interfaces.da.InterfaceIdMatchException;
import org.opendof.tools.repository.interfaces.da.NotFoundException;
import org.opendof.tools.repository.interfaces.da.SubRepositoryNode;
import org.opendof.tools.repository.interfaces.da.SubmitterData;
import org.opendof.tools.repository.interfaces.da.SubmitterException;
import org.opendof.tools.repository.interfaces.da.SubmitterPermissions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/* 
 * a*([ybnqiuxtdsogv]|\[\w+\])"
 * 
 * Basic dbus data types:
 * uint8        y
 * boolean      b
 * int16        n
 * uint16       q
 * int32        i
 * uint32       u
 * int64        x
 * uint64       t
 * float64      d
 * float32      h - not in xsd 
 * 
 * null terminated string like types
 * string       s
 * objpath      o
 * signature    g
 * 
 * Four container types
 * structure    r - but () used in signatures [] in xml
 * array        a   a followed by any single complete type       
 * variant      v   v followed by single complete type "i", "ai", "(ii)" ok, "ii" not. total msg depth <= 64
 * dict         {}  funky struct enforced as a name/value pair
 * 
 * schema questions
 * 1. description element missing xml:lang attribute
 * 2. what is data type "w"?
 * 3. what about enumeration? not in sample but saw this somewhere.
 */

/**
 * The Allseen Controller enforces the http:allseenalliance.org/schemas/extended-introspect.xsd
 * XML namespace rules for adding, reading, modifying and deleting interfaces from the persistent store.
 * <p> 
 * This class implements the <code>PersistenceController</code> interface which provides a relay for
 * requests from the <code>InterfaceServlet</code> to be routed to the proper namespace controller.
 * 
 * @see org.opendof.tools.repository.interfaces.core.PersistenceController 
 * @see org.opendof.tools.repository.interfaces.core.CoreController 
 * 
 */
public class AllseenController implements PersistenceController
{
    private static final String AllseenNameSpace = "org.allseen";
    private static final String FreedesktopNameSpace = "org.freedesktop";
    private static final String IidAttribute = "<interface name=\"";
    
    private final Logger log;
    // guarded by volatile, single thread init called only once
    private volatile DataAccessor da;
    private volatile String repoType;
    private final Map<String, XMLTranslator> translators;
    private volatile boolean userGroupOnly;

    /**
     * Allseen Controller default constructor.
     */
    public AllseenController()
    {
        log = LoggerFactory.getLogger(getClass());
        translators = new HashMap<String, XMLTranslator>();
    }

    /* ************************************************************************
     * PersistenceController implementation
    **************************************************************************/
    @Override
    public void init(CoreController core, String repoType, DataAccessor da, boolean userGroupOnly) throws Exception
    {
        this.da = da;
        this.repoType = repoType;
        this.userGroupOnly = userGroupOnly;
        initTranslations();
    }
    
    private void initTranslations(){    	
        String htmlXsl = this.getClass().getClassLoader().getResource("translation/allseen/html-translation.xsl").toString();
        String javaXsl = this.getClass().getClassLoader().getResource("translation/allseen/java-translation.xsl").toString();
        
    	translators.put("XML", new RawXMLTranslator());
    	translators.put("HTML", new XSLTTranslator(htmlXsl, "HTML"));
    	translators.put("Java", new XSLTTranslator(javaXsl, "Java"));
    }

    @Override
    public boolean isSupportingRepository(String iid)
    {
        if(iid.startsWith(AllseenNameSpace) || iid.startsWith(FreedesktopNameSpace))
            return true;
        return false;
    }
    
    @Override
    public String getIidFromXml(String xml)
    {
        int sidx = xml.indexOf(IidAttribute);
        if (sidx == -1)
            return null;
        sidx += IidAttribute.length();
        int eidx = xml.indexOf('\"', sidx);
        return xml.substring(sidx, eidx);
    }
    
    @Override
    public void destroy()
    {
    }

    @Override
    public List<InterfaceData> listInterfaces(AuthenticatedUser user, String submitter, String accessGroup, Boolean published) throws Exception
    {
        InterfaceIterable ii = da.listInterfaces(repoType, submitter, accessGroup, published);
        Iterator<InterfaceData> iter = ii.iterator();
        List<InterfaceData> list = new ArrayList<InterfaceData>();
        while (iter.hasNext())
        {
        	InterfaceData data = iter.next();
            if(JiraGroupRules.verifyReadAccess(user, data, userGroupOnly))
            {
                data.setName(getInterfaceIdFromXml(data.xml));
                list.add(data);
            }
        }
        ii.close();
        return list;
    }

    @Override
    public List<InterfaceData> listSharedInterfaces(AuthenticatedUser user, String submitter, Boolean published) throws Exception
    {
        InterfaceIterable ii = da.listSharedInterfaces(repoType, submitter, published);
        Iterator<InterfaceData> iter = ii.iterator();
        List<InterfaceData> list = new ArrayList<InterfaceData>();
        while (iter.hasNext())
        {
            InterfaceData data = iter.next();
            list.add(data);
        }
        ii.close();
        return list;
    }

    @Override
    public InterfaceData getInterface(AuthenticatedUser user, String iid, String version) throws Exception
    {
        try
        {
            return da.getInterface(iid, version == null ? "1" : version);
        } catch (Exception e)
        {
            log.warn(getClass().getSimpleName() + ".getInterface failed iid: " + iid + " version: " + version);
            throw e;
        }
    }
    
    private String getInterfaceIdFromXml(String xml)  throws Exception
    {
        int sidx = xml.indexOf(interfacetag) + interfacetag.length();
        if (sidx == -1)
            return "";
        int eidx = xml.indexOf('\"', sidx);
        return xml.substring(sidx, eidx);
    }

    private static final String interfacetag = "<interface name=\"";
    
    @Override
    public InterfaceData addInterface(AuthenticatedUser user, InterfaceData iface, String subRepoRdn) throws SubmitterException, AlreadyExistsException, NotFoundException, InsufficentRightsException, Exception
    {
        if (iface.version == null)
            throw new Exception("An Interface version must be given");
        String iid = getInterfaceIdFromXml(iface.xml);
        
        try
        {
            da.getInterface(iid, iface.version);
            throw new AlreadyExistsException(iid + " " + iface.version);
        } catch (NotFoundException e)
        {
            if (iface.creationDate == null || iface.lastModifiedDate == null)
            {
                Date date = new Date();
                //@formatter:off
                iface = new InterfaceData(
                                iface.getName(), iface.iid, iface.xml, iface.version, iface.submitter, iface.accessGroup, 
                                iface.getRepoType(), iface.creationDate == null ? date : iface.creationDate,
                                iface.lastModifiedDate == null ? date : iface.lastModifiedDate, iface.publish);
                //@formatter:off
            }
            da.addInterface(iface);
            log.info(iid + " " + iface.version + " added to working data store");
            return null;
        }
    }

    @Override
    public void publishInterface(AuthenticatedUser user, String iid, String version) throws NotFoundException, Exception
    {
        throw new Exception("Allseen publish not currently supported");
    }

    @Override
    public void addInterfaceSharePermission(AuthenticatedUser user, String iid, String version, String email,
                                            SubmitterPermissions.PermissionLevel permissionLevel) {

    }

    @Override
    public void deleteInterface(AuthenticatedUser user, String iid, String version) throws NotFoundException, Exception
    {
        if (iid == null)
            throw new Exception("An Interface ID must be given");
        if (version == null)
            throw new Exception("An Interface version must be given");

        da.deleteInterface(iid, version, null);
        log.info(iid + " " + version + " deleted");
    }
    
    @Override
    public void updateInterface(AuthenticatedUser user, String iid, String xml, String version, SubmitterData submitter, SubmitterData accessGroup, Boolean published) 
                    throws NotFoundException, ImmutableException, Exception
    {
        if (version == null)
            throw new Exception("An Interface version must be given");
        
        String giid = null;
        String xiid = null;
        if(xml != null)
            xiid = getIidFromXml(xml);
        if(iid != null)
            giid = iid;
        if(giid == null && xiid == null)
            throw new InterfaceIdMatchException("iid and xml are both null");
        else if(giid == null)
                giid = xiid;
        else
            if(!giid.equals(xiid))
                throw new InterfaceIdMatchException("iid: " + giid + " xml: " + xiid);
        
        InterfaceData idata = da.getInterface(iid, version);
        
        if(Objects.equals(user.email,idata.submitter.email)) {
            //Check if the user does not have write access
            Integer userPermission = da.getPermission(repoType, iid, version, user.email);
            if(userPermission != SubmitterPermissions.PermissionLevel.WRITE.getValue()) {
                throw new InsufficentRightsException(); // TODO: only the original submitter can update at this time.
            }
        }
        
        //@formatter:off
        da.updateInterface(
            new InterfaceData( 
                null, idata.iid, idata.xml, version, 
                idata.submitter, idata.accessGroup, idata.getRepoType(),  
                idata.creationDate, new Date(), idata.publish));
        //@formatter:on
        log.info(iid + " " + version + " updated inf data store " + (idata.publish ? "Published" : "Working"));
    }
    
    @Override
    public SubRepositoryNode getSubRepositoryNode(String repoType, String rdn) throws NotFoundException, Exception
    {
        return new SubRepositoryNode(CoreController.AllseenRepo, CoreController.AllseenRepo, "root", 0, DataAccessor.AnonymousGroup);
    }
    
    @Override
    public void deleteSubRepositoryNode(AuthenticatedUser user, String parentRdn, SubRepositoryNode node) throws NotFoundException, Exception
    {
        throw new Exception("Allseen repository does not support deleteSubRepositoryNode");
    }

    @Override
    public void addSubRepositoryNode(AuthenticatedUser user, String parentRdn, SubRepositoryNode child) throws NotFoundException, Exception
    {
        throw new Exception("Not currently supported on Allseen Repository");
    }
    
    @Override
    public void updateSubRepositoryNode(AuthenticatedUser user, String rdn, SubRepositoryNode node) throws NotFoundException, Exception
    {
        throw new Exception("Not currently supported on Allseen Repository");
    }

	@Override
	public String getURLIdentifier(String standardInterfaceID) {
		try {
			return URLEncoder.encode(standardInterfaceID, "utf-8");
		} catch (UnsupportedEncodingException e) {
			log.warn("Failed to encode allseen interface identifier " + standardInterfaceID, e);
			return standardInterfaceID;
		}
	}

	@Override
	public String getStandardIdentifier(String urlInterfaceID) {
		try {
			return URLDecoder.decode(urlInterfaceID, "utf-8");
		} catch (UnsupportedEncodingException e) {
			log.warn("Failed to decode allseen interface identifier " + urlInterfaceID, e);
			return urlInterfaceID;
		}
	}

	@Override
	public boolean versionSupported() {
		return true;
	}

	@Override
	public synchronized List<String> getTranslations() {
		return new ArrayList<String>(translators.keySet());
	}

	@Override
	public XMLTranslator getTranslator(String translation) throws Exception {
		if(!translators.containsKey(translation)){
			throw new Exception("Allseen repository does not support translation " + translation);
		}
		return translators.get(translation);
	}

    @Override
    public boolean hasAllocateSupport()
    {
        return false;
    }
}
