/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.core;

import java.util.*;
import java.util.Map.Entry;

import org.opendof.tools.repository.interfaces.da.*;
import org.opendof.tools.repository.interfaces.da.DataAccessor.SubmitterIterable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@SuppressWarnings("javadoc")
public class CoreController
{
    public static final String OpenDofRepo = "opendof";
    public static final String AllseenRepo = "allseen";

    public static final AuthenticatedUser CliUser = new AuthenticatedUser(DataAccessor.CliAdminName, DataAccessor.CliAdminEmail, DataAccessor.CliAdminDescription, null);

    public static final String ContextRealBaseKey = "opendof.tools.interface.repository.context-base";
    public static final String TranslationFolderKey = "opendof.tools.interface.repository.translations";
    public static final String CatalogFolderUrlKey = "opendof.tools.interface.repository.catalog-url";
    public static final String UserGroupOnlyKey = "opendof.tools.interface.repository.groups.user-only";
    public static final String PropertiesKey = "opendof.tools.interface.repository.properties";

    public static final String IrServletConfigKey = "opendof.tools.interface.repository.config";
    public static final String PersistenceClassKey = "opendof.tools.interface.repository.persist.class";
    public static final String ControllerClassBaseKey = "opendof.tools.interface.repository.controller";

    public static final String IrServletConfigDefault = "ir.properties"; // relative to deployed context classes
    public static final String TranslationFolderDefault = "/translation/"; // relative to WebContent
    public static final String CatalogFolderDefault = "/catalog/"; // relative to WebContent
    public static final String UserGroupOnlyDefault = "false";
    private static final String PersistenceClassDefault = "org.opendof.tools.repository.interfaces.mysql.MysqlDataAccessor";

    private final Logger log;
    // guarded by self
    private final Map<String, RepoType> controllers;
    private volatile DataAccessor da = null;
    private volatile Properties properties = null;

    public CoreController()
    {
        log = LoggerFactory.getLogger(getClass());
        controllers = new HashMap<String, RepoType>();
    }

    public Properties getProperties()
    {
        return properties;
    }

    //TODO: in version 1.x, remove this two repo only support

    public volatile String opendofRepoType;
    public volatile String allseenRepoType;

    public void init(Properties properties) throws Exception
    {
        StringBuilder sb = new StringBuilder("\n" + getClass().getSimpleName() + " starting core initialization:");
        this.properties = properties;
        String msg = "ok";
        try
        {
            String value = properties.getProperty(PersistenceClassKey, PersistenceClassDefault);
            msg = "failed to instantiate DataAccessor: " + value;
            sb.append("\n\t" + PersistenceClassKey + "=").append(value);
            Class<?> clazz = Class.forName(value);
            da = (DataAccessor) clazz.newInstance();
            msg = "failed to initialize DataAccessor: " + value;
            da.init(properties);
            boolean userGroupOnly = Boolean.parseBoolean(properties.getProperty(CoreController.UserGroupOnlyKey, CoreController.UserGroupOnlyDefault));
            sb.append("\t userGroupOnly: ").append(userGroupOnly);

            List<Entry<String, String>> list = getPropertiesForBaseKey(ControllerClassBaseKey, properties);
            for (Entry<String, String> entry : list)
            {
                String[] values = entry.getValue().split(" ");
                msg = "failed to instantiate Controller: " + values[0];
                sb.append("\n\t" + entry.getKey() + "=" + values[0] + " browserKey: " + values[1]);
                clazz = Class.forName(values[0]);
                PersistenceController controller = (PersistenceController) clazz.newInstance();
                msg = "failed to initialize Controller: " + value;
                controller.init(this, values[1], da, userGroupOnly);
                controllers.put(values[1], new RepoType(values[1], controller));
                if (values[0].contains("OpendofController"))
                    opendofRepoType = values[1];
                else
                    allseenRepoType = values[1];
            }
            log.debug(getClass().getSimpleName() + sb.toString());
        } catch (Exception e)
        {
            log.info(sb.toString());
            log.error(msg, e);
            throw e;
        }
    }

    public List<String> getRepositoryList()
    {
        return new ArrayList<String>(controllers.keySet());
    }

    public List<String> getSupportingRepositories(String iid)
    {
        List<String> repoTypes = new ArrayList<String>();
        for (Entry<String, RepoType> entry : controllers.entrySet())
        {
            RepoType rtype = entry.getValue();
            if (rtype.controller.isSupportingRepository(iid))
                repoTypes.add(rtype.browserKey);
        }
        return repoTypes;
    }

    public List<String> getIidFromXmlForRepositories(String xml)
    {
        List<String> iids = new ArrayList<String>();
        for (Entry<String, RepoType> entry : controllers.entrySet())
        {
            RepoType rtype = entry.getValue();
            String iid = rtype.controller.getIidFromXml(xml);
            if (iid != null)
                iids.add(iid);
        }
        return iids;
    }

    // TODO: don't really want to expose this, but going for legacy fixup 
    public DataAccessor getDataAccessor()
    {
        return da;
    }

    public void destroy()
    {
        for (Entry<String, RepoType> entry : controllers.entrySet())
            entry.getValue().controller.destroy();
        if (da != null)
        {
            da.close();
        }
    }

    public void addSubmitter(AuthenticatedUser user, SubmitterData submitter) throws AlreadyExistsException, Exception
    {
        //TODO: specialize out a SubmitterController if/when future features require it
        if(!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for SubmitterData APIs");
        da.addSubmitter(submitter);
    }
    
    public void deleteSubmitter(AuthenticatedUser user, String submitter) throws AlreadyExistsException, Exception
    {
        //TODO: specialize out a SubmitterController if/when future features require it
        if(!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for SubmitterData APIs");
        da.deleteSubmitter(submitter);
    }
    
    public List<SubmitterData> listSubmitters(AuthenticatedUser user) throws AlreadyExistsException, Exception
    {
        //TODO: specialize out a SubmitterController if/when future features require it
        if(!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for SubmitterData APIs");
        List<SubmitterData> list = new ArrayList<SubmitterData>();
        SubmitterIterable subIter = da.listSubmitters();
        Iterator<SubmitterData> siter = subIter.iterator();
        while(siter.hasNext())
            list.add(siter.next());
        subIter.close();
        return list;
    }
    
    public void updateSubmitter(AuthenticatedUser user, String submitter, SubmitterData newSubmitter) throws AlreadyExistsException, Exception
    {
        //TODO: specialize out a SubmitterController if/when future features require it
        if(!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for SubmitterData APIs");
        SubmitterData sdata = da.getSubmitter(submitter);
        String name = newSubmitter.name == null ? sdata.name : newSubmitter.name;
        String email = newSubmitter.email == null ? sdata.email : newSubmitter.email;
        String description = newSubmitter.description == null ? sdata.description : newSubmitter.description;
        Date date = newSubmitter.date == null ? sdata.date : newSubmitter.date;
        SubmitterData updateData = new SubmitterData(name, email, description, date, sdata.group);
        da.updateSubmitter(submitter, updateData);
    }
    
    public void addGroup(AuthenticatedUser user, String manager, SubmitterData group) throws AlreadyExistsException, NotFoundException, Exception
    {
        //TODO: specialize out a SubmitterController if/when future features require it
        if(!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for SubmitterData APIs");
        da.addGroup(manager, group);
    }
    
    public void deleteGroup(AuthenticatedUser user, String group, String member, boolean force) throws AlreadyExistsException, Exception
    {
        //TODO: specialize out a SubmitterController if/when future features require it
        if(!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for SubmitterData APIs");
        if(member == null)
            da.deleteGroup(group, force);
        else
            da.removeSubmitterFromGroup(member, group);
    }
    
    public List<SubmitterData> listMembers(AuthenticatedUser user, String group) throws AlreadyExistsException, Exception
    {
        //TODO: specialize out a SubmitterController if/when future features require it
        if(!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for SubmitterData APIs");
        return da.listMembers(group);
    }
    
    public List<GroupData> listGroups(AuthenticatedUser user) throws AlreadyExistsException, Exception
    {
        //TODO: specialize out a SubmitterController if/when future features require it
        if(!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for SubmitterData APIs");
        return da.listGroups();
    }
    
    public void updateGroup(AuthenticatedUser user, String group, String adminEmail, String description, Date date) throws Exception
    {
        //TODO: specialize out a SubmitterController if/when future features require it
        if(!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for SubmitterData APIs");
        da.updateGroup(group, adminEmail, description, date);
    }
    
    public void addSubmitterToGroup(AuthenticatedUser user, String member, String group) throws NotFoundException, Exception
    {
        //TODO: specialize out a SubmitterController if/when future features require it
        if(!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for SubmitterData APIs");
        da.addSubmitterToGroup(member, group);
    }
    
    public SubmitterData getSubmitterData(String submitter) throws NotFoundException, Exception
    {
        return da.getSubmitter(submitter);
    }
    
    public SubmitterData getGroupData(String group) throws NotFoundException, Exception
    {
        return da.getGroup(group);
    }

    /**
     * Return a list of interfaces based on the given filters. This method will only return interfaces that the user or group owns.
     *
     * @param user User requesting the list.  Must not be null.
     * @param repo  The repository to return the list for.  If null list all known repositories.
     * @param submitter Filter the list based on the submitter.  If null don't filter on submitter.
     * @param accessGroup Filter the list based on the accessGroup.  If null don't filter on accessGroup.
     * @param published Filter the list based on the published flag. If null don't filter on published.
     * @return the filtered list.
     * @throws Exception if the list could not be obtained
     */
    public List<InterfaceData> getSubmitterInterfaces(AuthenticatedUser user, String repo, String submitter, String accessGroup, Boolean published) throws Exception
    {
        List<InterfaceData> rvalue = new ArrayList<InterfaceData>();
        if(repo == null)
        {
            for(Entry<String, RepoType> entry : controllers.entrySet())
            {
                RepoType repoType = getRepoType(entry.getKey());
                List<InterfaceData> list = repoType.controller.listInterfaces(user, submitter, accessGroup, published);
                rvalue.addAll(list);
            }
        }else
        {
            RepoType repoType = getRepoType(repo);
            rvalue = repoType.controller.listInterfaces(user, submitter, accessGroup, published);
        }
        Collections.sort(rvalue);
        return rvalue;
    }

    /**
     * Return a list of interfaces based on the given filters.
     * 
     * @param user User requesting the list.  Must not be null.
     * @param repo  The repository to return the list for.  If null list all known repositories.
     * @param submitter Filter the list based on the submitter.  If null don't filter on submitter.
     * @param accessGroup Filter the list based on the accessGroup.  If null don't filter on accessGroup.
     * @param published Filter the list based on the published flag. If null don't filter on published.
     * @return the filtered list.
     * @throws Exception if the list could not be obtained
     */
    public List<InterfaceData> getAllInterfaces(AuthenticatedUser user, String repo, String submitter, String accessGroup, Boolean published) throws Exception
    {
    	List<InterfaceData> rvalue = new ArrayList<InterfaceData>();
    	if(repo == null)
    	{
    		for(Entry<String, RepoType> entry : controllers.entrySet())
    		{
    	        RepoType repoType = getRepoType(entry.getKey());
    	        List<InterfaceData> list = repoType.controller.listInterfaces(user, submitter, accessGroup, published);
                rvalue.addAll(list);
                list = repoType.controller.listSharedInterfaces(user, submitter, published);
                rvalue.addAll(list);
    		}
    	}else
    	{
            RepoType repoType = getRepoType(repo);
            rvalue = repoType.controller.listInterfaces(user, submitter, accessGroup, published);
    	}
    	Collections.sort(rvalue);
    	return rvalue;
    }
    	
    public InterfaceData getInterface(AuthenticatedUser user, String repo, String iid, String version) throws Exception
    {
        RepoType repoType = getRepoType(repo);
        return repoType.controller.getInterface(user, iid, version);
    }

    public InterfaceData addInterface(AuthenticatedUser user, String repo, InterfaceData iface, String subRepoRdn) throws AlreadyExistsException, Exception
    {
        RepoType repoType = getRepoType(repo);
        return repoType.controller.addInterface(user, iface, subRepoRdn);
    }

    public boolean hasAllocateSupport(String repoType) throws NotFoundException, Exception
    {
        RepoType repo = getRepoType(repoType);
        return repo.controller.hasAllocateSupport();
    }
    
    public SubRepositoryNode getSubRepositoryNode(String repoType, String name) throws NotFoundException, Exception
    {
        RepoType repo = getRepoType(repoType);
        return repo.controller.getSubRepositoryNode(repo.browserKey, name);
    }

    public void addSubRepositoryNode(AuthenticatedUser user, String parentRdn, SubRepositoryNode child) throws NotFoundException, Exception
    {
        RepoType repo = getRepoType(child.getRepoType());
        repo.controller.addSubRepositoryNode(user, parentRdn, child);
    }

    public void deleteSubRepositoryNode(AuthenticatedUser user, String parentRdn, SubRepositoryNode node) throws NotFoundException, Exception
    {
        RepoType repo = getRepoType(node.getRepoType());
        repo.controller.deleteSubRepositoryNode(user, parentRdn, node);
    }

    public void updateSubRepositoryNode(AuthenticatedUser user, String rdn, SubRepositoryNode node) throws NotFoundException, Exception
    {
        RepoType repo = getRepoType(node.getRepoType());
        repo.controller.updateSubRepositoryNode(user, rdn, node);
    }

    public void updateInterface(AuthenticatedUser user, String repo, String iid, String xml, String version, SubmitterData submitter, SubmitterData accessGroup, Boolean published) throws NotFoundException, ImmutableException, Exception
    {
        RepoType repoType = getRepoType(repo);
        repoType.controller.updateInterface(user, iid, xml, version, submitter, accessGroup, published);
    }

    public void publishInterface(AuthenticatedUser user, String repo, String iid, String version) throws NotFoundException, Exception
    {
        RepoType repoType = getRepoType(repo);
        repoType.controller.publishInterface(user, iid, version);
    }

    public void deleteInterface(AuthenticatedUser user, String repo, String iid, String version) throws NotFoundException, Exception
    {
        RepoType repoType = getRepoType(repo);
        repoType.controller.deleteInterface(user, iid, version);
    }

    public void setPermissions(AuthenticatedUser user, String repoType, String iid, String version, String granteeEmail, int permission) throws NotFoundException, Exception{
        InterfaceData ifaceData = getRepoType(repoType).controller.getInterface(user, iid, version);
        if(ifaceData.submitter.email.equals(user.email) || Objects.equals(DataAccessor.CliAdminEmail,user.email)){
            da.setPermissions(repoType, iid, version, granteeEmail, permission);
        } else {
            throw new NotFoundException("iid: " + iid + " not found");
        }
    }

    public void revokePermissions(AuthenticatedUser user, String repoType, String iid, String version, String granteeEmail) throws NotFoundException, Exception {
        InterfaceData ifaceData = getRepoType(repoType).controller.getInterface(user, iid, version);
        if(ifaceData.submitter.email.equals(user.email) || Objects.equals(DataAccessor.CliAdminEmail,user.email)){
            da.revokePermissions(repoType, iid, version, granteeEmail);
        } else {
            throw new NotFoundException("iid: " + iid + " not found");
        }
    }

    public List<SubmitterPermissions> getPermissions(AuthenticatedUser user, String repoType, String iid, String version) throws NotFoundException, Exception{
        InterfaceData ifaceData = getRepoType(repoType).controller.getInterface(user, iid, version);
        if(ifaceData.submitter.email.equals(user.email) || Objects.equals(DataAccessor.CliAdminEmail,user.email)){
            return da.getPermissions(repoType, iid, version);
        } else {
            throw new NotFoundException("iid: " + iid + " not found");
        }
    }

    public Integer getPermission(String repoType, String iid, String version, String granteeEmail) throws NotFoundException, Exception{
        try{
            return da.getPermission(repoType, iid, version, granteeEmail);
        } catch(Exception e)
        {
            throw new NotFoundException("iid: " + iid + " not found");
        }
    }

    
    /**
     * Get a URL friendly version of an interface identifier from the standard interface identifier format.
     * 
     * @param repoType The repository type.
     * @param standardInterfaceID The URL interface identifier.
     * @return A URL friendly version of an interface identifier.
     * @throws Exception Thrown if unable to find repository type.
     */
    public String getURLIdentifier(String repoType, String standardInterfaceID) throws Exception{
    	return getRepoType(repoType).controller.getURLIdentifier(standardInterfaceID);
    }
    
    /**
     * Get a standard interface identifier from the URL interface identifier format.
     * 
     * @param repoType the repository type.
     * @param urlInterfaceID The standard interface identifier.
     * @return A standard interface identifier.
     * @throws Exception Thrown if unable to find repository type.
     */
    public String getStandardIdentifier(String repoType, String urlInterfaceID) throws Exception{
    	return getRepoType(repoType).controller.getStandardIdentifier(urlInterfaceID);
    }
    
    /**
     * Whether or not the specified repoType supports interface definition versions.
     * 
     * @param repoType The repository type.
     * @return Whether or not the specified repoType supports interface definition versions.
     * @throws Exception thrown if call fails
     */
    public boolean getVersionSupported(String repoType) throws Exception{
    	return getRepoType(repoType).controller.versionSupported();
    }
    
    /**
     * Get a list of interface definition translations supported by the specified repository.
     * 
     * @param repoType The repository type.
     * @return Valid interface translations
     * @throws Exception Thrown if repoType is invalid or not accessible.
     */
    public List<String> getTranslations(String repoType) throws Exception{
    	return getRepoType(repoType).controller.getTranslations();
    }

    public RepoType getRepoType(String repoType) throws NotFoundException, Exception
    {
        RepoType repo = controllers.get(repoType);
        if (repo == null)
        {
            String msg = getClass().getSimpleName() + ".getInterface unknown repoType: " + repoType;
            log.warn(msg);
            throw new Exception(msg);
        }
        return repo;
    }
    
    public XMLTranslator getTranslator(String repoType, String translation) throws Exception{
        return getRepoType(repoType).controller.getTranslator(translation);
    }

    
    /* ************************************************************************
     * Helpers
    **************************************************************************/

    public class RepoType
    {
        public String browserKey;
        public PersistenceController controller;

        private RepoType(String browserKey, PersistenceController controller)
        {
            this.browserKey = browserKey;
            this.controller = controller;
        }
    }

    public static List<Entry<String, String>> getPropertiesForBaseKey(String baseKey, Properties properties)
    {
        ArrayList<Entry<String, String>> entries = new ArrayList<Entry<String, String>>();
        Hashtable<Integer, StringPair> orderingMap = new Hashtable<Integer, StringPair>();

        int found = 0;
        for (Entry<Object, Object> entry : properties.entrySet())
        {
            String key = (String) entry.getKey();
            int index = 0;
            if (key.startsWith(baseKey))
            {
                ++found;
                char[] chars = key.toCharArray();
                if (Character.isDigit(chars[chars.length - 1]))
                {
                    int strIndex = 0;
                    for (int i = chars.length - 1; i >= 0; i--)
                    {
                        if (!Character.isDigit(chars[i]))
                        {
                            strIndex = i + 1;
                            break;
                        }
                    }
                    index = Integer.parseInt(key.substring(strIndex));
                }
                orderingMap.put(index, new StringPair(entry));
            }
        }
        int i = 0;
        int hit = 0;
        do
        {
            StringPair pair = orderingMap.get(i);
            if (pair != null)
            {
                entries.add(pair);
                ++hit;
            }
            ++i;
        } while (hit < found);
        return entries;
    }

    public static class StringPair implements Entry<String, String>
    {
        private final String key;
        private String value;

        public StringPair(Entry<Object, Object> entry)
        {
            key = (String) entry.getKey();
            value = (String) entry.getValue();
        }

        @Override
        public String setValue(String value)
        {
            String old = this.value;
            this.value = value;
            return old;
        }

        // @formatter:off
        @Override
        public String getKey(){return key;}
        @Override
        public String getValue(){return value;}
        @Override
        public String toString(){return key + "=" + value;} // @Formatter:on
    }
}