/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.core;

import java.util.List;

import org.opendof.tools.repository.interfaces.da.*;

/**
 * The XML Persistence Controller Interface.
 * <p>Each unique repository type must have a custom persistence controller which enforces any required
 * rules for that repository type for the various XML persistence access.
 */
public interface PersistenceController
{
    /**
     * Initialize method.
     * <p>The controller implementation should be ready to accept calls on any 
     * of the other interface methods upon returning from this call. 
     * @param core  The CoreController that is calling you.  Will never be null.
     * @param repoType this controllers repository type.
     * @param da The persistence DataAccessor that you should use. 
     * @param userGroupOnly if true do not allow anonymous users to do anything. 
     * @throws Exception if unable to initialize the controller.
     */
    void init(CoreController core, String repoType, DataAccessor da, boolean userGroupOnly) throws Exception;
    
    /**
     * Is iid one I support.
     * @param iid the iid in question
     * @return true if its one I understand, false otherwise.
     */
    boolean isSupportingRepository(String iid);

    /**
     * Return the Interface ID of the XML.
     * @param xml the interface definition XML.
     * @return The Interface ID of the definition, if found in xml, null otherwise
     */
    String getIidFromXml(String xml);

    /**
     * Destroy yourself method.
     * <p> The controller implementation should release all its resources.
     * The <code>CoreController</code> will not call any methods on this interface
     * again upon returning from this call.
     */
    void destroy();
    
    /**
     *  List Interfaces method.
     *  
     *  <p>The list is affected by the <code>RequestData.AutenticatedUser</code> 
     *  which is used in conjunction with the following groups:
     *  <ul>
     *  <li>anonymous (an Interface Repository default group)
     *  <li>user      (same as above)
     *  <li>custom    (created by submitters)
     *  </ul>
     *   
     *  If the interface is in the anonymous group it will always be added to the list
     *  unless the site configuration has the <code>CoreController.UserGroupOnlyKey</code>
     *  set to true.
     *   
     *  <p>If the interface is in the user group it will only be added to the list if 
     *  <code>RequestData.AutenticatedUser</code> is non-null.
     *  
     *  <p>If the interface is in a submitter created custom group it will only be added 
     *  to the list if the request <code>RequestData.AutenticatedUser</code> is a member 
     *  of that custom group.
     *  
     *  <p>if the interface is not one of the two constant groups or a custom group then 
     *  the interface will only be added to the list if the request submitter matches
     *  the <code>RequestData.AutenticatedUser</code>.
     * @param user the user.  Maybe null for anonymous.
     * @param submitter Filter the list based on the submitter.  If null don't filter on submitter.
     * @param accessGroup Filter the list based on the accessGroup.  If null don't filter on accessGroup.
     * @param published Filter the list based on the published flag. If null don't filter on published.
     * @return List of filtered InterfaceData
     * @throws Exception if the list could not be obtained
     */
    //FIXME: needs some kind of sequencer
    List<InterfaceData> listInterfaces(AuthenticatedUser user, String submitter, String accessGroup, Boolean published) throws Exception;

    /**
     * List Shared Interfaces method.
     *
     * <p>The list is all the interfaces the user has at least read access. This method will not list interfaces in which the user is the owner.</p>
     *
     * @param user the user. May be null for anonymous.
     * @param userEmail Filter the list based on the submitter. May not be null.
     * @param published Filter the list based on the published flag. If null don't filter on published.
     * @return List of shared InterfaceData
     * @throws Exception thrown if the list of shared interfaces could not be found.
     */
    List<InterfaceData> listSharedInterfaces(AuthenticatedUser user, String userEmail, Boolean published) throws Exception;

    /**
     * Get the requested Interface.
     * @param user the user.  Maybe null for anonymous.
     * @param iid the iid of the interface to return.  Must not be null.
     * @param version the version of the iid to return.  If null version "1" is used. 
     * @return the Interface data.
     * @throws Exception if the interface data could not be obtained.
     */
    InterfaceData getInterface(AuthenticatedUser user, String iid, String version) throws Exception;

    /**
     * Add a Working Interface.
     * @param user the user.  Maybe null for anonymous.
     * @param iface data for the interface to be added.
     * @param subRepoRdn the relative distinguished name of the sub-repository allocation pool.
     * @return the given iid to add if not null.  Otherwise allocate a new iid.
     * @throws SubmitterException if the given submitter of the interface is not setup properly. 
     * @throws AlreadyExistsException if trying to add a working interface ID/version that already exists.
     * @throws InterfaceIdMatchException if iid and xml are both supplied and they don't match.
     * @throws NotFoundException given submitter is not known. 
     * @throws Exception if the interface could not be added.
     */
    InterfaceData addInterface(AuthenticatedUser user, InterfaceData iface, String subRepoRdn) throws Exception;
    
    /**
     * Update a working or published interface definition.
     * <p>The iid and xml arguments have the following rules:
     * <ul>
     * <li>if iid is null, look to xml for iid.
     * <li>if iid is null, and xml is null throw InterfaceIdMatchException.
     * <li>if iid is given and xml is null, use iid.
     * <li>if iid is given and xml is given the two must match or throw InterfaceIdMatchException.
     * </ul>
     * @param user the user.  Maybe null for anonymous.
     * @param iid the ID being updated.  If null, the iid is obtained from the xml
     * @param xml the xml to push. if null no change otherwise set the given xml.
     * @param version the version of the xml to push; 
     * @param submitter new submitter.  If null, no change otherwise set the given submitter.
     * @param accessGroup new accessGroup.  If null, no change otherwise set the given accessGroup. 
     * @param published new published value.  If null, no change otherwise set the given flag. 
     * @throws SubmitterException if the given submitter of the interface is setup properly. 
     * @throws NotFoundException if the requested InterfaceID is not in the repository.
     * @throws InterfaceIdMatchException if iid and xml are both supplied and they don't match.
     * @throws ImmutableException if the update is for a published interface and invalid
     * portions of that interface would be modified.
     * @throws Exception if the interface could not be updated.
     */
    void updateInterface(AuthenticatedUser user, String iid, String xml, String version, SubmitterData submitter, SubmitterData accessGroup, Boolean published) throws Exception;
    
    /**
     * Update a working or published interface definition.
     * @param user the user.  Maybe null for anonymous.
     * @param iid the interface ID to publish; 
     * @param version the version of the iid to publish; 
     * @throws NotFoundException if the requested InterfaceID is not in the repository.
     * @throws ImmutableException if the update is for a published interface and invalid
     * portions of that interface would be modified.
     * @throws Exception if the interface could not be updated.
     */
    void publishInterface(AuthenticatedUser user, String iid, String version) throws Exception;

    /**
     * Update a working interface definition to shared.
     * @param user the user.  Maybe null for anonymous.
     * @param iid the interface ID to publish.
     * @param version the version of the iid to publish.
     * @param permissionLevel the permission to add.
     * @param email the email address of the user that the InterfaceID should be shared with.
     * @throws NotFoundException if the requested InterfaceID is not in the repository.
     * @throws ImmutableException if the update is for a published interface and invalid
     * portions of that interface would be modified.
     */
    void addInterfaceSharePermission(AuthenticatedUser user, String iid, String version,
                                     String email, SubmitterPermissions.PermissionLevel permissionLevel)
            throws NotFoundException, ImmutableException;
    /**
     * Delete a working or published interface definition.
     * @param user the user.  Maybe null for anonymous.
     * @param iid the interface ID to delete; 
     * @param version the version of the iid to delete; 
     * @throws NotFoundException if the requested InterfaceID is not in the repository.
     * @throws Exception if the interface could not be updated.
     */
    void deleteInterface(AuthenticatedUser user, String iid, String version) throws Exception;

    /**
     * Reports if the Repository supports IID allocation.
     * @return true if supported, false otherwise
     */
    boolean hasAllocateSupport();
    
    /**
     * Get Sub-Repository Node.
     * @param repoType the repository type to apply the RDN against.
     * @param rdn the relative distinguished name (RDN). May be null.
     * @return the requested node.
     * @throws NotFoundException given repoType/name is not known. 
     * @throws Exception if the node could not be obtained.
     */
    SubRepositoryNode getSubRepositoryNode(String repoType, String rdn) throws Exception;
    
    /**
     * Add a Sub-Repository Node.
     * @param user the user.  Maybe null for anonymous.
     * @param parentRdn the relative distinguished name (RDN) to the parent. May be null (depth 0 node).
     * @param child the node to be added to the given parent.  Must not be null.
     * @throws AlreadyExistsException if the parent/child node already exists.
     * @throws NotFoundException given repoType/name is not known. 
     * @throws Exception if the node could not be obtained.
     */
    void addSubRepositoryNode(AuthenticatedUser user, String parentRdn, SubRepositoryNode child) throws Exception;
    
    /**
     * Delete a Sub-Repository Node.
     * @param user the user.  Maybe null for anonymous.
     * @param parentRdn the rdn to the parent node holding the child to be deleted.
     * @param node the child information needed to delete the node.
     * @throws AlreadyExistsException if the parent/child node already exists.
     * @throws NotFoundException given repoType/name is not known. 
     * @throws Exception if the node could not be obtained.
     */
    void deleteSubRepositoryNode(AuthenticatedUser user, String parentRdn, SubRepositoryNode node) throws Exception;
    
    /**
     * Update a Sub-Repository Node's Label.
     * @param user the user.  Maybe null for anonymous.
     * @param rdn the rdn to update.  Must not be null.
     * @param node the update data.  Must not be null.
     * @throws NotFoundException given repoType/name is not known. 
     * @throws Exception if the node could not be obtained.
     */
    void updateSubRepositoryNode(AuthenticatedUser user, String rdn, SubRepositoryNode node) throws Exception;
    
    /**
     * Get a URL friendly version of an interface identifier from the standard interface identifier format.
     * 
     * The URL friendly interface identifier must not contain any reserved or invalid URL characters.
     * 
     * @param standardInterfaceID The URL interface identifier.
     * @return A URL friendly version of an interface identifier.
     */
    String getURLIdentifier(String standardInterfaceID);
    
    /**
     * Get a standard interface identifier from the URL interface identifier format.
     * @param urlInterfaceID The standard interface identifier.
     * @return A standard interface identifier.
     */
    String getStandardIdentifier(String urlInterfaceID);
    
    /**
     * Whether or not the controller supports versions.
     * 
     *  @return true if interface definition versions are supported by the controller. False otherwise.
     */
    boolean versionSupported();
    
    /**
     * Get a list of interface definition translations supported by this repository.
     *  
     * @return The list of translations.
     */
    List<String> getTranslations();
    
    /**
     * Get an XML translator for the requested translation format.
     * 
     * @param translation The requested translation format.
     * @return The XMLTranslator.
     * @throws Exception if the translator could not be obtained.
     */
    XMLTranslator getTranslator(String translation) throws Exception;
}
