/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.da;

import java.util.Date;
import java.util.List;
import java.util.Properties;

@SuppressWarnings("javadoc")
public interface DataAccessor
{
    String NonServletInitKey = "opendof.tools.repository.persist.cli-init";
    
    String AdminName = "OpenDOF Admin";
    String AdminEmail = "admin@opendof.org";
    String AdminDescription = "OpenDOF Manager, general inquiries only";
    
    String AnonymousGroupName = "anonymous";
    String AnonymousGroupEmail = AnonymousGroupName;
    String AnonymousGroupDescription = "Anonymous users group";
    
    String UserGroupName = "user";
    String UserGroupEmail = UserGroupName;
    String UserGroupDescription = "Authenticated users group";

    String PrivateGroupName = "private";
    String PrivateGroupEmail = PrivateGroupName;
    String PrivateGroupDescription = "Private interface";
    
    String OpenDofAllocatorGroupName = "opendof-allocator";
    String OpenDofAllocatorGroupEmail = OpenDofAllocatorGroupName;
    String OpenDofAllocatorGroupDescription = "DOF IID 1 and 2 byte size allocation group";
    
    String CliAdminName = "cli-admin";
    String CliAdminEmail = CliAdminName;
    String CliAdminDescription = "Root user CLI Privileges";
    
    //@formatter:off
    SubmitterData AnonymousGroup = new SubmitterData(
                    AnonymousGroupName, AnonymousGroupEmail, 
                    AnonymousGroupDescription);
    
    SubmitterData UserGroup = new SubmitterData(
                    UserGroupName, UserGroupEmail, 
                    UserGroupDescription);
    
    SubmitterData PrivateGroup = new SubmitterData(
                    PrivateGroupName, PrivateGroupEmail, 
                    PrivateGroupDescription);
    
    SubmitterData OpenDofAllocationGroup = new SubmitterData(
                    OpenDofAllocatorGroupName, OpenDofAllocatorGroupEmail, 
                    OpenDofAllocatorGroupDescription);
    
    SubmitterData OpendofAdmin = new SubmitterData(
                    AdminName, AdminEmail, 
                    AdminDescription);
    //@formatter:on
    
	void init(Properties properties) throws Exception;
	void close();

	// submitter
	
	/**
	 * @param submitterData the submitter information.
	 * @throws AlreadyExistsException if the given submitter already exists
	 * @throws Exception if the submitter could not be added
	 */
    void addSubmitter(SubmitterData submitterData) throws AlreadyExistsException, Exception;
    
    /**
     * @param submitterEmail the requested submitter.
     * @return the requested submitter data.
     * @throws NotFoundException if the given submitter already exists
     * @throws Exception if the submitter could not be added
     */
	SubmitterData getSubmitter(String submitterEmail) throws NotFoundException, Exception;
	
    /**
     * @param submitterEmail the requested submitter.
     * @param submitterData the submitter information.
     * @throws NotFoundException if the given submitter already exists
     * @throws Exception if the submitter could not be updated
     */
	void updateSubmitter(String submitterEmail, SubmitterData submitterData) throws NotFoundException, Exception;
	
    /**
     * @param submitterEmail the requested submitter.
     * @throws Exception if the submitter could not be deleted
     */
    void deleteSubmitter(String submitterEmail) throws Exception;
    
    /**
     * Warning, potential persistence memory leak, caller must call close on the 
     * iterable or iterator obtained from returned Iterable.
     * @return Iterable from which an iterator of submitters can be obtained.
     * @throws Exception if the list can not be obtained
     */
    SubmitterIterable listSubmitters() throws Exception;

	// group
    void addGroup(String managerEmail, SubmitterData group) throws AlreadyExistsException, NotFoundException, Exception;
    SubmitterData getGroup(String group) throws NotFoundException, Exception;
    void updateGroup(String group, String adminEmail, String description, Date date) throws NotFoundException, Exception;
    void deleteGroup(String group, boolean force) throws NotFoundException, Exception;
	void addSubmitterToGroup(String member, String group) throws NotFoundException, Exception;
    void removeSubmitterFromGroup(String member, String group) throws Exception;
    boolean isMember(String member, String group) throws Exception;
    
    /**
     * Warning, potential persistence memory leak, caller must call close on the 
     * iterable or iterator obtained from returned Iterable.
     * @param group The group to list members
     * @return Iterable from which an iterator of submitters can be obtained.
     * @throws NotFoundException if the group does not exist in the data store
     * @throws Exception if the list can not be obtained
     */
    List<SubmitterData> listMembers(String group) throws NotFoundException, Exception;
    /**
     * @return List of all submitters in the group.
     * @throws Exception if the list can not be obtained
     */
    List<GroupData> listGroups() throws Exception;

	// interface
    void addInterface(InterfaceData interfaceData) throws AlreadyExistsException, Exception;
	InterfaceData getInterface(String iid, String version) throws NotFoundException, Exception;
    void updateInterface(InterfaceData interfaceData) throws NotFoundException, Exception;
    void deleteInterface(String iid, String version, SubRepositoryNode node) throws Exception;


    void setPermissions(String repoType, String iid, String version, String granteeEmail, int permission) throws NotFoundException, Exception;
	void revokePermissions(String repoType, String iid, String version,  String granteeEmail) throws NotFoundException, Exception;
    List<SubmitterPermissions> getPermissions(String repoType, String iid, String version) throws NotFoundException, Exception;
    Integer getPermission(String repoType, String iid, String version, String granteeEmail) throws NotFoundException, Exception;

  
    /**
     * Warning, potential persistence memory leak, caller must call close on the 
     * iterable or iterator obtained from returned Iterable.
     * @param repoType the repository type to return interfaces for.  Must not be null.
     * @param submitterName the stored submitterName to return interfaces for. If null, do not filter on submitter.
     * @param accessGroup the access group to return interfaces for. If null, do not filter on access group.
     * @param published the published state to return interfaces for.  If null, do not filter on published.
     * @return Iterable from which an iterator of interfaces can be obtained.
     * @throws Exception if the list can not be obtained
     */
    InterfaceIterable listInterfaces(String repoType, String submitterName, String accessGroup, Boolean published) throws Exception;

    /**
     * Lists all the interfaces that are shared with the user.
     * @param repoType the repository type to return interfaces for.  Must not be null.
     * @param userEmail the email used as the identifier for the authenticated user.
     * @param published the published state to return interfaces for.  If null, do not filter on published
     * @return Iterable from which an iterator of interfaces can be obtained.
     * @throws Exception if the list cannot be obtained.
     */
    InterfaceIterable listSharedInterfaces(String repoType, String userEmail, Boolean published) throws Exception;

    // reservation
    
    /**
     *  Allocate an IID from the named sub-repository node's pool.
     * @param iface Required interface information for insertion to the interface table. Must not be null
     * @param rdn the Relative Distinguished Name of the allocation pool node.  Must not be null
     * @return the interfaceData added to the dn.
     * @throws NotFoundException if the given rdn was not found.
     * @throws Exception if the node can not be obtained.
     */
    InterfaceData allocateInterfaceId(InterfaceData iface, String rdn) throws NotFoundException, Exception;
    
    
    /**
     * Return the requested node.
     * @param repoType the repository type. Must not be null.
     * @param rdn the '/' separated relative distinguished name from the repoType root.  Must not be null.
     * @return the node requested in the rdn.
     * @throws NotFoundException if the given rdn was not found.
     * @throws Exception if the node could not be retrieved
     */
    SubRepositoryNode getSubRepoNode(String repoType, String rdn) throws NotFoundException, Exception;
    
    /**
     * Add a SubRepo node.
     * @param parentRdn the parent to add the node to. May be null.
     * @param child the child to be added to given parent. Must not be null.
     * @throws AlreadyExistsException if the node already exists.
     * @throws NotFoundException if the parentRdn can not be found.
     * @throws Exception if the node could not be added.
     */
    void addSubRepoNode(String parentRdn, SubRepositoryNode child) throws AlreadyExistsException, NotFoundException, Exception;
    void deleteSubRepoNode(String parentRdn, SubRepositoryNode node) throws NotFoundException, Exception;
    void updateSubRepoNode(String rdn, String repoType, SubRepositoryNode node) throws NotFoundException, Exception;
        
    /* ************************************************************************
     * List Iterable interface declarations and iid size enum 
    **************************************************************************/
    interface SubmitterIterable extends Iterable<SubmitterData>
    {
        void close() throws Exception;
    }

    interface SubmitterPermissionsIterable extends Iterable<SubmitterPermissions>
    {
        void close() throws Exception;
    }

    interface GroupIterable extends Iterable<SubmitterData>
    {
        void close() throws Exception;
    }
    
    interface InterfaceIterable extends Iterable<InterfaceData>
    {
        void close() throws Exception;
    }
    
    enum SizeCategory
    {
        OneByte, TwoByte, FourByte;
        
        public static SizeCategory getSize(int ordinal)
        {
            switch(ordinal)
            {
                case 0:
                    return OneByte;
                case 1 :
                    return TwoByte;
                case 2:
                    return FourByte;
                default:
                    throw new IllegalArgumentException("ordinal: " + ordinal + " is not a valid size category value");
            }
        }
    }
}
