package org.opendof.tools.repository.interfaces.opendof;

import java.io.ByteArrayInputStream;
import java.io.Writer;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.TreeMap;

import javax.xml.parsers.SAXParserFactory;

import org.apache.xerces.jaxp.SAXParserFactoryImpl;
import org.apache.xerces.jaxp.SAXParserImpl;
import org.apache.xerces.util.XMLCatalogResolver;
import org.opendof.core.oal.DOFInterfaceID;
import org.opendof.tools.repository.interfaces.core.XMLTranslator;
import org.opendof.tools.repository.interfaces.opendof.iface.compute.CodeBlock;
import org.opendof.tools.repository.interfaces.opendof.iface.compute.EventInstance;
import org.opendof.tools.repository.interfaces.opendof.iface.compute.ExceptionInstance;
import org.opendof.tools.repository.interfaces.opendof.iface.compute.Form;
import org.opendof.tools.repository.interfaces.opendof.iface.compute.MethodInstance;
import org.opendof.tools.repository.interfaces.opendof.iface.compute.PropertyInstance;
import org.opendof.tools.repository.interfaces.opendof.iface.compute.TypeInstance;
import org.opendof.tools.repository.interfaces.opendof.saxParser.InterfaceElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler.OpendofErrorHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.event.EventElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.event.EventsElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.exception.ExceptionElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.exception.ExceptionsElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.method.MethodElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.method.MethodsElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.property.PropertiesElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.property.PropertyElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.DofTypeBase;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

/**
 * Translate xml file DOF interface specifications to the set of text that describes the interface, and is suitable to be placed in a C-compatible .h file. 
 *
 */
public class CLanguageTranslator implements XMLTranslator {
    private final Logger log;

    private final String catalogPath;
    private TreeMap<Short, TypeInstance> mapTypeInstances; // TreeMap allows descendingMap use, as an option
    private HashMap<Short, Short> mapCodeBlockTypeReferences;
    private TreeMap<Short, PropertyInstance> mapPropertyInstances;
    private TreeMap<Short, MethodInstance> mapMethodInstances;
    private TreeMap<Short, EventInstance> mapEventInstances;
    private TreeMap<Short, ExceptionInstance> mapExceptionInstances;
    private String interfaceName = null;
    
	/**
	 * Constructor
	 * 
	 * @param catalogPath the catalog path
	 */
    public CLanguageTranslator(String catalogPath) {
        log = LoggerFactory.getLogger(getClass());
		this.catalogPath = catalogPath;
		this.mapTypeInstances = new TreeMap<>();
	    this.mapCodeBlockTypeReferences = new HashMap<>();
		this.mapPropertyInstances = new TreeMap<>();
		this.mapMethodInstances = new TreeMap<>();
		this.mapEventInstances = new TreeMap<>();
		this.mapExceptionInstances = new TreeMap<>();
	}

    /**
     * Compute the type reference used in the final code block, from the xml-file supplied type reference value 
     * @param xmlSuppliedTypeReference the xml supplied type reference
     * @return translated type reference value
     * @throws Exception thrown if the code block could not be determined
     */
    public short getCodeBlockTypeReference(short xmlSuppliedTypeReference) throws Exception {
    	Short codeBlockTypeReference = this.mapCodeBlockTypeReferences.get(xmlSuppliedTypeReference);
    	if (codeBlockTypeReference == null)
    		throw new Exception("code block type reference lookup failed");
    	return codeBlockTypeReference.shortValue();
    }
    
	@Override
	public void toHTML(String xml, Writer writer) throws Exception {
		// TODO Auto-generated method stub
	}

	@Override
	public boolean hasLanguageSupport() {
		return false;
	}
	
	@Override
	public boolean hasRawSupport() {
		return true;
	}

	@Override
	public String getTranslationName() {
		return "C";
	}

	@Override
	public boolean hasHTMLSupport() {
		return false;
	}

	@Override
	public String toRaw(String xml, Writer writer) throws Exception {
//		log.debug("DEBUG mode LOG MESSAGE: CLanguageTranslator.toRaw() called");

        // TODO: from a parameter (?), identify requested language; throw exception if not available; add support later
        // note: this reflects the value true/false returned by this.hasLanguageSupport()
        String strRequestedLanguage = "en";

        // process xml to establish InterfaceElement class object
        OpenDofSaxHandler handler;
        {
	        // convert xml to InputSource
	        byte[] xmlBytes = xml.getBytes();
	        ByteArrayInputStream bais = new ByteArrayInputStream(xmlBytes);
	        InputSource inSource = new InputSource(bais);
	        
	        // setup oasis catalog
	        XMLCatalogResolver resolver = new XMLCatalogResolver();
	        resolver.setPreferPublic(true);
	        resolver.setCatalogList(new String[] { this.catalogPath });
	
	        // parse xml
//          System.setProperty("javax.xml.parsers.SAXParserFactory", "org.apache.xerces.jaxp.SAXParserFactoryImpl"); // System already has this property set
	        SAXParserFactoryImpl spf = (SAXParserFactoryImpl) SAXParserFactory.newInstance();
	        spf.setNamespaceAware(true);
	        spf.setValidating(false); // can not seem to get cataloging to work
	        SAXParserImpl saxParser = (SAXParserImpl) spf.newSAXParser();
	        XMLReader xmlReader = saxParser.getXMLReader();
	        xmlReader.setProperty("http://apache.org/xml/properties/internal/entity-resolver", resolver);
	        xmlReader.setErrorHandler(new OpendofErrorHandler());
	        handler = new OpenDofSaxHandler();
//          OpenDofSaxHandler.setPublish(false); // default is false; REVIEW: true not needed, we are only consuming the xml
	        xmlReader.setContentHandler(handler);
	        xmlReader.parse(inSource); // modifies content in handler
        }

        // note: if an element of the xml file (in handler) does not include "en" for language, try to output whichever language is present
        //       if more than one non-en language is available, use the first one

        
	    // compute and store the C Language information for the OpenDOF interface described in the InterfaceElement class object
        Form.initializeNextDistinguishingCharacter();
        boolean success = false;
        StringBuilder buildAcc = null;
        try {
            // setup highest level
            InterfaceElement ie = handler.getInterface(); // ie holds all required information
//			ie.validate(); //No need to do full validation as we are reading an existing interface.
            MetadataElements mde = ie.getMetadata();

            // get interface name, e.g. "Status;" announce it in a header
            this.interfaceName = Form.getInterfaceElementDisplayName(mde, strRequestedLanguage);
            if (this.interfaceName==null || this.interfaceName.isEmpty())
                this.interfaceName = "Interface_" + Form.nextDistinguishingCharacter();
            buildAcc = new StringBuilder(); // Acc: accumulate text, from empty
            buildAcc.append("/* C Language Information for " + this.interfaceName + " generated on " + new Date() + " */" + Form.newline +
                            Form.newline);

            // setup five reports, in their C comment block
			this.mapTypeInstances.clear();  // in case we are called more than once
			this.mapCodeBlockTypeReferences.clear();
			this.mapPropertyInstances.clear();
			this.mapMethodInstances.clear();
			this.mapEventInstances.clear();
			this.mapExceptionInstances.clear();
			TypedefsElement te = ie.getTypedefs();
			List<DofTypeBase> typeElements = te!=null ? te.getTypeDefs() : null;
			PropertiesElement pe = ie.getProperties();
			List<PropertyElement> propertyElements = pe!=null ? pe.getProperties() : null;
			MethodsElement me = ie.getMethods();
			List<MethodElement> methodElements = me!=null ? me.getMethods() : null;
			EventsElement ee = ie.getEvents();
			List<EventElement> eventElements = ee!=null ? ee.getEvents() : null;
			ExceptionsElement expe = ie.getExceptions();
			List<ExceptionElement> exceptionElements = expe!=null ? expe.getExceptions() : null;

			// start the five reports, in their C comment block
			buildAcc.append("/*" + Form.newline);

			// report Types information in comments block (while also filling this.mapTypeInstances and this.mapCodeBlockTypeReferences 
			// 		note: this section must be early, because others use this.mapTypeInstances
			buildAcc.append(Form.indentLevel1 + "Types" + Form.newline);
			if (!typeElements.isEmpty()) {
			    for (DofTypeBase typeElement : typeElements) {
			    	// typeElement: from original xml.parse() operation; extract TypeInstance object from typeElement
			    	TypeInstance typeInstance = new TypeInstance(typeElement, strRequestedLanguage, this.mapTypeInstances);
			    	
				    // Store in as received order, for displaying type information in C comment block, then later for forming interface definition bytes. 
			    	//     Also for use when properties, and similar, refer to types, i.e. by reference value; interface guarantees unique xmlSuppliedTypeReference.
			    	short xmlSuppliedTypeReference = typeInstance.getTypeReference(); // e.g. 0 (is known as type-id in the xml)
			    	// the above is not the registered Type Identifier from Protocol Spec PDU Type.23 (e.g. uint8=>00, uint16=>02, structure=0x7801)
				    this.mapTypeInstances.put(xmlSuppliedTypeReference, typeInstance);
			    }
			    
			    // translate TypeInstance xml supplied type reference to code block type reference
			    short overrideRef = 0;
			    Iterator<Entry<Short, TypeInstance>> iterRef = this.mapTypeInstances.entrySet().iterator();
			    while (iterRef.hasNext()) {
			    	TypeInstance typeInstance = iterRef.next().getValue();
			    	short xmlSpecifiedTypeReference = typeInstance.getTypeReference(); // xml guarantees unique ref value
			    	this.mapCodeBlockTypeReferences.put(xmlSpecifiedTypeReference, overrideRef++);
			    }
			    
			    // output info to C comment block; interface type reference ordering
			    Iterator<Entry<Short, TypeInstance>> iter = this.mapTypeInstances.entrySet().iterator();
			    while (iter.hasNext()) {
			    	TypeInstance typeInstance = iter.next().getValue();
			    	
//			    	// temporarily
//			    	short xmlSpecifiedRef = typeInstance.getTypeReference();
//			    	short codeBlockTypeReference = this.getCodeBlockTypeReference(xmlSpecifiedRef);
//			    	buildAcc.append(xmlSpecifiedRef + " " + codeBlockTypeReference);
			    	
			        buildAcc.append(Form.indentLevel2 + typeInstance.getCodeName() + ": "+ typeInstance.getDecoratedTypeName() + Form.newline +
			        		        Form.justifyDescription(typeInstance.getName() + " - " + typeInstance.getDescription()) + // trailing newline is included
	                                Form.newline);
			    }
			}

			// report Properties information in comments block
			buildAcc.append(Form.indentLevel1 + "Properties" + Form.newline);
			if (propertyElements!=null && !propertyElements.isEmpty()) {
			    for (PropertyElement propertyElement : propertyElements) {
			    	// propertyElement: from original xml.parse() operation

			    	// extract PropertyInstance object from propertyElement
			    	PropertyInstance propertyInstance = new PropertyInstance(this, propertyElement, strRequestedLanguage);
			    	
				    // store propertyInstance for everything to use that follows; the interface guarantees a unique itemId
			    	short itemId = propertyInstance.getItemId();
				    this.mapPropertyInstances.put(itemId, propertyInstance);
			    }
			    
			    // output info to C comment block, in itemId order
			    Iterator<Entry<Short, PropertyInstance>> iter = mapPropertyInstances.entrySet().iterator();
			    while (iter.hasNext()) {
			    	PropertyInstance propertyInstance = iter.next().getValue();
			        buildAcc.append(Form.indentLevel2 + propertyInstance.getCodeName() + ": "+ propertyInstance.getTypeName() + Form.newline +
			        		        Form.justifyDescription(propertyInstance.getName() + " - " + propertyInstance.getDescription()) + // trailing newline is included
                                    Form.newline);
			    }
			}

			// report Events information in comments block
			buildAcc.append(Form.indentLevel1 + "Events" + Form.newline);
			if (eventElements!=null && !eventElements.isEmpty()) {
			    for (EventElement eventElement : eventElements) {
			    	// eventElement: from original xml.parse() operation
			    	
			    	// extract EventInstance object from eventElement
			    	EventInstance eventInstance = new EventInstance(this, eventElement, strRequestedLanguage);
			    	
				    // store eventInstance for everything to use that follows; the interface guarantees a unique itemId
			    	short itemId = eventInstance.getItemId();
				    this.mapEventInstances.put(itemId, eventInstance);
			    }

			    // output info to C comment block, in itemId order
			    Iterator<Entry<Short, EventInstance>> iter = this.mapEventInstances.entrySet().iterator();
			    while (iter.hasNext()) {
			    	EventInstance eventInstance = iter.next().getValue();
			    	
			    	String strEventOutputParamTypes = new String("(");
			    	for (Integer ref : eventInstance.getXMLBasedOutputTypeReferences()) {
			    		Short shortRef = ref.shortValue(); // Passing Integer or int to this.mapTypeInstances.get() returns null;
			    		TypeInstance typeInstance = this.mapTypeInstances.get(shortRef);
			    		if (typeInstance == null)
			    			throw new Exception("Event processing finds no stored type reference");
				    	String typeName = typeInstance.getCodeName();
				    	typeName = typeName!=null ? typeName : "Type";
			    		strEventOutputParamTypes += (typeName + ", ");
			    	}
			    	strEventOutputParamTypes += ")"; // (): potential property types of returned event
			    	
			        buildAcc.append(Form.indentLevel2 + eventInstance.getCodeName() + ": " + strEventOutputParamTypes + Form.newline +
			        		        Form.justifyDescription(eventInstance.getName() + " - " + eventInstance.getDescription()) + // trailing newline is included
	                                Form.newline);
			    }
			}
			
			// report Methods information in comments block
			buildAcc.append(Form.indentLevel1 + "Methods" + Form.newline);
			if (methodElements!=null && !methodElements.isEmpty()) {
			    for (MethodElement methodElement : methodElements) {
			    	// methodElement: from original xml.parse() operation
			    	
			    	// extract MethodInstance object from methodElement
			    	MethodInstance methodInstance = new MethodInstance(this, methodElement, strRequestedLanguage);
			    	
				    // store methodInstance for everything to use that follows; the interface guarantees a unique itemId
			    	short itemId = methodInstance.getItemId();
				    this.mapMethodInstances.put(itemId, methodInstance);
			    }
			    
			    // output info to C comment block, in itemId order
			    Iterator<Entry<Short, MethodInstance>> iter = mapMethodInstances.entrySet().iterator();
			    while (iter.hasNext()) {
			    	MethodInstance methodInstance = iter.next().getValue();

			    	String strMethodInputParamTypes = new String("(");
			    	for (Integer ref : methodInstance.getXmlBasedInputTypeReferences()) {
			    		Short shortRef = ref.shortValue(); // Passing Integer or int to this.mapTypeInstances.get() returns null;
				    	String typeName = this.mapTypeInstances.get(shortRef).getCodeName();
				    	typeName = typeName!=null ? typeName : "Type";
			    		strMethodInputParamTypes += (typeName + ", ");
			    	}
			    	strMethodInputParamTypes += ")"; // (): potential property types of returned event
			    	
			    	String strMethodOutputParamTypes = new String("(");
			    	for (Integer ref : methodInstance.getXmlBasedOutputTypeReferences()) {
			    		Short shortRef = ref.shortValue(); // Passing Integer or int to this.mapTypeInstances.get() returns null
			    		TypeInstance typeInstance = this.mapTypeInstances.get(shortRef);
			    		if (typeInstance == null)
			    			throw new Exception("Method processing finds no stored type reference");
				    	String typeName = typeInstance.getCodeName();
				    	typeName = typeName!=null ? typeName : "Type";
				    	strMethodOutputParamTypes += (typeName + ", ");
			    	}
			    	strMethodOutputParamTypes += ")"; // (): potential property types of returned event
			    	
			        buildAcc.append(Form.indentLevel2 + methodInstance.getCodeName() + ": " + strMethodInputParamTypes + " -> " + strMethodOutputParamTypes + Form.newline +
			        		        Form.justifyDescription(methodInstance.getName() + " - " + methodInstance.getDescription()) + // trailing newline is included
	                                Form.newline);
			    }
			}
			
			// report Exceptions information in comments block
			buildAcc.append(Form.indentLevel1 + "Exceptions" + Form.newline);
			if (exceptionElements!=null && !exceptionElements.isEmpty()) {
			    for (ExceptionElement exceptionElement : exceptionElements) {
			    	// exceptionElement: from original xml.parse() operation
			    	
			    	// extract ExceptionInstance object from exceptionElement
			    	ExceptionInstance exceptionInstance = new ExceptionInstance(this, exceptionElement, strRequestedLanguage);
			    	
				    // store eventInstance for everything to use that follows; the interface guarantees a unique itemId
			    	short itemId = exceptionInstance.getItemId();
				    this.mapExceptionInstances.put(itemId, exceptionInstance);
			    }

			    // output info to C comment block, in itemId order
			    Iterator<Entry<Short, ExceptionInstance>> iter = mapExceptionInstances.entrySet().iterator();
			    while (iter.hasNext()) {
			    	ExceptionInstance exceptionInstance = iter.next().getValue();
			    	
			    	String strExceptionOutputParamTypes = new String("(");
			    	for (Integer ref : exceptionInstance.getXmlBasedOutputTypeReferences()) {
			    		Short shortRef = ref.shortValue(); // Passing Integer or int to this.mapTypeInstances.get() returns null;
			    		TypeInstance typeInstance = this.mapTypeInstances.get(shortRef);
			    		if (typeInstance == null)
			    			throw new Exception("Exception processing finds no stored type reference");
				    	String typeName = typeInstance.getCodeName();
				    	typeName = typeName!=null ? typeName : "Type";
			    		strExceptionOutputParamTypes += (typeName + ", ");
			    	}
			    	strExceptionOutputParamTypes += ")"; // (): potential property types of returned event
			    	
			        buildAcc.append(Form.indentLevel2 + exceptionInstance.getCodeName() + ": " + strExceptionOutputParamTypes + Form.newline + // (): potential property of returned exception
			        		        Form.justifyDescription(exceptionInstance.getName() + " - " + exceptionInstance.getDescription()) + // trailing newline is included
	                                Form.newline);
			    }
			}
			
			// close out the five reports, in their C comment block
			buildAcc.append("*/" + Form.newline +
					        Form.newline);

			// output itemID #define(s)
			String codeLabelPrefix = this.interfaceName.trim().toUpperCase().replace(" ", "") + "_";

			Iterator<Entry<Short, PropertyInstance>> iterProp = mapPropertyInstances.entrySet().iterator();
		    while (iterProp.hasNext()) {
		    	PropertyInstance propertyInstance = iterProp.next().getValue();
		    	int itemId = propertyInstance.getItemId();
				String interfaceElementName = propertyInstance.getCodeName().trim().replace(" ", "_");
				buildAcc.append("#define " + codeLabelPrefix + interfaceElementName + " " + itemId + Form.newline);
		    }

			Iterator<Entry<Short, EventInstance>> iterEvent = mapEventInstances.entrySet().iterator();
		    while (iterEvent.hasNext()) {
		    	EventInstance eventInstance = iterEvent.next().getValue();
		    	int itemId = eventInstance.getItemId();
				String interfaceElementName = eventInstance.getCodeName().trim().replace(" ", "_");
				buildAcc.append("#define " + codeLabelPrefix + interfaceElementName + " " + itemId + Form.newline);
		    }
		    
			Iterator<Entry<Short, MethodInstance>> iterMethod = mapMethodInstances.entrySet().iterator();
		    while (iterMethod.hasNext()) {
		    	MethodInstance methodInstance = iterMethod.next().getValue();
		    	int itemId = methodInstance.getItemId();
				String interfaceElementName = methodInstance.getCodeName().trim().replace(" ", "_");
				buildAcc.append("#define " + codeLabelPrefix + interfaceElementName + " " + itemId + Form.newline);
		    }
		    
			Iterator<Entry<Short, ExceptionInstance>> iterException = mapExceptionInstances.entrySet().iterator();
		    while (iterException.hasNext()) {
		    	ExceptionInstance exceptionInstance = iterException.next().getValue();
		    	int itemId = exceptionInstance.getItemId();
				String interfaceElementName = exceptionInstance.getCodeName().trim().replace(" ", "_");
				buildAcc.append("#define " + codeLabelPrefix + interfaceElementName + " " + itemId + Form.newline);
		    }

			buildAcc.append(Form.newline);
		    
			// output iid byte array code block
			DOFInterfaceID iid = ie.getIid();
			byte [] bytes = iid.getBytes();
			String iidStandardForm = Form.bytesToCommaSeparatedHex(bytes);
			buildAcc.append("uint8 " + // uint8 because we are describing a byte array
                                       codeLabelPrefix + "IDBytes[] = { " + iidStandardForm + " }" + ";" + Form.newline);
			
			// compute interface byte array code block
			String bytesString = new CodeBlock(this).computeAndFormatByteArray();
			    
			// output interface byte array code block; write all results to writer
			buildAcc.append("uint8 "+ codeLabelPrefix + "Bytes[] = {" + Form.newline + bytesString + "};" + Form.newline); // final newline is only for form
			
			writer.write(buildAcc.toString());
			success = true;
			log.debug("CLanguageTranslation.toRaw() issues complete interface information for interfaceId { " + iidStandardForm + " }");
        } catch (NullPointerException npe) {
        	log.debug("CLanguageTranslation.toRaw() sees null pointer exception");
        	throw npe;
		} catch (Exception e) {
			throw e;
		} finally {
			if (buildAcc != null) {
				if (!success) {
					buildAcc.append(Form.newline + "incomplete processing" + Form.newline);
					log.error("Incomplete processing: computed interface bytes structure is not fully populated");
				}
//				System.out.print(buildAcc.toString()); // temporarily
			}
		}
        
		return "text/plain";
	}

	/**
	 * Return mapTypeInstances
	 * @return mapTypeInstances
	 */
    public TreeMap<Short, TypeInstance> getMapTypeInstances() {
	    return this.mapTypeInstances;
    }

	/**
	 * Return mapPropertyInstances
	 * @return mapPropertyInstances
	 */
    public TreeMap<Short, PropertyInstance> getMapPropertyInstances() {
    	return this.mapPropertyInstances;
    }

	/**
	 * Return mapMethodInstances
	 * @return mapMethodInstances
	 */
    public TreeMap<Short, MethodInstance> getMapMethodInstances() {
    	return this.mapMethodInstances;
    }
    
	/**
	 * Return mapEventInstances
	 * @return mapEventInstances
	 */
    public TreeMap<Short, EventInstance> getMapEventInstances() {
    	return this.mapEventInstances;
    }
    
	/**
	 * Return mapExceptionInstances
	 * @return mapExceptionInstances
	 */
    public TreeMap<Short, ExceptionInstance> getMapExceptionInstances() {
    	return this.mapExceptionInstances;
    }
	
}