/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof;

import org.opendof.tools.repository.interfaces.core.AuthenticatedUser;
import org.opendof.tools.repository.interfaces.da.DataAccessor;
import org.opendof.tools.repository.interfaces.da.InterfaceData;

@SuppressWarnings("javadoc")
public class JiraGroupRules
{
    public JiraGroupRules()
    {
    }
    
    /**
     *  Verify Read Access.
     *  
     *  <p>Read access is based on the following groups:
     *  <ul>
     *  <li>anonymous (an Interface Repository default group)
     *  <li>user      (same as above)
     *  <li>custom    (created by submitters)
     *  </ul>
     *   
     *  If the interface is in the anonymous group it will always be added to the list
     *  unless the site configuration has the <code>CoreController.UserGroupOnlyKey</code>
     *  set to true.
     *   
     *  <p>If the interface is in the user group it will only be added to the list if 
     *  the requesting submitter is authenticated.
     *  
     *  <p>If the interface is in a submitter created custom group it will only be added 
     *  to the list if the request submitter is a member of that custom group.
     *  
     *  <p>if the interface is not one of the two constant groups or a custom group then 
     *  the interface will only be added to the list if the request submitter matches
     *  the interface's submitter.
     *
     * @param authUser the authenticated user.
     * @param interfaceData the interface data.
     * @param userGroupOnly filters on groups only.
     *
     * @return List of InterfaceData
     */
    public static boolean verifyReadAccess(AuthenticatedUser authUser, InterfaceData interfaceData, boolean userGroupOnly)
    {
        if(authUser == null)
        {
            if(userGroupOnly)   // is site configuration set for user group only
                return false;
            if(interfaceData.accessGroup == null || DataAccessor.PrivateGroup.email.equals(interfaceData.accessGroup.email))
                return false; // its private
            if(!DataAccessor.AnonymousGroup.email.equals(interfaceData.accessGroup.email))
                return false;   // the interface is not assigned to the anonymous group
            return true;
        }
        if(interfaceData.accessGroup != null)
        {
            if(DataAccessor.AnonymousGroup.email.equals(interfaceData.accessGroup.email))
                return true;   // the interface is not assigned to the anonymous group 
            if(DataAccessor.UserGroup.email.equals(interfaceData.accessGroup.email))
                return true;   // the interface is not assigned to the anonymous group
            if(authUser.email.equals(interfaceData.submitter.email))
                return true;    // it's the private group
            return false;  //TODO: must be a custom group add support here later
        }
        if(authUser.email.equals(interfaceData.submitter.email))
            return true;  // assume null accessGroup is private
        return false;
    }
}
