/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof;

import org.apache.xerces.jaxp.SAXParserFactoryImpl;
import org.apache.xerces.jaxp.SAXParserImpl;
import org.apache.xerces.util.XMLCatalogResolver;
import org.opendof.core.oal.DOFInterfaceID;
import org.opendof.tools.repository.interfaces.core.AuthenticatedUser;
import org.opendof.tools.repository.interfaces.core.CoreController;
import org.opendof.tools.repository.interfaces.core.PersistenceController;
import org.opendof.tools.repository.interfaces.core.PropertiesFile;
import org.opendof.tools.repository.interfaces.core.RawXMLTranslator;
import org.opendof.tools.repository.interfaces.core.XMLTranslator;
import org.opendof.tools.repository.interfaces.core.XSLTTranslator;
import org.opendof.tools.repository.interfaces.da.AlreadyExistsException;
import org.opendof.tools.repository.interfaces.da.DataAccessor;
import org.opendof.tools.repository.interfaces.da.DataAccessor.InterfaceIterable;
import org.opendof.tools.repository.interfaces.da.ImmutableException;
import org.opendof.tools.repository.interfaces.da.InsufficentRightsException;
import org.opendof.tools.repository.interfaces.da.InterfaceData;
import org.opendof.tools.repository.interfaces.da.InterfaceIdMatchException;
import org.opendof.tools.repository.interfaces.da.NotFoundException;
import org.opendof.tools.repository.interfaces.da.SubRepositoryNode;
import org.opendof.tools.repository.interfaces.da.SubmitterData;
import org.opendof.tools.repository.interfaces.da.SubmitterPermissions;
import org.opendof.tools.repository.interfaces.opendof.saxParser.InterfaceElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements.LanguageString;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler.OpendofErrorHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

import javax.xml.parsers.SAXParserFactory;
import java.io.ByteArrayInputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.net.URLDecoder;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;

/**
 * The Opendof Controller enforces the http:opendof.org/schema/interface-repository.xsd
 * XML namespace rules for adding, reading, modifying and deleting interfaces from the persistent store.
 * <p>
 * This class implements the <code>PersistenceController</code> interface which provides a relay for
 * requests from the <code>InterfaceServlet</code> to be routed to the proper namespace controller.
 *
 * @see org.opendof.tools.repository.interfaces.core.PersistenceController
 * @see org.opendof.tools.repository.interfaces.core.CoreController
 */
public class OpendofController implements PersistenceController {

    private static final String IidAttribute = "iid=\"";
    private static final char DOT = '.';
    private static final char LBRACE = '[';
    private static final char COLON = ':';
    private static final char LPAREN = '}';
    private static final char QUOTE = '\"';
    private static final char FSLASH = '/';
    private static final int RADIX = 16;
    private static final int BYTE = 256;
    private static final int USHORT = 65536;
    private static final char SPACE = ' ';
    private static final char LCBRACE = '{';

    private final Object syncObj = new Object();
    private final Logger log;
    private volatile DataAccessor da = null;
    private volatile String repoType = null;
    private volatile String catalogPath = null;
    private volatile boolean userGroupOnly = false;
    private final Map<String, XMLTranslator> translators;

    /**
     * OpendofController default constructor
     */
    public OpendofController() {
        log = LoggerFactory.getLogger(getClass());
        translators = new HashMap<>();
    }

    @Override
    public void destroy() {
    }

    @Override
    public void init(CoreController core, String repoType, DataAccessor da, boolean userGroupOnly) throws Exception {
        //        this.core = core;
        this.da = da;
        this.repoType = repoType;
        String catalogFolder = core.getProperties().getProperty(CoreController.CatalogFolderUrlKey, CoreController.CatalogFolderDefault);
        catalogPath = PropertiesFile.combinePaths(catalogFolder, repoType, true, false);
        synchronized (syncObj) {
            catalogPath += ".xml";
        }
        this.userGroupOnly = userGroupOnly;
        initTranslations();
        try {
            initReservedGroups();
        } catch (Exception ignored) {
        } // if already exists, great
    }

    private void initReservedGroups() throws Exception {
        SubmitterData manager = new SubmitterData(DataAccessor.AdminName, DataAccessor.AdminEmail, DataAccessor.AdminDescription, new Date());
        da.addSubmitter(manager);

        SubmitterData anonymous = new SubmitterData(DataAccessor.AnonymousGroupName, DataAccessor.AnonymousGroupEmail, DataAccessor.AnonymousGroupDescription, new Date(), true);
        da.addGroup(manager.email, anonymous);

        SubmitterData user = new SubmitterData(DataAccessor.UserGroupName, DataAccessor.UserGroupEmail, DataAccessor.UserGroupDescription, new Date(), true);
        da.addGroup(manager.email, user);

        SubmitterData priv = new SubmitterData(DataAccessor.PrivateGroupName, DataAccessor.PrivateGroupEmail, DataAccessor.PrivateGroupDescription, new Date(), true);
        da.addGroup(manager.email, priv);

        SubmitterData opendof = new SubmitterData(DataAccessor.OpenDofAllocatorGroupName, DataAccessor.OpenDofAllocatorGroupEmail, DataAccessor.OpenDofAllocatorGroupDescription, new Date(), true);
        da.addGroup(manager.email, opendof);

        SubmitterData cliAdmin = new SubmitterData(DataAccessor.CliAdminName, DataAccessor.CliAdminEmail, DataAccessor.CliAdminDescription, new Date());
        da.addSubmitter(cliAdmin);

    }

    private void initTranslations() {
        URL htmlXslURL = this.getClass().getClassLoader().getResource("translation/opendof/html-translation.xsl");
        URL javaXslURL = this.getClass().getClassLoader().getResource("translation/opendof/java-translation.xsl");

        if(htmlXslURL == null || javaXslURL == null){
            throw new IllegalStateException("Missing init files: translation/opendof/html-translation.xsl and/or translation/opendof/java-translation.xsl");
        }

        translators.put("XML", new RawXMLTranslator());
        String htmlXsl = htmlXslURL.toString();
        translators.put("HTML", new XSLTTranslator(htmlXsl, "HTML"));
        String javaXsl = javaXslURL.toString();
        translators.put("Java", new XSLTTranslator(javaXsl, "Java"));
        translators.put("C", new CLanguageTranslator(catalogPath));
    }

    @Override
    public boolean isSupportingRepository(String iid) {
        return iid.contains("[") || iid.contains(".");
    }

    @Override
    public String getIidFromXml(String xml) {
        try {
            return getDofInterfaceIdFromXml(xml).toStandardString();
        } catch (Exception ignored) {
            return null;
        }
    }

    /**
     * Looks for <code>DofUriBase</code> anywhere in the given uri and replaces
     * it in place with <code>DOFInterfaceID.toStandardForm</code> if found.
     *
     * @param uri to be converted
     * @return input uri if exact matching is not found, otherwise input uri with in place replacement of standard form.
     */
    public static String dofUriToStandardForm(String uri) {
        try {
            //Decode incase any URL Encoded characters have been used.
            uri = URLDecoder.decode(uri, "utf-8");
        } catch (UnsupportedEncodingException ignored) {
            return uri;
        }

        try {
            //Already a valid standard string
            return DOFInterfaceID.create(uri).toStandardString();
        } catch (RuntimeException ignored) {
            //Expected
        }

        // strict checking if actually a dof.<registry>.<hex> uri
        // early out with least amount of instructions if not dofUriForm
        try {
            int dot1 = 0; // 1 past first dot, first char of registry number
            int dot2 = uri.indexOf(DOT, dot1); // delimit registry number
            if (dot2 == -1 || dot1 == dot2)
                return uri;
            int dof0 = 0;
            StringBuilder sb = new StringBuilder(uri.substring(0, dof0)).append(LBRACE);
            int length = 0;
            // make sure registry only has 1 or 2 base 10 digits 
            while (dot1 < dot2) {
                char c = uri.charAt(dot1);
                dot1++;
                sb.append(c);
                if (!Character.isDigit(c))
                    return uri;
                ++length;
                if (length > 2)
                    return uri;
            }
            sb.append(":{");
            // dot1 and dot2 at second dot i.e. 'dof.dd.'
            // make sure ID only has  1 to 8 base 16 digits
            ++dot1; // move to past 2nd dot to expected first hex char
            length = 0;
            int uriLength = uri.length();
            while (dot1 < uriLength) {
                char c = Character.toUpperCase(uri.charAt(dot1++));
                dot2 = Character.digit(c, RADIX); // capture hex value or -1;
                if (dot2 == -1) {
                    if (length == 0)
                        return uri; // not hex at the 'H' - dof.<registry>.H
                    --dot1; // step back to the first character of the rest of the original uri
                    break; // end of hex chars in string  - dof.<registry>.HH&cmd=other
                }
                sb.append(c);
                if (++length > 8) {
                    --dot1; // step back to the first character of the rest of the original uri
                    break;
                }
            }
            if (length == 0)
                return uri; // only found dof.<registry>.
            // finish up the standard form and append any remaining of the original uri string. 
            sb.append("}]");
            // dot1 is at the first character past ID hex portion of string or past the end
            if (dot1 == uriLength)
                return DOFInterfaceID.create(sb.toString()).toStandardString();
            return DOFInterfaceID.create(sb.toString()).toStandardString() + uri.substring(dot1);
        } catch (Exception ignored) {
            return uri;
        }
    }

    /**
     * @param standardForm dof standard form in.
     * @return dof uri formated string
     */
    private static String StandardFormToDofUri(String standardForm) {
        // [1:{01}]
        try {
            int idx0 = 1;
            int idx1 = standardForm.indexOf(COLON);
            StringBuilder sb = new StringBuilder().append(standardForm.substring(idx0, idx1)).append(DOT);
            idx0 = idx1 + 2; // step past ":{";
            idx1 = standardForm.indexOf(LPAREN);
            sb.append(standardForm.substring(idx0, idx1));
            return sb.toString();
        } catch (Exception ignored) {
            return standardForm;
        }
    }

    private static DOFInterfaceID getDofInterfaceIdFromXml(String xml) throws Exception {
        int sidx = xml.indexOf(IidAttribute);
        if (sidx == -1)
            throw new Exception("Did not find " + IidAttribute + " in the file");
        sidx += IidAttribute.length();
        int eidx = xml.indexOf(QUOTE, sidx);
        String iidStr = xml.substring(sidx, eidx);
        return DOFInterfaceID.create(iidStr);
    }

    // FIXME: the following synchronized is required because of the static stack variable found in the top of the OpenDofSaxHandler
    // at some point we should remove the current requirements of that static variable.  At that time, remove this synchronized. 
    private synchronized InterfaceElement validateIncoming(String xml, boolean publish) throws Exception {
        // setup oasis catalog
        XMLCatalogResolver resolver = new XMLCatalogResolver();
        resolver.setPreferPublic(true);
        resolver.setCatalogList(new String[]{catalogPath});

        System.setProperty("javax.xml.parsers.SAXParserFactory", "org.apache.xerces.jaxp.SAXParserFactoryImpl");
        SAXParserFactoryImpl spf = (SAXParserFactoryImpl) SAXParserFactory.newInstance();
        spf.setNamespaceAware(true);
        spf.setValidating(false); // can not seem to get cataloging to work
        SAXParserImpl saxParser = (SAXParserImpl) spf.newSAXParser();
        XMLReader xmlReader = saxParser.getXMLReader();
        xmlReader.setProperty("http://apache.org/xml/properties/internal/entity-resolver", resolver);
        xmlReader.setErrorHandler(new OpendofErrorHandler());
        OpenDofSaxHandler handler = new OpenDofSaxHandler();
        OpenDofSaxHandler.setPublish(publish);
        xmlReader.setContentHandler(handler);

        byte[] bytes = xml.getBytes();
        ByteArrayInputStream bais = new ByteArrayInputStream(bytes);

        Reader reader = new InputStreamReader(bais, "UTF-8");
        InputSource inSource = new InputSource(reader);
        xmlReader.parse(inSource);
        return handler.getInterface();
    }

    @Override
    public List<InterfaceData> listInterfaces(AuthenticatedUser user, String submitter, String accessGroup, Boolean published) throws Exception {
        InterfaceIterable ii = da.listInterfaces(repoType, submitter, accessGroup, published);
        Iterator<InterfaceData> iter = ii.iterator();
        List<InterfaceData> list = new ArrayList<>();
        while (iter.hasNext()) {
            InterfaceData data = iter.next();
            if (JiraGroupRules.verifyReadAccess(user, data, userGroupOnly)) {
                data.setName(getInterfaceNameFromXml(data.xml, data.iid));
                list.add(data);
            }
        }
        ii.close();
        return list;
    }

    @Override
    public List<InterfaceData> listSharedInterfaces(AuthenticatedUser user, String submitter, Boolean published) throws Exception {
        InterfaceIterable ii = da.listSharedInterfaces(repoType, user.email, published);
        Iterator<InterfaceData> iter = ii.iterator();
        List<InterfaceData> list = new ArrayList<>();
        while (iter.hasNext()) {
            InterfaceData data = iter.next();
            data.setName(getInterfaceNameFromXml(data.xml, data.iid));
            list.add(data);
        }
        ii.close();
        return list;
    }

    private String getInterfaceNameFromXml(String xml, String iid) {
        if (xml == null)
            return iid;
        InterfaceElement parsed;
        try {
            parsed = validateIncoming(xml, false);
        } catch (Exception e) {
            log.debug("Failed to validate existing interface " + iid, e);
            return iid;
        }
        MetadataElements metadata = parsed.getMetadata();
        if (metadata == null) {
            return iid;
        }
        List<LanguageString> names = metadata.getDisplayNames();
        if (names.isEmpty())
            return iid;
        for (LanguageString name : names) {
            if (name.getAttrValue().equals("en"))
                return name.getCdata();
        }
        return names.get(0).getCdata();
    }

    @Override
    public InterfaceData getInterface(AuthenticatedUser user, String iid, String version) throws Exception {
        try {
            iid = dofUriToStandardForm(iid);
            InterfaceData interfaceData = da.getInterface(iid, version == null ? "1" : version);
            Integer permissionLevel = null;
            try{
                permissionLevel = da.getPermission(repoType, iid, version == null ? "1" : version, user.email);
            } catch(Exception ignored){
                log.debug("Failed to get permission for user: {} on iid: {}", user.email, iid );
            }

            if (permissionLevel != null || JiraGroupRules.verifyReadAccess(user, interfaceData, userGroupOnly)) {
                interfaceData.setName(getInterfaceNameFromXml(interfaceData.xml, iid));
                return interfaceData;
            }
            throw new NotFoundException("iid: " + iid + " not found");
        } catch (Exception e) {
            log.warn(getClass().getSimpleName() + ".getInterface failed iid: " + iid + " version: " + (version == null ? "null" : version), e);
            throw e;
        }
    }

    @Override
    public void deleteInterface(AuthenticatedUser user, String iid, String version) throws Exception {
        if (iid == null)
            throw new Exception("An Interface ID must be given");
        if (version == null)
            throw new Exception("An Interface version must be given");
        DOFInterfaceID diid = DOFInterfaceID.create(iid);
        iid = diid.toStandardString();
        String registry = String.valueOf(diid.getRegistry());
        String byteSize;
        long id = diid.getIdentifier();
        if (id < BYTE)
            byteSize = "1";
        else if (id < USHORT)
            byteSize = "2";
        else
            byteSize = "4";

        if (user == null)
            throw new InsufficentRightsException("insufficent rights to delete the interface: " + iid);

        SubRepositoryNode node = da.getSubRepoNode(CoreController.OpenDofRepo, registry + FSLASH + byteSize);
        InterfaceData idata = da.getInterface(iid, version);
        if (user.email.equals(CoreController.CliUser.email)) {
            if (idata.publish && user.rights == null)
                throw new InsufficentRightsException("cli admin must use force for this operation");
            da.deleteInterface(iid, version, node);
            log.info(iid + SPACE + version + " deleted");
            return;
        }
        if (idata.publish)
            throw new InsufficentRightsException("insufficent rights to delete the published interface: " + iid);

        if (!user.email.equals(idata.submitter.email))
            throw new InsufficentRightsException("You must be the owner of the working interface: " + iid + " to delete it");

        da.deleteInterface(iid, version, node);
        log.info(iid + SPACE + version + " deleted");
    }

    @Override
    public void publishInterface(AuthenticatedUser user, String iid, String version) throws Exception {
        if (iid == null)
            throw new Exception("An Interface ID must be given");
        if (version == null)
            throw new Exception("An Interface version must be given");
        iid = DOFInterfaceID.create(iid).toString();
        InterfaceData idata = da.getInterface(iid, version);
        if (idata.publish)
            throw new Exception(iid + " is already Published");

        if (!user.email.equals(idata.submitter.email))
            throw new InsufficentRightsException("You must be the owner of the working interface: " + iid + " to publish it");

        //@formatter:off
        da.updateInterface(
                new InterfaceData(
                        null, idata.iid, idata.xml, version,
                        idata.submitter, idata.accessGroup,
                        idata.getRepoType(),
                        idata.creationDate, new Date(), true));
        //@formatter:on
        log.info(iid + SPACE + version + " moved to published");
    }

    @Override
    public void addInterfaceSharePermission(AuthenticatedUser user, String iid, String version, String email,
                                            SubmitterPermissions.PermissionLevel permissionLevel) {

    }

    @SuppressWarnings("null")
    @Override
    public void updateInterface(AuthenticatedUser user, String iid, String xml, String version, SubmitterData submitter, SubmitterData accessGroup, Boolean published) throws Exception {
        if (version == null)
            throw new Exception("An Interface version must be given");
        // The iid and xml arguments have the following rules:
        // if iid is null, look to xml for iid.
        // if iid is null, and xml is null throw InterfaceIdMatchException.
        // if iid is given and xml is null, use iid.
        // if iid is given and xml is given the two must match or throw InterfaceIdMatchException.

        DOFInterfaceID xiid = null;
        if (xml != null)
            xiid = getDofInterfaceIdFromXml(xml);
        DOFInterfaceID giid = null;
        if (iid != null)
            giid = DOFInterfaceID.create(iid);
        if (xml == null)
            xiid = giid;
        if (giid == null && xiid == null)
            throw new InterfaceIdMatchException("iid and xml are both null");
        else if (giid == null)
            giid = xiid;
        else if (!Objects.equals(giid,xiid))
            throw new InterfaceIdMatchException("iid: " + giid.toStandardString() + " xml: " + xiid.toStandardString());
        if (user == null)
            throw new InsufficentRightsException("Insufficient rights to update interface: " + xiid);

        InterfaceData idata = da.getInterface(giid.toStandardString(), version);

        //Check if the user is not the admin or the submitter
        if (!Objects.equals(user.email,DataAccessor.CliAdminEmail) && !Objects.equals(user.email,idata.submitter.email)) {
            //Check if the user does not have write access
            Integer userPermission = da.getPermission(repoType, iid, version, user.email);
            if(userPermission != SubmitterPermissions.PermissionLevel.WRITE.getValue()) {
                throw new InsufficentRightsException("Insufficient rights to update interface: " + xiid);
            }

        }


        InterfaceElement elementsIn = null;
        if (xml != null) {
            elementsIn = validateIncoming(xml, idata.publish);
            elementsIn.validate();
        }
        if (idata.xml != null) {
            if (idata.publish) {
                InterfaceElement elementsExisting = validateIncoming(idata.xml, true);
                if (elementsIn != null && !elementsExisting.validPublish(elementsIn)) {
                    if (!user.email.equals(DataAccessor.CliAdminEmail) && user.rights == null)
                        throw new ImmutableException("attempting to modify an immutable element/attribute of a published interface");
                }
            }
        }

        String pchanged = "unchanged";

        if (published == null)
            published = idata.publish;
        else {
            if (!user.email.equals(DataAccessor.CliAdminEmail) && user.rights == null)
                throw new InsufficentRightsException("Insufficent rights to update the published field");
            pchanged = String.valueOf(published);
        }

        SubmitterData submitterData = idata.submitter;
        String schanged = "unchanged";
        if (submitter != null) {
            submitterData = submitter;
            schanged = submitter.email;
        }

        SubmitterData accessGroupData = idata.accessGroup;
        String gchanged = "unchanged";
        if (accessGroup != null) {
            accessGroupData = accessGroup;
            gchanged = accessGroup.name;
        }

        String xchanged = "unchanged";
        if (xml == null)
            xml = idata.xml;
        else
            xchanged = "updated";

        InterfaceData updatedInterfaceData = new InterfaceData(null, giid.toStandardString(), xml, version, submitterData, accessGroupData, idata.getRepoType(), idata.creationDate, new Date(), published);
        da.updateInterface(updatedInterfaceData);
        log.info("updated the " + (idata.publish ? "Published" : "Working") + " interface " + giid.toStandardString() + " xml: " + xchanged + " submitter: " + schanged + " group: " + gchanged + " published: " + pchanged);
    }

    @Override
    public InterfaceData addInterface(AuthenticatedUser user, InterfaceData iface, String subRepoRdn) throws Exception {
        if (iface.version == null)
            throw new Exception("An Interface version must be given");

        DOFInterfaceID xiid = null;
        if (iface.xml != null)
            xiid = getDofInterfaceIdFromXml(iface.xml);
        DOFInterfaceID giid = null;
        if (iface.iid != null)
            giid = DOFInterfaceID.create(iface.iid);
        if (giid != null && xiid != null) {
            if (!giid.equals(xiid))
                throw new InterfaceIdMatchException("iid: " + giid.toStandardString() + " and xml: " + xiid.toStandardString() + " do not match");
        }
        if (giid == null)
            giid = xiid;

        if (giid != null) {
            try {
                da.getInterface(giid.toStandardString(), iface.version);
                throw new AlreadyExistsException(giid.toStandardString() + SPACE + iface.version);
            } catch (NotFoundException ignored) { // expected case, do nothing
            }
        } else {
            iface = allocateIid(user, iface, subRepoRdn);
            log.info(iface.iid + SPACE + iface.version + " added to working data store via normal allocation");
            return iface;
        }
        if (!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights to add a given iid.");
        iface = allocateIid(user, iface, subRepoRdn);
        log.info(giid.toStandardString() + SPACE + iface.version + " added to data store via admin cli");
        return iface;
    }

    private InterfaceData allocateIid(AuthenticatedUser user, InterfaceData iface, String rdn) throws Exception {
        if (rdn == null || rdn.length() == 0)
            rdn = "1/4";
        SubRepositoryNode node = da.getSubRepoNode(iface.getRepoType(), rdn);
        if (node.getDepth() == 1) {
            rdn += rdn.endsWith("/") ? "4" : "/4";
        }
        if (iface.creationDate == null || iface.lastModifiedDate == null) {
            Instant date = Instant.now();
            //@formatter:off
            iface = new InterfaceData(
                    iface.getName(), iface.iid, iface.xml, iface.version, iface.submitter, iface.accessGroup,
                    iface.getRepoType(), iface.creationDate == null ? new Date(date.toEpochMilli()) : iface.creationDate,
                    iface.lastModifiedDate == null ? new Date(date.toEpochMilli()) : iface.lastModifiedDate, iface.publish);
            //@formatter:off
        }

        if (DataAccessor.CliAdminEmail.equals(user.email))
            return da.allocateInterfaceId(iface, rdn);

        if (node.getName().equals("4")) {
            if (DataAccessor.UserGroupName.equals(node.getGroup().name) && iface.submitter.name != null)
                return da.allocateInterfaceId(iface, rdn);
            throw new InsufficentRightsException("insufficent rights for repoType: " + iface.getRepoType() + " rdn: " + rdn);
        }
        if (DataAccessor.OpenDofAllocatorGroupName.equals(String.valueOf(iface.submitter.group)))
            return da.allocateInterfaceId(iface, rdn);
        throw new InsufficentRightsException("insufficent rights for repoType: " + iface.getRepoType() + " rdn: " + rdn);
    }

    private void addGivenIid(AuthenticatedUser user, InterfaceData iface, String rdn) throws Exception {
        iface = allocateIid(user, iface, rdn);
        log.info(iface.iid + SPACE + iface.version + " added to working data store via a given iid");
    }

    @Override
    public SubRepositoryNode getSubRepositoryNode(String repoType, String rdn) throws Exception {
        return da.getSubRepoNode(repoType, rdn);
    }

    @Override
    public void addSubRepositoryNode(AuthenticatedUser user, String parentRdn, SubRepositoryNode child) throws Exception {
        if (!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for adding sub-repository nodes");
        if (child.getGroup() != null) {
            SubmitterData group = da.getSubmitter(child.getGroup().name);
            child = new SubRepositoryNode(child.getRepoType(), child.getName(), child.getLabel(), child.getDepth(), group);
        }
        da.addSubRepoNode(parentRdn, child);
    }

    @Override
    public void deleteSubRepositoryNode(AuthenticatedUser user, String parentRdn, SubRepositoryNode node) throws Exception {
        if (!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for deleting sub-repository nodes");
        da.deleteSubRepoNode(parentRdn, node);
    }

    @Override
    public void updateSubRepositoryNode(AuthenticatedUser user, String rdn, SubRepositoryNode node) throws Exception {
        if (!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights for updating sub-repository nodes");
        da.updateSubRepoNode(CoreController.OpenDofRepo, rdn, node);
    }

    @Override
    public String getURLIdentifier(String standardIterfaceID) {
        return StandardFormToDofUri(standardIterfaceID);
    }

    @Override
    public String getStandardIdentifier(String urlInterfaceID) {
        return dofUriToStandardForm(urlInterfaceID);
    }

    @Override
    public boolean versionSupported() {
        return false;
    }

    @Override
    public synchronized List<String> getTranslations() {
        return new ArrayList<>(translators.keySet());
    }

    @Override
    public XMLTranslator getTranslator(String translation) throws Exception {
        if (!translators.containsKey(translation)) {
            throw new Exception("Opendof repository does not support translation " + translation);
        }
        return translators.get(translation);
    }

    @Override
    public boolean hasAllocateSupport() {
        return true;
    }

    /**
     * @param user with rights for the synch call.
     * @throws Exception if the synch could not be performed
     */
    //TODO: one shot code delete me next version
    // this should only be ran one time after v1.0 database has been imported into v2.0 to fix up holes table to pre-existing interfaces in the database. 
    public void synch(AuthenticatedUser user) throws Exception {
        if (!DataAccessor.CliAdminEmail.equals(user.email))
            throw new InsufficentRightsException("insufficent rights to preform interface table synch");

        // any existing interfaces are already public, thus set there accessGroup to anon
        InterfaceIterable ii = da.listInterfaces(repoType, null, null, null);
        for (InterfaceData data : ii) {
            //@formatter:off
            InterfaceData updateData = new InterfaceData(
                    null, data.iid, data.xml, data.version,
                    data.submitter, DataAccessor.AnonymousGroup, data.getRepoType(),
                    data.creationDate, data.lastModifiedDate, data.publish);
            //@formatter:on
            da.updateInterface(updateData);
        }
        ii.close();

        List<InterfaceData> list = listInterfaces(user, null, null, null);
        for (InterfaceData iface : list) {
            DOFInterfaceID iid = DOFInterfaceID.create(iface.iid);
            String iids = iid.toStandardString();
            int idx = iids.indexOf(LCBRACE);
            iids = iids.substring(++idx);
            iids = iids.substring(0, iids.length() - 2);
            int numberOfBytes;
            switch (iids.length()) {
                case 2:
                    numberOfBytes = 1;
                    break;
                case 4:
                    numberOfBytes = 2;
                    break;
                case 8:
                    numberOfBytes = 4;
                    break;
                default:
                    throw new Exception("unknown iid number of bytes: " + iid.toStandardString());
            }

            int registry = iid.getRegistry();
            if (registry != 1) {
                log.warn("\n\nWARNING: You have an Interface with an unsupported registry: " + iid.toStandardString() + "\n You need to manually cleanup this interface with the cli and direct SQL\n");
                continue;
            }
            String rdn = registry + "/" + numberOfBytes;
            addGivenIid(user, iface, rdn);
        }
    }
}
