package org.opendof.tools.repository.interfaces.opendof.iface.compute;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.TreeMap;

import org.opendof.tools.repository.interfaces.opendof.CLanguageTranslator;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection.FieldData;

/**
 * Compute the C language structure of bytes that represents an actual DOF interface.
 */
public class CodeBlock {
	private CLanguageTranslator clt;
	private TreeMap<Short, TypeInstance> mapTypeInstances = null;
	private TreeMap<Short, PropertyInstance> mapPropertyInstances = null;
	private TreeMap<Short, EventInstance> mapEventInstances = null;
	private TreeMap<Short, MethodInstance> mapMethodInstances = null;
	private TreeMap<Short, ExceptionInstance> mapExceptionInstances = null;
	
	public CodeBlock(CLanguageTranslator clt) {
		this.clt = clt;
	}

	/**
	 * Compute and format the type definition portion of the structure of bytes that represents an actual DOF interface.
	 * @return The string representation of bytes.
	 * @throws Exception
	 */
	private String computeAndFormatTypeDefinitionBytes() throws Exception {
		// types count, expressed as a compressed 16 bit number in comma-separated hex bytes
		String retString = new String(Form.indentLevel2 + Form.binaryC16Bit(this.mapTypeInstances.size()) + Form.newline);

		// list individual types, one display line per types
	    Iterator<Entry<Short, TypeInstance>> iter = this.mapTypeInstances.entrySet().iterator();
	    while (iter.hasNext()) {
	    	TypeInstance typeInstance = iter.next().getValue();
	    	// typeId, expressed as a compUnsign16 bit number in comma-separated byte(s)
        	short typeId = typeInstance.getTypeId();
	    	retString += (Form.indentLevel2 + Form.binaryC16Bit(typeId)); // trailing comma, no newline
	        if (!typeInstance.isSimpleType()) {
	        	boolean trailingSpace = true;
	        	switch (typeId) {
	        	case TypeInstance.STRING_TYPEID: // 0x31 stringd typeId, comprUnsign16 pdu len, comprUnsign24 max char len, comprUnsign16 MIB enum number
	        		int maxCharLength = typeInstance.getLength();
	        		int enumMIB = typeInstance.getEncoding();
	        		int pduLength = Form.binaryC24BitLength(maxCharLength) + Form.binaryC16BitLength(enumMIB);
	        		retString += (" " + Form.binaryC16Bit(pduLength) + " " + Form.binaryC24Bit(maxCharLength) + " " + Form.binaryC16Bit(enumMIB)); // trailing comma, no newline
	        		// Note: the xml can supply a minimum required char length, but there is no place to place that value in these bytes
	        		break;
	        	case TypeInstance.BLOB_TYPEID: // 0x49 blob typeId, comprUnsign16 pdu len, comprUnsign24 max blob len, comprUnsign24 min blob len
	        		int maxBlobLength = typeInstance.getLength();
	        		int minBlobLength = typeInstance.getMinLength();
	        		boolean lengthsEqual = maxBlobLength==minBlobLength;
	        		pduLength = Form.binaryC24BitLength(maxBlobLength) + (lengthsEqual ? 0:Form.binaryC24BitLength(minBlobLength));
	        		retString += (" " + Form.binaryC16Bit(pduLength) + " " + Form.binaryC24Bit(maxBlobLength)); // trailing comma, no newline
	        		if (!lengthsEqual)
	        			retString += (" " + Form.binaryC24Bit(minBlobLength));
	        		break;
	        	case TypeInstance.ARRAY_TYPEID: // 0x4d array typeId, comprUnsign16 pdu len, comprUnsign16 ref to element subtype, comprUnsign16 max element count, comprUnsign16 min element count
	        		short refElementReferredTo = typeInstance.getRefersTo();
	        		short codeBlockTypeReference = this.clt.getCodeBlockTypeReference(refElementReferredTo);
	        		int maxElementCount = typeInstance.getLength();
	        		int minElementCount = typeInstance.getMinLength();
	        		boolean elementCountsEqual = maxElementCount==minElementCount;
	        		pduLength = Form.binaryC16BitLength(codeBlockTypeReference) + Form.binaryC16BitLength(maxElementCount) + (elementCountsEqual ? 0:Form.binaryC16BitLength(minElementCount));
	        		retString += (" " + Form.binaryC16Bit(pduLength) + " " + Form.binaryC16Bit(codeBlockTypeReference) + " " + Form.binaryC16Bit(maxElementCount)); // trailing comma, no newline
	        		if (!elementCountsEqual)
	        			retString += (" " + Form.binaryC16Bit(minElementCount)); // trailing comma, no newline
	        		break;
	        	case TypeInstance.REFERENCE_TYPEID:
	        		refElementReferredTo = typeInstance.getRefersTo();
	        		codeBlockTypeReference = this.clt.getCodeBlockTypeReference(refElementReferredTo);
	        		pduLength = Form.binaryC16BitLength(codeBlockTypeReference);
	        		retString += (" " + Form.binaryC16Bit(pduLength) + " " + Form.binaryC16Bit(codeBlockTypeReference)); // trailing comma, no newline
	               	// note: No example available, but we known that, at output time, reference type gets handled the same as nullable type; so we do that.
	        		break;
	        	case TypeInstance.NULLABLE_TYPEID: // 0x53 nullable typeId, comprUnsign16 pdu len, comprUnsign16 ref to subtype 
	        		refElementReferredTo = typeInstance.getRefersTo();
	        		codeBlockTypeReference = this.clt.getCodeBlockTypeReference(refElementReferredTo);
	        		pduLength = Form.binaryC16BitLength(codeBlockTypeReference);
	        		retString += (" " + Form.binaryC16Bit(pduLength) + " " + Form.binaryC16Bit(codeBlockTypeReference)); // trailing comma, no newline
	        		break;
	        	case TypeInstance.STRUCTURE_TYPEID:
	        		List<FieldData> fields = typeInstance.getFields();
	        		
	        		pduLength = Form.binaryC16BitLength(fields.size());
	        		for (FieldData field : fields) {
	        			refElementReferredTo = (short)field.getReference();
		        		codeBlockTypeReference = this.clt.getCodeBlockTypeReference(refElementReferredTo);
	        			pduLength += Form.binaryC16BitLength(codeBlockTypeReference);
	        		}
	        		retString += (" " + Form.binaryC16Bit(pduLength) + " " + Form.binaryC16Bit(fields.size())); // trailing comma, no newline
	        		
	        		for (FieldData field : fields) {
	        			refElementReferredTo = (short)field.getReference();
		        		codeBlockTypeReference = this.clt.getCodeBlockTypeReference(refElementReferredTo);
		        		retString += (" " + Form.binaryC16Bit(codeBlockTypeReference));
	        		}
	        		break;
	        	default:
	        		trailingSpace = false;
	        		break;
	        	}
	        	if (trailingSpace)
	        		retString += (" "); // trailing space, no newline
	        }
//	        retString += Form.indentLevel3 + Form.indentLevel3 + Form.indentLevel3 + Form.indentLevel3 + Form.indentLevel3 + clt.getCodeBlockTypeReference(typeInstance.getTypeReference()); // temporarily
        	retString += (Form.newline);
	    }
	    return retString;
    }

	/**
	 * Compute and format the properties portion of the structure of bytes that represents an actual DOF interface.
	 * @return The string representation of bytes.
	 * @throws Exception
	 */
	private String computeAndFormatPropertiesBytes() throws Exception {
		// properties count, expressed as a compressed 16 bit number in comma-separated hex bytes
		String retString = new String(Form.indentLevel2 + Form.binaryC16Bit(this.mapPropertyInstances.size()) + Form.newline);

		// list individual method itemIds, one display line per itemId
	    Iterator<Entry<Short, PropertyInstance>> iter = this.mapPropertyInstances.entrySet().iterator();
	    while (iter.hasNext()) {
	    	// itemId, byte W|R, comprUnsign16 type ref
	    	PropertyInstance propertyInstance = iter.next().getValue();
        	short itemId = propertyInstance.getItemId();
	    	retString += (Form.indentLevel2 + Form.binaryC16Bit(itemId)); // trailing comma, no newline
	    	
	    	byte [] bits = new byte[] { (byte)(propertyInstance.getReadable()==true ? 1:0) };
	    	bits[0] |= (byte)(propertyInstance.getWriteable()==true ? 2:0);

	    	short refReferredTo = propertyInstance.getCodeBlockTypeReference();
	    	retString += (" " + Form.bytesToCommaSeparatedHex(bits) + " " + Form.binaryC16Bit(refReferredTo) + Form.newline); // trailing comma and newline
	    }
		return retString;
	}

	/**
	 * Compute and format the events portion of the structure of bytes that represents an actual DOF interface.
	 * @return The string representation of bytes.
	 * @throws Exception
	 */
	private String computeAndFormatEventsBytes() throws Exception {
		// events count, expressed as a compressed 16 bit number in comma-separated hex bytes
		String retString = new String(Form.indentLevel2 + Form.binaryC16Bit(this.mapEventInstances.size()) + Form.newline);

		// list individual event itemIds, one display line per itemId
	    Iterator<Entry<Short, EventInstance>> iter = this.mapEventInstances.entrySet().iterator();
	    while (iter.hasNext()) {
	    	// itemId, comprUnsign16 outputTypeRefs count, list of comprUnsign16 references to output type
	    	EventInstance eventInstance = iter.next().getValue();
        	short itemId = eventInstance.getItemId();
	    	retString += (Form.indentLevel2 + Form.binaryC16Bit(itemId) + Form.newline); // trailing comma and newline

	    	short [] outputTypeRefs = eventInstance.getCodeBlockOutputTypeReferences();
	    	int outputTypeRefsCount = outputTypeRefs.length;
    		retString += (" " + Form.binaryC16Bit(outputTypeRefsCount)); // trailing comma, no newline
    		for (short outputTypeRef : outputTypeRefs)
        		retString += (" " + Form.binaryC16Bit(outputTypeRef)); // trailing comma
	    	
    		retString += Form.newline;
	    }
		return retString;
	}

	/**
	 * Compute and format the methods portion of the structure of bytes that represents an actual DOF interface.
	 * @return The string representation of bytes.
	 * @throws Exception
	 */
	private String computeAndFormatMethodsBytes() throws Exception {
		// methods count, expressed as a compressed 16 bit number in comma-separated hex bytes
		String retString = new String(Form.indentLevel2 + Form.binaryC16Bit(this.mapMethodInstances.size()) + Form.newline);

		// list individual method itemIds, one display line per itemId
	    Iterator<Entry<Short, MethodInstance>> iter = this.mapMethodInstances.entrySet().iterator();
	    while (iter.hasNext()) {
	    	// itemId, comprUnsign16 inputTypeRefs count, list of comprUnsign16 references to input type, comprUnsign16 outputTypeRefs count, list of comprUnsign16 references to output type
	    	MethodInstance methodInstance = iter.next().getValue();
        	short itemId = methodInstance.getItemId();
        	retString += (Form.indentLevel2 + Form.binaryC16Bit(itemId)); // trailing comma, no newline
        	
	    	short [] inputTypeRefs = methodInstance.getCodeBlockInputTypeReferences();
	    	int inputTypeRefsCount = inputTypeRefs.length;
    		retString += (" " + Form.binaryC16Bit(inputTypeRefsCount)); // trailing comma, no newline
    		for (short inputTypeRef : inputTypeRefs)
        		retString += (" " + Form.binaryC16Bit(inputTypeRef)); // trailing comma
        	
	    	short [] outputTypeRefs = methodInstance.getCodeBlockOutputTypeReferences();
	    	int outputTypeRefsCount = outputTypeRefs.length;
    		retString += (" " + Form.binaryC16Bit(outputTypeRefsCount)); // trailing comma, no newline
    		for (short outputTypeRef : outputTypeRefs)
        		retString += (" " + Form.binaryC16Bit(outputTypeRef)); // trailing comma
        	
    		retString += Form.newline;
	    }
		return retString;
	}

	/**
	 * Compute and format the exceptions portion of the structure of bytes that represents an actual DOF interface.
	 * @return The string representation of bytes.
	 * @throws Exception
	 */
	private String computeAndFormatExceptionsBytes() throws Exception {
		// exceptions count, expressed as a compressed 16 bit number in comma-separated hex bytes
		String retString = new String(Form.indentLevel2 + Form.binaryC16Bit(this.mapExceptionInstances.size()) + Form.newline);

		// list individual exception itemIds, one display line per itemId
	    Iterator<Entry<Short, ExceptionInstance>> iter = this.mapExceptionInstances.entrySet().iterator();
	    while (iter.hasNext()) {
	    	// itemId, comprUnsign16 outputTypeRefs count, list of comprUnsign16 references to output type
	    	ExceptionInstance exceptionInstance = iter.next().getValue();
        	short itemId = exceptionInstance.getItemId();
	    	retString += (Form.indentLevel2 + Form.binaryC16Bit(itemId)); // trailing comma, no newline
    		
	    	short [] outputTypeRefs = exceptionInstance.getCodeBlockOutputTypeReferences();
	    	int outputTypeRefsCount = outputTypeRefs.length;
    		retString += (" " + Form.binaryC16Bit(outputTypeRefsCount)); // trailing comma, no newline
    		for (short outputTypeRef : outputTypeRefs)
        		retString += (" " + Form.binaryC16Bit(outputTypeRef)); // trailing comma
    		
    		retString += Form.newline;
	    }
	    if (iter.hasNext())
	    	retString += (Form.newline);
		return retString;
	}

	/**
	 * Compute and format full structure of bytes that represents an actual DOF interface.
	 * @return The string representation of the structure.
	 * @throws Exception thrown if the byte array could not be calculated.
	 */
	public String computeAndFormatByteArray() throws Exception {
		String retString = new String("");
		this.mapTypeInstances = this.clt.getMapTypeInstances();
		retString += this.computeAndFormatTypeDefinitionBytes();
		this.mapPropertyInstances = this.clt.getMapPropertyInstances();
		retString += this.computeAndFormatPropertiesBytes();
		this.mapEventInstances = this.clt.getMapEventInstances();
		retString += this.computeAndFormatEventsBytes();
		this.mapMethodInstances = this.clt.getMapMethodInstances();
		retString += this.computeAndFormatMethodsBytes();
		this.mapExceptionInstances = this.clt.getMapExceptionInstances();
		retString += this.computeAndFormatExceptionsBytes();
		return retString;
	}

}