package org.opendof.tools.repository.interfaces.opendof.iface.compute;

import java.util.ArrayList;
import java.util.List;

import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements.LanguageString;

/**
 * A set of static constants and methods to facilitate computation and display of interface elements  
 */
public class Form {
	public static String indentLevel1 = "  ";
	public static String indentLevel2 = indentLevel1.concat(indentLevel1);
	public static String indentLevel3 = indentLevel2.concat(indentLevel1);
	public static String newline = "\r\n";

	private static char nextDistinguishingChar = 'A';
	
	static public void initializeNextDistinguishingCharacter() {
		Form.nextDistinguishingChar = 'A';
	}
	
	static public String nextDistinguishingCharacter() {
		String retString = String.valueOf(Form.nextDistinguishingChar);
		Form.nextDistinguishingChar++; 
		return retString;
	}
	
    /**
     * Output an array of strings, justified to the given margins, from one input string.
     * @param left The left margin.
     * @param right The right margin (as an offset from the left edge).
     * @param value The string to convert to a justified array of strings.
     * @return The array of strings.
     */
    static private ArrayList<String> justifyToArray(int left, int right, String value)
    {
        /* Break the value into segments no greater than 'right - length'. */
        int width = right - left;
        ArrayList<String> lines = new ArrayList<String>();
        while (value.length() > 0)
        {
            if (value.length() < width)
            {
                lines.add(value);
                break;
            }

            String strAtWidth = value.substring(0, width);
            int p = strAtWidth.lastIndexOf(' ');
            if (p < 0)
                p = value.indexOf(' ', width);

            if (p < 0)
            {
                lines.add(value);
                break;
            }

            String line = value.substring(0, p);
            lines.add(line);
            value = value.substring(p + 1);
        }

        String header = "";
        for (int i = 0; i < left; i++)
            header = header + " ";
        
        for (int i=0; i<lines.size(); i++) {
        	String str = header + lines.get(i);
        	lines.set(i, str);
        }
        
        return lines;
    }   

    /**
     * Return one string that holds a set of text lines that are justified to the given margins, from one input string.
     * @param value The string to convert to a justified set of text lines.
     * @return The justified set of text lines.
     */
    static public String justifyDescription(String value) {
    	ArrayList<String> lines = Form.justifyToArray(Form.indentLevel3.length(), 70, value);
    	String retString = new String();
        for (String line : lines)
        	retString += (line + Form.newline);
        return retString;
    }
	
    /**
     * Return display display string selected by requested language (if available), then by first offered language
     * @param langStrings list of offered strings, by language; cannot be null
     * @param strRequestedLanguage the requested language, i.e. en, ja
     * @return selected string: null if langStrings is empty or if all display string lookups are null
     * @throws Exception
     */
    static private String getSelectedDisplayString_RequestedLanguageFirstThenAsFirstOffered(List<LanguageString> langStrings, String strRequestedLanguage) throws Exception {
        String retString = null;
        String firstFoundString = null;
        for (LanguageString langString : langStrings) {
            String cData = langString.getCdata();
            if (firstFoundString==null && cData!=null)
                firstFoundString = cData;
            String strLanguage = langString.getAttrValue();
            if (strLanguage.equals(strRequestedLanguage)) {
                retString = cData;
                break;
            }
        }

        // use the most we can get: not null and not empty
        if (firstFoundString!=null &&
            (retString==null || retString.isEmpty())) {
            retString = firstFoundString;
        }

        return retString;
    }

    /**
     * Return display name of the given metadata element.
     * @param meElement The metadata element.
     * @param strRequestedLanguage Specify the language to extract from the metadata element.
     * @return The display name.
     * @throws Exception thrown if the element display name could not be calculated.
     */
    static public String getInterfaceElementDisplayName(MetadataElements meElement, String strRequestedLanguage) throws Exception {
        List<LanguageString> elementDisplayNames = meElement.getDisplayNames();
        return Form.getSelectedDisplayString_RequestedLanguageFirstThenAsFirstOffered(elementDisplayNames, strRequestedLanguage);
    }

    /**
     * Return description of the given metadata element.
     * @param meElement The metadata element.
     * @param strRequestedLanguage Specify the language to extract from the metadata element.
     * @return The description.
     * @throws Exception thrown if the element description could not be calculated
     */
    static public String getInterfaceElementDescription(MetadataElements meElement, String strRequestedLanguage) throws Exception {
        List<LanguageString> elementDescriptions = meElement.getDescriptions();
        return Form.getSelectedDisplayString_RequestedLanguageFirstThenAsFirstOffered(elementDescriptions, strRequestedLanguage);
    }
    
	/**
	 * From a byte array, return a set of comma separated hex strings, 1 hex string per byte. 
	 * @param bytes The byte array to process.
	 * @return The set of comma separated hex strings.
	 */
	static public String bytesToCommaSeparatedHex(byte [] bytes) {
		String retString = new String("");
        for (int i = 0; i < bytes.length; i++)
        {
            String s = Integer.toHexString(bytes[i]);
            if (s.length() > 2)
                s = s.substring(s.length() - 2);
            else if (s.length() == 1)
                s = "0" + s;
            s = s.toUpperCase(); // 0xf8 => 0xF8
            
        	if (retString.endsWith(","))
        		retString += " ";
            retString += ("0x" + s + ",");
        }
        return retString;
	}
	
    /**
     * From an integer value, return a compressed 16-bit number in comma separated hex form. A trailing comma is applied.
     * @param value The integer value to process.
     * @return The number in comma-separated hex form.
     */
    static public String binaryC16Bit(int value) {
        if (value <= 0x7F)
            return Form.bytesToCommaSeparatedHex(new byte[] { (byte)(value & 0xFF) });
        else
            return Form.bytesToCommaSeparatedHex((new byte[] { (byte)(0x80 | ((value >> 8) & 0xFF)), (byte)(value & 0xFF) }));
    }
    
    /**
     * From an integer value, return a compressed 24-bit number in comma separated hex form. A trailing comma is applied.
     * @param value The integer value to process.
     * @return The number in comma-separated hex form.
     */
    static public String binaryC24Bit(int value)
    {
        if (value <= 0x7F)
        	return Form.bytesToCommaSeparatedHex(new byte[] { (byte)(value & 0xFF) });
        else if (value <= 0x3FFF)
        	return Form.bytesToCommaSeparatedHex((new byte[] { (byte)(0x80 | ((value >> 8) & 0xFF)), (byte)(value & 0xFF) }));
        else
        	return Form.bytesToCommaSeparatedHex((new byte[] { (byte)(0xC0 | ((value >> 16) & 0xFF)), (byte)((value >> 8) & 0xFF), (byte)(value & 0xFF) }));
    }
    
    
    /**
     * Determine the compressed length of a 16-bit number.
     * @param value The number.
     * @return The compressed length.
     */
    static public int binaryC16BitLength(int value)
    {
        if (value <= 0x7F)
            return 1;
        else
            return 2;
    }
    
    /**
     * Determine the compressed length of a 24-bit number.
     * @param value The number.
     * @return The compressed length.
     */
    static public int binaryC24BitLength(int value) {
        if (value <= 0x7F)
            return 1;
        else if (value <= 0x3FFF)
            return 2;
        else
            return 3;
    }

}