package org.opendof.tools.repository.interfaces.opendof.iface.compute;

import java.util.ArrayList;
import java.util.List;
import java.util.TreeMap;

import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.DofTypeBase;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection.DofArrayElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection.DofBlobElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection.DofStructureElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection.FieldData;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.other.DofStringElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.reference.DofNullableElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.reference.DofReference;

/**
 * Temporary storage for a type definition.
 * 
 * Note: the reference number of this type, supplied in the interface definition) is not stored here
 */
public class TypeInstance
{
	public static final short STRING_TYPEID = 0x31;
	public static final short BLOB_TYPEID = 0x49;
	public static final short ARRAY_TYPEID = 0x4D;
	public static final short REFERENCE_TYPEID = 0x51;
	public static final short NULLABLE_TYPEID = 0x53;
	public static final short STRUCTURE_TYPEID = 0x7801;
	
	/* Local copy of caller's mapTypeInstances. We receive it incomplete, but it is sufficient for looking up previous types that this type might refer to. */
    private TreeMap<Short, TypeInstance> mapTypeInstances;
    
    private final DofTypeBase typeElement;
	private final String typename;

    // DOF Common Types protocol specification: Compressed Unsigned 16 bits, which equates to 2**15 max
    private short typeId;        // 0x00 for uint8, 0x02 for int 8, etc.
    private final short typeReference; // reference to type position in the type context field (i.e. type count followed by list of types)

    // metadata, considered optional; these defaults are generally overwritten by processing xml-supplied data
    private String codename = "";
    private String name = "";
    private String description = "";

    // DOF Common Types protocol specification: Compressed Unsigned 24 bits, which equates to 2**23 max
    private int length = 0; // String: encoded length; General string: max char length; Blob: length; 

    // DOF Object Model Types protocol specification: Compressed Unsigned 16 bits, equates to 2**15 max
    private short encoding = -1;

    private int min_length = -1; // where is the max value of this specified?
    
    private short refersTo = -1;
    private List<FieldData> fields = null;

    public short getTypeId() {
    	return this.typeId;
    }

    public short getTypeReference() {
    	return this.typeReference;
    }

    public short getRefersTo() {
    	return this.refersTo;
    }
    
    public String getName() {
    	return this.name;
    }
    
    public String getDescription() {
    	return this.description;
    }

    public boolean isSimpleType() {
    	return (this.typeId % 2) == 0;
    }
    
    public int getLength() {
    	return this.length;
    }
    public int getMinLength() {
    	return this.min_length;
    }
    
    public short getEncoding() {
    	return this.encoding;
    }

    public String getCodeName() {
    	return this.codename;
    }
    
    public List<FieldData> getFields() {
    	return this.fields;
    }

    /**
     * Return type name. For other than simple types, add referenced type information,
     * for example: "uint32," or "structure{ uint8, blob }"
     * @return The type name.
     * @throws Exception thrown if the decorated type name could not be calculated
     */
    public String getDecoratedTypeName() throws Exception {
    	String retString = this.typename;
    	if (!this.isSimpleType()) {
    		retString += ("( ");
    		
    		switch (this.typename) {
    		case "string":
    			retString += (this.encoding + ", " + this.length + " ");
    			break;
    		case "blob":
    			retString += (this.min_length + ", " + this.length + " "); 
    			break;
    		case "array":
    			TypeInstance typeInstance = this.mapTypeInstances.get(this.refersTo); 
    			retString += (this.min_length + ", " + this.length + ", " + typeInstance.getCodeName() + " ");
    			break;
    		case "reference":
    			typeInstance = this.mapTypeInstances.get(this.refersTo);
    			retString += typeInstance.getCodeName();
               	// note: No example available, but we known that, at output time, reference type gets handled the same as nullable type; so we do that.
    			break;
    		case "nullable":
    			typeInstance = this.mapTypeInstances.get(this.refersTo);
    			retString += (typeInstance.getCodeName() + " ");
    			break;
    		case "structure":
    			for (FieldData fieldData : this.fields) {
    				short structRef = (short)(fieldData.getReference());
    				typeInstance = this.mapTypeInstances.get(structRef);
    				if (typeInstance == null)
    					throw new Exception("type instance reference not found");
    				retString += (typeInstance.getCodeName() + ", ");
    			}
    			break;
    		default:
    			break;
    		}
    		
    		retString += ")"; 
    	}
     	return retString;
    }

    /**
     * Constructor
	 * @param typeElement the element type
	 * @param strRequestedLanguage the language to generate
	 * @param mapTypeInstances the map of type isntances
	 * @throws Exception thrown if the TypeInstance could not be created
	 *
     */
    public TypeInstance(DofTypeBase typeElement, String strRequestedLanguage, TreeMap<Short, TypeInstance> mapTypeInstances) throws Exception {
    	this.mapTypeInstances = mapTypeInstances;
		this.typeElement = typeElement;
    	this.typename = typeElement.getTypeAsString(); // e.g. "uint8"
    	this.typeReference = (short)typeElement.getTypeId(); // this is the special transition: the xml contains a reference to type but calls that value a typeId
    	
    	// where possible, override default metadata codename, name, and description
    	MetadataElements mElements = typeElement.getMetadata();    	
    	if (mElements != null) {
     		String localCodeName = mElements.getCodeName(); // e.g. "Status Type"
    		this.codename = localCodeName!=null ? localCodeName : "Type_" + Form.nextDistinguishingCharacter();
    		String localName = Form.getInterfaceElementDisplayName(mElements, strRequestedLanguage); // e.g. "Status"
    		if (localName != null)
    			this.name = localName;
    		String localDescription = Form.getInterfaceElementDescription(mElements, strRequestedLanguage); // e.g. "This is ..."
    		if (localDescription != null)
    			this.description = localDescription;
    	}
        this.computeDetail();
	}

	private void computeDetail() throws Exception {
        if (this.typename.equals("uint8"))
        	this.typeId = 0x00;
        else if (this.typename.equals("uint16"))
        	this.typeId = 0x02;
        else if (this.typename.equals("uint32"))
        	this.typeId = 0x04;
        else if (this.typename.equals("uint64"))
        	this.typeId = 0x06;
        else if (this.typename.equals("int8"))
        	this.typeId = 0x10;
        else if (this.typename.equals("int16"))
        	this.typeId = 0x12;
        else if (this.typename.equals("int32"))
        	this.typeId = 0x14;
        else if (this.typename.equals("int64"))
        	this.typeId = 0x16;
        else if (this.typename.equals("float32"))
        	this.typeId = 0x20;
        else if (this.typename.equals("float64"))
        	this.typeId = 0x22;
        else if (this.typename.equals("datetime"))
        	this.typeId = 0x40;
        else if (this.typename.equals("boolean"))
        	this.typeId = 0x42;
        else if (this.typename.equals("iid"))
        	this.typeId = 0x44;
        else if (this.typename.equals("oid"))
        	this.typeId = 0x46;
        else if (this.typename.equals("uuid"))
        	this.typeId = 0x48;
        else if (this.typename.equals("string")) {
        	this.typeId = STRING_TYPEID;
            if (!this.typeElement.getClass().isAssignableFrom(DofStringElement.class))
            	throw new Exception("typeElement (DofTypeBase) is not Assignable from class " + DofStringElement.class);
           	DofStringElement dofStringElement = (DofStringElement)typeElement;
        	this.length = dofStringElement.getLength() & 0x007fffff;  // max length of the String being described
        	this.encoding = (short)dofStringElement.getEncoding();
        } else if (typename.equals("blob")) {
            typeId = TypeInstance.BLOB_TYPEID;
            if (!this.typeElement.getClass().isAssignableFrom(DofBlobElement.class))
            	throw new Exception("typeElement (DofTypeBase) is not Assignable from class " + DofBlobElement.class);
           	DofBlobElement dofBlobElement = (DofBlobElement)typeElement;
       	 	this.length = dofBlobElement.getLength() & 0x007fffff;      // max length of the Blob
       	 	this.min_length = dofBlobElement.getMinLength().intValue(); // min length of the Blob
        } else if (typename.equals("array")) {
            typeId = TypeInstance.ARRAY_TYPEID;
            if (!this.typeElement.getClass().isAssignableFrom(DofArrayElement.class))
            	throw new Exception("typeElement (DofTypeBase) is not Assignable from class " + DofArrayElement.class);
           	DofArrayElement dofArrayElement = (DofArrayElement)typeElement;
           	this.length = dofArrayElement.getLength();			// max length of the Array
           	this.min_length = dofArrayElement.getMinLength();	// min length of the Array
           	this.refersTo = (short)dofArrayElement.getReference();
        } else if (typename.equals("reference")) {
            typeId = REFERENCE_TYPEID;
            if (!this.typeElement.getClass().isAssignableFrom(DofReference.class))
            	throw new Exception("typeElement (DofTypeBase) is not Assignable from class " + DofReference.class);
           	DofReference dofReference = (DofReference)typeElement;
           	int ref = dofReference.getReference();
           	this.refersTo = (short)ref;
           	// note: No example available, but we known that, at output time, reference type gets handled the same as nullable type; so we do that.
        } else if (typename.equals("nullable")) {
            typeId = TypeInstance.NULLABLE_TYPEID;
            if (!this.typeElement.getClass().isAssignableFrom(DofNullableElement.class))
            	throw new Exception("typeElement (DofTypeBase) is not Assignable from class " + DofNullableElement.class);
           	DofNullableElement dofNullableElement = (DofNullableElement)typeElement;
           	int ref = dofNullableElement.getReference();
           	this.refersTo = (short)ref;
        } else if (typename.equals("structure")) {
            typeId = STRUCTURE_TYPEID;
            if (!this.typeElement.getClass().isAssignableFrom(DofStructureElement.class))
            	throw new Exception("typeElement (DofTypeBase) is not Assignable from class " + DofStructureElement.class);
           	DofStructureElement dofStructureElement = (DofStructureElement)typeElement;
           	
           	List<FieldData> listFieldData = dofStructureElement.getFields();
           	if (this.fields==null && !listFieldData.isEmpty())
           		this.fields = new ArrayList<FieldData>();
           	for (FieldData fieldData : listFieldData) {
           		this.fields.add(fieldData);
           	}
        } else
            throw new Exception("Unrecognized type.");
	}
	
}