/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.slf4j.LoggerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Metadata list Element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class MetadataElements extends OpenDofSaxHandler
{
    private static final String XmlLangQname = "xml:lang";
    
    private static final String CodeNameElement = "code-name";
    private static final String KnownValueElement = "known-value";
    private static final String DisplayNameElement = "display-name";
    private static final String DescriptionElement = "description";
    private static final String NamespaceAbbreviation = "md:";
    
    private volatile String codeName;
    private final List<LanguageString> descriptions;
    private final List<LanguageString> displayNames;
    private final List<LanguageString> knownValues;
    
    //working variables follow
    private StringBuilder cdata;
    private LanguageString langString;
    
    /**
     * Default MetadataElements constructor 
     */
    public MetadataElements()
    {
        descriptions = new ArrayList<LanguageString>();
        displayNames = new ArrayList<LanguageString>();
        knownValues = new ArrayList<LanguageString>();
    }
    
    public MetadataElements(String codeName)
    {
        this.codeName = codeName;
        descriptions = new ArrayList<LanguageString>();
        displayNames = new ArrayList<LanguageString>();
        knownValues = new ArrayList<LanguageString>();
    }
    
    public String getCodeName() {
    	return codeName;
    }

    public String getCdata()
    {
        return cdata.toString();
    }
    
    public List<LanguageString> getDescriptions()
    {
        synchronized (descriptions)
        {
            return new ArrayList<LanguageString>(descriptions);
        }
    }
    
    public void addDescription(LanguageString value)
    {
        synchronized (descriptions)
        {
            descriptions.add(value);
        }
    }
    
    public List<LanguageString> getDisplayNames()
    {
        synchronized (displayNames)
        {
            return new ArrayList<LanguageString>(displayNames);
        }
    }
    
    public void addDisplayName(LanguageString value)
    {
        synchronized (displayNames)
        {
            displayNames.add(value);
        }
    }
    
    public List<LanguageString> getKnownValues()
    {
        synchronized (knownValues)
        {
            return new ArrayList<LanguageString>(knownValues);
        }
    }
    
    public void addKnownValue(LanguageString value)
    {
        synchronized (knownValues)
        {
            knownValues.add(value);
        }
    }

    
    /**
     * Validate that this MetadataElements element meets OpenDOF requirements
     * @param codeNameRequired set true if code name is required. 
     * @param aloDisplayName set true if at least one display name is required.
     * @param aloDescription set true if at least one description is required. 
     * @param location string identifying the containing element of this metadata.
     * @throws SAXException if the validation fails.
     */
    public void validate(boolean codeNameRequired, boolean aloDisplayName, boolean aloDescription, String location) throws SAXException
    {
        Hashtable<String, String> uniqueMap = new Hashtable<String, String>();
        if(codeName != null && codeName.length() < 1)
            throw new SAXException("codename with empty string");
        if(codeNameRequired)
        {
            if(codeName == null || codeName.length() < 1)
            {
                //TODO: fix these
                if(OpenDofSaxHandler.isPublish())
                    LoggerFactory.getLogger(getClass()).warn("MetadataElement published without a code-name");
//                    throw new SAXException("codename is required at: " + location);
                LoggerFactory.getLogger(getClass()).warn("MetadataElement without a code-name");
            }
        }
        synchronized (descriptions)
        {
            if(aloDescription)
            {
                    if(descriptions.size() == 0)
                        throw new SAXException("At least one description is required at: " + location);
            }
            for(LanguageString lang : descriptions)
            {
                lang.validate();
                if(uniqueMap.get(lang.attrValue) != null)
                    throw new SAXException("description with duplicate language: " + lang.attrValue);
                uniqueMap.put(lang.attrValue, lang.attrValue);
            }
        }
        synchronized (displayNames)
        {
            if(aloDisplayName)
            {
                if(displayNames.size() == 0)
                    throw new SAXException("At least one display name is required at: " + location);
            }
            uniqueMap.clear();
            for(LanguageString lang : displayNames)
            {
                lang.validate();
                if(uniqueMap.get(lang.attrValue) != null)
                    throw new SAXException("displayName with duplicate language: " + lang.attrValue);
                uniqueMap.put(lang.attrValue, lang.attrValue);
            }
        }
    }
    
    public static boolean oneOfYours(String element)
    {
        if(CodeNameElement.equals(element))
            return true;
        if(KnownValueElement.equals(element))
            return true;
        if(DisplayNameElement.equals(element))
            return true;
        return DescriptionElement.equals(element);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        cdata = new StringBuilder();
        if(localName.equals(DisplayNameElement) || localName.equals(DescriptionElement) || localName.equals(KnownValueElement))
        {
            stack.push(localName);
            int attrCount = 1;
            if(localName.equals(KnownValueElement))
            	attrCount = 2;
            langString = new LanguageString(attributes, attrCount);
            langString.currentPath = currentPath();
            if(DescriptionElement.equals(localName))
            {
                synchronized (descriptions)
                {
                    descriptions.add(langString);
                }
                return;
            }
            else if(DisplayNameElement.equals(localName))
            {
                synchronized (displayNames)
                {
                    displayNames.add(langString);
                }
                return;
            }
            else if(KnownValueElement.equals(localName))
            {
                synchronized (knownValues)
                {
                    knownValues.add(langString);
                }
                return;
            }
        }
        if(localName.equals(CodeNameElement))
        {
            stack.push(localName);
            return;
            
        }
        throw new SAXException(currentPath() + "."+localName + " unexpected localName: " + localName);
   }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        if(isMyTag(CodeNameElement) || isMyTag(DisplayNameElement) || isMyTag(DescriptionElement) || isMyTag(KnownValueElement))
        {
            cdata.append(new String(ch, start, length));
            return;
        }
        String value = new String(ch, start, length);
        throw new SAXException(currentPath() + " characters unknown current stack: " + value);
    }
    
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if(localName.equals(CodeNameElement))
        {
            codeName = cdata.toString();
            stack.pop();
            return;
        }
                        
        if(localName.equals(DisplayNameElement) || localName.equals(DescriptionElement) || localName.equals(KnownValueElement))
        {
            langString.cdata = cdata.toString();
            stack.pop();
            return;
        }
        throw new SAXException(currentPath() + "."+localName + " unexpected endElement: " + localName);
    }
    
    public static class LanguageString
    {
        public volatile String cdata;
        public final String attrUri;
        public final String attrQname;
        public final String attrValue;
        public String currentPath;
        
        public LanguageString(Attributes attributes, int attrCount) throws SAXException
        {
            if(attributes.getLength() != attrCount)
                throw new SAXException("expected " + attrCount + " ns attributes got: " + attributes.getLength());
            attrUri = attributes.getURI(0);
            attrQname = attributes.getQName(0);
            attrValue = attributes.getValue(0);
        }
        
        public LanguageString(String attrValue, String cdata)
        {
            this.attrUri = InterfaceElement.W3Namespace;
            this.attrQname = "xml:lang";
            this.attrValue = attrValue;
            this.cdata = cdata;
        }

        public String getAttrQname()
        {
            return attrQname;
        }
        
        public String getAttrValue()
        {
            return attrValue;
        }
        
        public String getCdata()
        {
            return cdata;
        }
        
        /**
         * Validate if the language cdata string is valid.
         * @throws SAXException if the string is invalid
         */
        public void validate() throws SAXException
        {
            if(cdata == null || cdata.length() < 1)
                throw new SAXException("cdata is null or empty string");
            if(attrUri == null || attrUri.length() < 1)
                throw new SAXException("attrUri is null or empty string");
            if(!attrUri.equals(InterfaceElement.W3Namespace))
                throw new SAXException("attrUri unexpected: " + attrUri);
            if(attrQname == null || attrQname.length() < 1)
                throw new SAXException("attrQname is null or empty string");
            if(!attrQname.equals(XmlLangQname))
                throw new SAXException("attrQname unexpected: " + attrQname);
            if(attrValue == null || attrValue.length() < 1)
                throw new SAXException("attrValue is null or empty string");
        }
        
        @Override
        public String toString()
        {
            return "attrUri: " + attrUri + " attrQname: " + attrQname + " attrValue: " + attrValue + " cdata: " + cdata;
        }
        
        public void export(StringBuilder sb, int level, String tag)
        {
            element(sb, tag, attrQname, attrValue, cdata, true, level);
        }
    }
    
    /**
     * Export the MetadataElements element as XML
     * @param sb  the StringBuilder to add the XML to.  Must not be null.
     * @param level the indent level for this element.
     * @return the StringBuilder handed in with this element's XML appended.
     */
    public StringBuilder export(StringBuilder sb, int level)
    {
        if(codeName != null && codeName.length() > 0)
            element(sb, NamespaceAbbreviation+CodeNameElement, codeName, true, level);
        synchronized (knownValues)
        {
            for(LanguageString ls : knownValues)
                ls.export(sb, level, NamespaceAbbreviation+KnownValueElement);
        }
        synchronized (displayNames)
        {
            for(LanguageString ls : displayNames)
                ls.export(sb, level, NamespaceAbbreviation+DisplayNameElement);
        }
        synchronized (descriptions)
        {
            for(LanguageString ls : descriptions)
                ls.export(sb, level, NamespaceAbbreviation+DescriptionElement);
        }
        return sb;
    }
}
