/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.event;

import java.util.ArrayList;
import java.util.List;

import org.opendof.tools.repository.interfaces.opendof.saxParser.InterfaceElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler.ChildComplete;
import org.opendof.tools.repository.interfaces.opendof.saxParser.method.MethodElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection.FieldData;
import org.slf4j.LoggerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Event Element SAX handler.  
 */
public class EventElement extends OpenDofSaxHandler implements ChildComplete
{
    /** The Event Element tag */
    public static final String EventElement = "event";
    private static final String MyPath = InterfaceElement.InterfaceElement + "." + EventsElement.EventsElement + "." + EventElement;

    private final List<FieldData> outputs;
    /** The itemId of this event */
    public volatile int itemId;
    private volatile MetadataElements metadata;

    // working
    private FieldData currentParameter;
    private final ChildComplete childComplete;

    /**
     * EventElement constructor
     * @param childComplete the child completion callback to my parent.  Must not be null.
     */
    public EventElement(ChildComplete childComplete)
    {
        this.childComplete = childComplete;
        outputs = new ArrayList<FieldData>();
    }

    public EventElement(int itemId, MetadataElements metadata)
    {
        childComplete = null;
        outputs = new ArrayList<FieldData>();
        this.itemId = itemId;
        this.metadata = metadata;
    }

    public int getItemId()
    {
        return itemId;
    }

    public MetadataElements getMetadata()
    {
        return metadata;
    }

    public List<FieldData> getOutputs()
    {
        synchronized (outputs)
        {
            return new ArrayList<FieldData>(outputs);
        }
    }

    public void addOutput(FieldData output)
    {
        synchronized (outputs)
        {
            outputs.add(output);
        }
    }

    /**
     * Validate that this Event element meets OpenDOF requirements
     * @param typedefs the typedefs structure to validate typeIds with.  Must not be null
     * @throws SAXException if the validation fails.
     */
    public void validate(TypedefsElement typedefs) throws SAXException
    {
        if (metadata == null)
        {
            if (OpenDofSaxHandler.isPublish())
                throw new SAXException("EventElement.metadata is null");
            LoggerFactory.getLogger(getClass()).warn("MetadataElement.metadata is null");
        }
        String location = "interface.methods";
        if (metadata != null)
            metadata.validate(true, false, false, location);
        if (itemId < 0 || itemId > 32767)
            throw new SAXException("event.itemId is less than 0 or greater than 32767, " + location);
        synchronized (outputs)
        {
            for (FieldData param : outputs)
                param.validate(typedefs, location);
        }
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (localName.equals(EventElement))
        {
            // The stack here is "interface,events"
            if (attributes.getLength() != 1)
                throw new SAXException(currentPath() + "."+localName + " expected exactly 1 attibutes");
            int index = attributes.getIndex("", InterfaceElement.ItemIdAttr);
            if (index < 0)
                throw new SAXException(currentPath() + "."+localName + " could not find property: " + InterfaceElement.ItemIdAttr);
            itemId = Integer.parseInt(attributes.getValue(index));
            stack.push(EventElement);
            return;
        }

        if (stack.peek().equals(EventElement))
        {
            // The stack here is "interface,events,event"
            // valid children are metadata, outputs
            if (MetadataElements.oneOfYours(localName))
            {
                if (metadata == null)
                    metadata = new MetadataElements();
                metadata.startElement(uri, localName, qName, attributes);
                return;
            }
        }
        if (currentPath().startsWith(MyPath))
        {
            if (localName.equals(MethodElement.OutputsElement))
            {
                // The stack here is "interface,events,event"
                stack.push(MethodElement.OutputsElement);
                return;
            }
            if (localName.equals(FieldData.OutputElement))
            {
                // The stack here is "interface,events,event,outputs"
                currentParameter = new FieldData(this);
                currentParameter.startElement(uri, localName, qName, attributes);
                return;
            }
            // The stack here is "interface,events,event,outputs,typeref,..."
            currentParameter.startElement(uri, localName, qName, attributes);
            return;
        }
        throw new SAXException(currentPath() + "."+localName + " startElement unknown localName for this level: " + localName);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        if (currentPath().startsWith(MyPath))
        {
            // The stack here is "interface,events,event,..."
            if (currentParameter == null)
            {
                // The stack here is "interface,events,event,<metadata>
                metadata.characters(ch, start, length);
                return;
            }
            // The stack here is "interface,events,event,<iputs || outputs>"
            currentParameter.characters(ch, start, length);
            return;
        }
        String value = new String(ch, start, length);
        throw new SAXException(currentPath() + " characters unexpected: " + value);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (localName.equals(EventElement))
        {
            stack.pop();
            childComplete.setChildComplete(childComplete, this);
            return;
        }
        if (localName.equals(MethodElement.OutputsElement))
        {
            stack.pop();
            return;
        }
        if (currentPath().startsWith(MyPath))
        {
            // The stack here is "interface,events,event,..."
            if (currentParameter == null)
            {
                // The stack here is "interface,events,event,<metadata>
                metadata.endElement(uri, localName, qName);
                return;
            }
            // The stack here is "interface,events,event,outputs"
            currentParameter.endElement(uri, localName, qName);
            return;
        }
        throw new SAXException(currentPath() + "."+localName + " endElement unknown localName for this level: " + localName);
    }

    @Override
    public void setChildComplete(Object listener, Object obj)
    {
        synchronized (outputs)
        {
            outputs.add((FieldData) obj);
        }
        currentParameter = null;
    }

    /**
     * Export this Event element as XML
     * @param sb  the StringBuilder to add the XML to.  Must not be null.
     * @param level the indent level for this element.
     * @return the StringBuilder handed in with this element's XML appended.
     */
    public StringBuilder export(StringBuilder sb, int level)
    {
        synchronized (outputs)
        {
            element(sb, EventElement, InterfaceElement.ItemIdAttr, "" + itemId, null, metadata == null && outputs.size() == 0, level);
            exportNames(sb, null, metadata, true, level);
            if (outputs.size() > 0)
            {
                ++level;
                element(sb, MethodElement.OutputsElement, CType.Open, level);
                ++level;
                for (FieldData param : outputs)
                    param.export(sb, FieldData.OutputElement, level);
                --level;
                element(sb, MethodElement.OutputsElement, CType.Close, level);
                --level;
            }
            return element(sb, EventElement, CType.Close, level);
        }
    }

    /**
     * Validate if published/immutable fields have been modified. 
     * @param other The master/published element to be compared against.
     * @return true if an immutable field has changed, false otherwise.
     */
    public boolean validPublish(EventElement other)
    {
        if (itemId != other.itemId)
            return false;
        try
        {
            synchronized (outputs)
            {
                synchronized (other.outputs)
                {
                    for (int i = 0; i < outputs.size(); i++)
                    {
                        FieldData field = outputs.get(i);
                        if (!field.validPublish(other.outputs.get(i)))
                            return false;
                    }
                }
            }
        } catch (Exception e)
        {
            return false;
        }
        return true;
    }
}
