/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.event;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.opendof.tools.repository.interfaces.opendof.saxParser.InterfaceElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler.ChildComplete;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Events Element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class EventsElement extends OpenDofSaxHandler implements ChildComplete
{
    /** The Events Element tag */
    public static final String EventsElement = "events";

    public static final String MyTag = "event";
    private static final String EventsPath = InterfaceElement.InterfaceElement + "," + MyTag;

    private final List<EventElement> events;

    // working
    private EventElement currentEvent;

    /**
     * Default Events element constructor 
     */
    public EventsElement()
    {
        events = new ArrayList<EventElement>();
    }

    public List<EventElement> getEvents()
    {
        synchronized (events)
        {
            return new ArrayList<EventElement>(events);
        }
    }

    public void addEvent(EventElement event)
    {
        synchronized (events)
        {
            events.add(event);
        }
    }

    public void uniqueItemId(Hashtable<Integer, Integer> uniqueMap) throws SAXException
    {
        synchronized (events)
        {
            for (EventElement data : events)
            {
                Integer itemId = data.itemId;
                if (uniqueMap.get(itemId) != null)
                    throw new SAXException(currentPath() + " ItemId is not unique: " + itemId);
                uniqueMap.put(itemId, itemId);
            }
        }
    }

    /**
     * Validate that this Events element meets OpenDOF requirements
     * @param typedefs the typedefs structure to validate typeIds with.  Must not be null
     * @throws SAXException if the validation fails.
     */
    public void validate(TypedefsElement typedefs) throws SAXException
    {
        synchronized (events)
        {
            for (EventElement event : events)
                event.validate(typedefs);
        }
    }

    public static boolean oneOfYours(String element)
    {
        return MyTag.equals(element);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (localName.equals(EventsElement))
        {
            stack.push(localName);
            if (attributes.getLength() != 0)
                throw new SAXException(currentPath() + "."+localName + " " + EventsElement + " should not have any attributes");
            currentEvent = new EventElement(this);
            return;
        }
        if (localName.equals(EventElement.EventElement))
            // The stack here is: "interface,events"
            currentEvent = new EventElement(this);

        // The stack here is "interface,events,..."  let the children determine error
        currentEvent.startElement(uri, localName, qName, attributes);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        if (isMyTag(EventsElement))
        {
            // The stack here is: "interface,events"
            String value = new String(ch, start, length);
            throw new SAXException(currentPath() + " characters unexpected: " + value);
        }
        // The stack here is: "interface,events,..."
        // let children deal with any errors
        currentEvent.characters(ch, start, length);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (localName.equals(EventsElement))
        {
            // The stack here is: "interface,events"
            stack.pop();
            return;
        }
        // The stack here is: "interface,events,..."
        // let children deal with any errors
        currentEvent.endElement(uri, localName, qName);
    }

    @Override
    public void setChildComplete(Object listener, Object obj)
    {
        synchronized (events)
        {
            events.add((EventElement) obj);
        }
        currentEvent = null;
    }

    /**
     * Export this Events element as XML
     * @param sb  the StringBuilder to add the XML to.  Must not be null.
     * @param level the indent level for this element.
     * @return the StringBuilder handed in with this element's XML appended.
     */
    public StringBuilder export(StringBuilder sb, int level)
    {
        element(sb, EventsElement, CType.Open, level);
        ++level;
        synchronized (events)
        {
            for (EventElement event : events)
                event.export(sb, level);
        }
        --level;
        return element(sb, EventsElement, CType.Close, level);
    }

    /**
     * Validate if published/immutable fields have been modified. 
     * @param other The master/published element to be compared against.
     * @return true if an immutable field has changed, false otherwise.
     */
    public boolean validPublish(EventsElement other)
    {
        try
        {
            synchronized (events)
            {
                synchronized (other.events)
                {
                    if (other.events.size() != events.size())
                        return false;
                    for (int i = 0; i < events.size(); i++)
                    {
                        EventElement method = events.get(i);
                        if (!method.validPublish(other.events.get(i)))
                            return false;
                    }
                }
            }
        } catch (Exception e)
        {
            return false;
        }
        return true;
    }
}
