/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.exception;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.opendof.tools.repository.interfaces.opendof.saxParser.InterfaceElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler.ChildComplete;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Exceptions Element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class ExceptionsElement extends OpenDofSaxHandler implements ChildComplete
{
    /** The Exceptions Element tag */
    public static final String ExceptionsElement = "exceptions";

    public static final String MyTag = "exception";
    public static final String ExceptionsPath = InterfaceElement.InterfaceElement + "," + MyTag;

    private final List<ExceptionElement> exceptions;

    // working
    private ExceptionElement currentException;

    /**
     * Default Exceptions element constructor 
     */
    public ExceptionsElement()
    {
        exceptions = new ArrayList<ExceptionElement>();
    }

    public List<ExceptionElement> getExceptions()
    {
        synchronized (exceptions)
        {
            return new ArrayList<ExceptionElement>(exceptions);
        }
    }

    public void addException(ExceptionElement exception)
    {
        synchronized (exceptions)
        {
            exceptions.add(exception);
        }
    }

    public void exceptionMethod(ExceptionElement exception)
    {
        synchronized (exceptions)
        {
            exceptions.add(exception);
        }
    }

    public void uniqueItemId(Hashtable<Integer, Integer> uniqueMap) throws SAXException
    {
        synchronized (exceptions)
        {
            for (ExceptionElement data : exceptions)
            {
                Integer itemId = data.itemId;
                if (uniqueMap.get(itemId) != null)
                    throw new SAXException("ItemId is not unique: " + itemId);
                uniqueMap.put(itemId, itemId);
            }
        }
    }

    /**
     * Validate that this Exceptions element meets OpenDOF requirements
     * @param typedefs the typedefs structure to validate typeIds with.  Must not be null
     * @throws SAXException if the validation fails.
     */
    public void validate(TypedefsElement typedefs) throws SAXException
    {
        synchronized (exceptions)
        {
            for (ExceptionElement ex : exceptions)
                ex.validate(typedefs);
        }
    }

    public static boolean oneOfYours(String element)
    {
        return MyTag.equals(element);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (localName.equals(ExceptionsElement))
        {
            stack.push(localName);
            if (attributes.getLength() != 0)
                throw new SAXException(currentPath() + "."+localName + " " + ExceptionsElement + " should not have any attributes");
            currentException = new ExceptionElement(this);
            return;
        }
        if (localName.equals(ExceptionElement.ExceptionElement))
            // The stack here is: "interface,exceptions"
            currentException = new ExceptionElement(this);

        // The stack here is "interface,exceptions,..."  let the children determine error
        currentException.startElement(uri, localName, qName, attributes);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        if (isMyTag(ExceptionsElement))
        {
            // The stack here is: "interface,exceptions"
            String value = new String(ch, start, length);
            throw new SAXException(currentPath() + " characters unexpected: " + value);
        }
        // The stack here is: "interface,exceptions,..."
        // let children deal with any errors
        currentException.characters(ch, start, length);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (localName.equals(ExceptionsElement))
        {
            // The stack here is: "interface,exceptions"
            stack.pop();
            return;
        }
        // The stack here is: "interface,exceptions,..."
        // let children deal with any errors
        currentException.endElement(uri, localName, qName);
    }

    @Override
    public void setChildComplete(Object listener, Object obj)
    {
        synchronized (exceptions)
        {
            exceptions.add((ExceptionElement) obj);
        }
        currentException = null;
    }

    /**
     * Export this Exceptions element as XML
     * @param sb  the StringBuilder to add the XML to.  Must not be null.
     * @param level the indent level for this element.
     * @return the StringBuilder handed in with this element's XML appended.
     */
    public StringBuilder export(StringBuilder sb, int level)
    {
        element(sb, ExceptionsElement, CType.Open, level);
        ++level;
        synchronized (exceptions)
        {
            for (ExceptionElement exception : exceptions)
                exception.export(sb, level);
        }
        --level;
        return element(sb, ExceptionsElement, CType.Close, level);
    }

    /**
     * Validate if published/immutable fields have been modified. 
     * @param other The master/published element to be compared against.
     * @return true if an immutable field has changed, false otherwise.
     */
    public boolean validPublish(ExceptionsElement other)
    {
        try
        {
            synchronized (exceptions)
            {
                synchronized (other.exceptions)
                {
                    if (other.exceptions.size() != exceptions.size())
                        return false;
                    for (int i = 0; i < exceptions.size(); i++)
                    {
                        ExceptionElement exception = exceptions.get(i);
                        if (!exception.validPublish(other.exceptions.get(i)))
                            return false;
                    }
                }
            }
        } catch (Exception e)
        {
            return false;
        }
        return true;
    }
}
