/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.method;

import java.util.ArrayList;
import java.util.List;

import org.opendof.tools.repository.interfaces.opendof.saxParser.InterfaceElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler.ChildComplete;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection.FieldData;
import org.slf4j.LoggerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Method Element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class MethodElement extends OpenDofSaxHandler implements ChildComplete
{
    /** The Method Element tag */
    public static final String MethodElement = "method";
    public static final String InputsElement = "inputs";
    public static final String OutputsElement = "outputs";

    private static final String MyPath = InterfaceElement.InterfaceElement + "." + MethodsElement.MethodsElement + "." + MethodElement;

    private final List<FieldData> inputs;
    private final List<FieldData> outputs;
    public volatile int itemId;
    private volatile MetadataElements metadata;

    // working
    private FieldData currentParameter;
    private final ChildComplete childComplete;
    private boolean inputsActive;

    /**
     * MethodElement constructor
     * @param childComplete the child completion callback to my parent.  Must not be null.
     */
    public MethodElement(ChildComplete childComplete)
    {
        inputs = new ArrayList<FieldData>();
        outputs = new ArrayList<FieldData>();
        this.childComplete = childComplete;
    }

    public MethodElement(int itemId, MetadataElements metadata)
    {
        childComplete = null;
        inputs = new ArrayList<FieldData>();
        outputs = new ArrayList<FieldData>();
        this.itemId = itemId;
        this.metadata = metadata;
    }

    public int getItemId()
    {
        return itemId;
    }

    public MetadataElements getMetadata()
    {
        return metadata;
    }

    public List<FieldData> getInputs()
    {
        synchronized (inputs)
        {
            return new ArrayList<FieldData>(inputs);
        }
    }

    public void addInput(FieldData input)
    {
        synchronized (inputs)
        {
            inputs.add(input);
        }
    }

    public List<FieldData> getOutputs()
    {
        synchronized (outputs)
        {
            return new ArrayList<FieldData>(outputs);
        }
    }

    public void addOutput(FieldData output)
    {
        synchronized (outputs)
        {
            outputs.add(output);
        }
    }

    /**
     * Validate that this Method element meets OpenDOF requirements
     * @param typedefs the typedefs structure to validate typeIds with.  Must not be null
     * @throws SAXException if the validation fails.
     */
    public void validate(TypedefsElement typedefs) throws SAXException
    {
        if (metadata == null)
        {
            if (OpenDofSaxHandler.isPublish())
                throw new SAXException("MethodData.metadata is null");
            LoggerFactory.getLogger(getClass()).warn("MethodData.metadata is null");
        }
        String location = "interface.methods";
        if (metadata != null)
            metadata.validate(true, false, false, location);
        if (itemId < 0 || itemId > 32767)
            throw new SAXException("ExceptionData.itemId is less than 0 or greater than 32767");
        synchronized (inputs)
        {
            for (FieldData param : inputs)
                param.validate(typedefs, location);
        }
        synchronized (outputs)
        {
            for (FieldData param : outputs)
                param.validate(typedefs, location);
        }
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (localName.equals(MethodElement))
        {
            // The stack here is "interface,methods"
            if (attributes.getLength() != 1)
                throw new SAXException(currentPath() + "."+localName + " expected exactly 1 attibutes");
            int index = attributes.getIndex("", InterfaceElement.ItemIdAttr);
            if (index < 0)
                throw new SAXException(currentPath() + "."+localName + " could not find property: " + InterfaceElement.ItemIdAttr);
            itemId = Integer.parseInt(attributes.getValue(index));
            stack.push(MethodElement);
            return;
        }

        if (stack.peek().equals(MethodElement))
        {
            // The stack here is "interface,methods,method"
            // valid children are metadata, inputs and outputs
            if (MetadataElements.oneOfYours(localName))
            {
                if (metadata == null)
                    metadata = new MetadataElements();
                metadata.startElement(uri, localName, qName, attributes);
                return;
            }
        }
        if (currentPath().startsWith(MyPath))
        {
            if (localName.equals(InputsElement))
            {
                // The stack here is "interface,methods,method"
                inputsActive = true;
                stack.push(InputsElement);
                return;
            }
            if (localName.equals(OutputsElement))
            {
                // The stack here is "interface,methods,method"
                inputsActive = false;
                stack.push(OutputsElement);
                return;
            }
            if (FieldData.oneOfYours(localName))
            {
                // The stack here is "interface,methods,method,<inputs|outputs>"
                currentParameter = new FieldData(this, inputsActive, false);
                currentParameter.startElement(uri, localName, qName, attributes);
                return;
            }
            // The stack here is "interface,methods,method,<inputs|outputs>,<input|output>,..."
            currentParameter.startElement(uri, localName, qName, attributes);
            return;
        }
        throw new SAXException(currentPath() + "."+localName + " startElement unknown localName for this level: " + localName);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        if (currentPath().startsWith(MyPath))
        {
            // The stack here is "interface,methods,method,..."
            if (currentParameter == null)
            {
                // The stack here is "interface,methods,method,<metadata>
                metadata.characters(ch, start, length);
                return;
            }
            // The stack here is "interface,methods,method,<iputs || outputs>"
            currentParameter.characters(ch, start, length);
            return;
        }
        String value = new String(ch, start, length);
        throw new SAXException(currentPath() + " characters unexpected: " + value);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (localName.equals(MethodElement))
        {
            stack.pop();
            childComplete.setChildComplete(childComplete, this);
            return;
        }
        if (localName.equals(InputsElement) || localName.equals(OutputsElement))
        {
            stack.pop();
            return;
        }
        if (currentPath().startsWith(MyPath))
        {
            // The stack here is "interface,methods,method,..."
            if (currentParameter == null)
            {
                // The stack here is "interface,methods,method,<metadata>
                metadata.endElement(uri, localName, qName);
                return;
            }
            // The stack here is "interface,methods,method,<inputs || outputs>"
            currentParameter.endElement(uri, localName, qName);
            return;
        }
        throw new SAXException(currentPath() + "."+localName + " endElement unknown localName for this level: " + localName);
    }

    @Override
    public void setChildComplete(Object listener, Object obj)
    {
        FieldData parameter = (FieldData) obj;
        if (parameter.inputs)
        {
            synchronized (inputs)
            {
                inputs.add(parameter);
            }
        } else
        {
            synchronized (outputs)
            {
                outputs.add(parameter);
            }
        }
        currentParameter = null;
    }

    /**
     * Export this Method element as XML
     * @param sb  the StringBuilder to add the XML to.  Must not be null.
     * @param level the indent level for this element.
     * @return the StringBuilder handed in with this element's XML appended.
     */
    public StringBuilder export(StringBuilder sb, int level)
    {
        synchronized (inputs)
        {
            synchronized (outputs)
            {
                //@formatter:off
        element(
            sb, MethodElement, InterfaceElement.ItemIdAttr, ""+itemId, null, 
            metadata == null && inputs.size() == 0 && outputs.size() == 0, level);
        //@formatter:on
                exportNames(sb, null, metadata, true, level);
                if (inputs.size() > 0)
                {
                    ++level;
                    element(sb, InputsElement, CType.Open, level);
                    ++level;
                    for (FieldData param : inputs)
                        param.export(sb, FieldData.InputElement, level);
                    --level;
                    element(sb, InputsElement, CType.Close, level);
                    --level;
                }

                if (outputs.size() > 0)
                {
                    ++level;
                    element(sb, OutputsElement, CType.Open, level);
                    ++level;
                    for (FieldData param : outputs)
                        param.export(sb, FieldData.OutputElement, level);
                    --level;
                    element(sb, OutputsElement, CType.Close, level);
                    --level;
                }
                return element(sb, MethodElement, CType.Close, level);
            }
        }
    }

    /**
     * Validate if published/immutable fields have been modified.
     * @param method the method
     * @return true if an immutable field has changed, false otherwise.
     */
    public boolean validPublish(MethodElement method)
    {
        if (itemId != method.itemId)
            return false;
        try
        {
            synchronized (inputs)
            {
                synchronized (method.inputs)
                {
                    for (int i = 0; i < inputs.size(); i++)
                    {
                        FieldData field = inputs.get(i);
                        if (!field.validPublish(method.inputs.get(i)))
                            return false;
                    }
                }
            }
            synchronized (outputs)
            {
                synchronized (method.outputs)
                {
                    for (int i = 0; i < outputs.size(); i++)
                    {
                        FieldData field = outputs.get(i);
                        if (!field.validPublish(method.outputs.get(i)))
                            return false;
                    }
                }
            }
        } catch (Exception e)
        {
            return false;
        }
        return true;
    }
}
