/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.method;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler.ChildComplete;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Methods Element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class MethodsElement extends OpenDofSaxHandler implements ChildComplete
{
    /** The Methods Element tag */
    public static final String MethodsElement = "methods";

    private final List<MethodElement> methods;

    // working
    private MethodElement currentMethod;

    /**
     * Default Methods element constructor 
     */
    public MethodsElement()
    {
        methods = new ArrayList<MethodElement>();
    }

    public List<MethodElement> getMethods()
    {
        synchronized (methods)
        {
            return new ArrayList<MethodElement>(methods);
        }
    }

    public void addMethod(MethodElement method)
    {
        synchronized (methods)
        {
            methods.add(method);
        }
    }

    public void uniqueItemId(Hashtable<Integer, Integer> uniqueMap) throws SAXException
    {
        synchronized (methods)
        {
            for (MethodElement data : methods)
            {
                Integer itemId = data.itemId;
                if (uniqueMap.get(itemId) != null)
                    throw new SAXException("ItemId is not unique: " + itemId);
                uniqueMap.put(itemId, itemId);
            }
        }
    }

    /**
     * Validate that this Methods element meets OpenDOF requirements
     * @param typedefs the typedefs structure to validate typeIds with.  Must not be null
     * @throws SAXException if the validation fails.
     */
    public void validate(TypedefsElement typedefs) throws SAXException
    {
        synchronized (methods)
        {
            for (MethodElement method : methods)
                method.validate(typedefs);
        }
    }

    public static boolean oneOfYours(String element)
    {
        return MethodsElement.equals(element);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (localName.equals(MethodsElement))
        {
            stack.push(localName);
            if (attributes.getLength() != 0)
                throw new SAXException(currentPath() + "."+localName + " " + MethodsElement + " should not have any attributes");
            currentMethod = new MethodElement(this);
            return;
        }
        if (localName.equals(MethodElement.MethodElement))
            // The stack here is: "interface,methods"
            currentMethod = new MethodElement(this);

        // The stack here is "interface,methods,..."  let the children determine error
        currentMethod.startElement(uri, localName, qName, attributes);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        if (isMyTag(MethodsElement))
        {
            // The stack here is: "interface,methods"
            String value = new String(ch, start, length);
            throw new SAXException(currentPath() + " characters unexpected: " + value);
        }
        // The stack here is: "interface,methods,..."
        // let children deal with any errors
        currentMethod.characters(ch, start, length);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (localName.equals(MethodsElement))
        {
            // The stack here is: "interface,methods"
            stack.pop();
            return;
        }
        // The stack here is: "interface,methods,..."
        // let children deal with any errors
        currentMethod.endElement(uri, localName, qName);
    }

    @Override
    public void setChildComplete(Object listener, Object obj)
    {
        synchronized (methods)
        {
            methods.add((MethodElement) obj);
        }
        currentMethod = null;
    }

    /**
     * Export this Methods element as XML
     * @param sb  the StringBuilder to add the XML to.  Must not be null.
     * @param level the indent level for this element.
     * @return the StringBuilder handed in with this element's XML appended.
     */
    public StringBuilder export(StringBuilder sb, int level)
    {
        element(sb, MethodsElement, CType.Open, level);
        ++level;
        synchronized (methods)
        {
            for (MethodElement method : methods)
                method.export(sb, level);
        }
        --level;
        return element(sb, MethodsElement, CType.Close, level);
    }

    /**
     * Validate if published/immutable fields have been modified. 
     * @param other The master/published element to be compared against.
     * @return true if an immutable field has changed, false otherwise.
     */
    public boolean validPublish(MethodsElement other)
    {
        try
        {
            synchronized (methods)
            {
                synchronized (other.methods)
                {
                    if (other.methods.size() != methods.size())
                        return false;
                    for (int i = 0; i < methods.size(); i++)
                    {
                        MethodElement method = methods.get(i);
                        if (!method.validPublish(other.methods.get(i)))
                            return false;
                    }
                }
            }
        } catch (Exception e)
        {
            return false;
        }
        return true;
    }
}
