/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.property;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler.ChildComplete;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.DofTypeBase;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Properties Element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class PropertiesElement extends OpenDofSaxHandler implements ChildComplete
{
    /** The Properties Element tag */
    public static final String PropertiesElement = "properties";

    private final List<PropertyElement> properties;

    // working
    private PropertyElement currentProperty;

    /**
     * Default Properties element constructor 
     */
    public PropertiesElement()
    {
        properties = new ArrayList<PropertyElement>();
    }

    public List<PropertyElement> getProperties()
    {
        synchronized (properties)
        {
            return new ArrayList<PropertyElement>(properties);
        }
    }

    public void addProperty(PropertyElement property)
    {
        synchronized (properties)
        {
            properties.add(property);
        }
    }

    public void uniqueItemId(Hashtable<Integer, Integer> uniqueMap) throws SAXException
    {
        synchronized (properties)
        {
            for (PropertyElement data : properties)
            {
                Integer itemId = data.itemId;
                if (uniqueMap.get(itemId) != null)
                    throw new SAXException("ItemId is not unique: " + itemId);
                uniqueMap.put(itemId, itemId);
            }
        }
    }

    /**
     * Validate that this Properties element meets OpenDOF requirements
     * @param typedefs the typedefs structure to validate typeIds with.  Must not be null
     * @throws SAXException if the validation fails.
     */
    public void validate(TypedefsElement typedefs) throws SAXException
    {
        synchronized (properties)
        {
            for (PropertyElement prop : properties)
                prop.validate(typedefs);
        }
    }

    public static boolean oneOfYours(String element)
    {
        return PropertiesElement.equals(element);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (localName.equals(PropertiesElement))
        {
            // The stack here is: "interface"
            stack.push(localName);
            if (attributes.getLength() != 0)
                throw new SAXException(currentPath() + "."+localName + " " + PropertiesElement + " should not have any attributes");
            return;
        }
        if (localName.equals(PropertyElement.PropertyElement))
            // The stack here is: "interface.properties"
            currentProperty = new PropertyElement(this);

        // The stack here is "interface,properties,..."  let the children determine error
        currentProperty.startElement(uri, localName, qName, attributes);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        if (isMyTag(PropertiesElement))
        {
            // The stack here is: "interface,properties"
            String value = new String(ch, start, length);
            throw new SAXException(currentPath() + " characters unexpected: " + value);
        }
        // The stack here is: "interface,properties,..."
        // let children deal with any errors
        currentProperty.characters(ch, start, length);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (localName.equals(PropertiesElement))
        {
            // The stack here is: "interface,properties"
            stack.pop();
            return;
        }
        // The stack here is: "interface,properties,..."
        // let children deal with any errors
        currentProperty.endElement(uri, localName, qName);
    }

    @Override
    public void setChildComplete(Object listener, Object obj)
    {
        synchronized (properties)
        {
            properties.add((PropertyElement) obj);
        }
        currentProperty = null;
    }

    /**
     * Export this Properties element as XML
     * @param sb  the StringBuilder to add the XML to.  Must not be null.
     * @param level the indent level for this element.
     * @return the StringBuilder handed in with this element's XML appended.
     */
    public StringBuilder export(StringBuilder sb, int level)
    {
        element(sb, PropertiesElement, CType.Open, level);
        ++level;
        synchronized (properties)
        {
            for (PropertyElement prop : properties)
                prop.export(sb, level);
        }
        --level;
        return element(sb, PropertiesElement, CType.Close, level);
    }

    /**
     * Validate if published/immutable fields have been modified. 
     * @param pe The master/published element to be compared against.
     * @return true if an immutable field has changed, false otherwise.
     */
    public boolean validPublish(PropertiesElement pe)
    {
        try
        {
            synchronized (properties)
            {
                synchronized (pe.properties)
                {
                    if (pe.properties.size() != properties.size())
                        return false;
                    for (int i = 0; i < properties.size(); i++)
                    {
                        PropertyElement prop = properties.get(i);
                        if (!prop.validPublish(pe.properties.get(i)))
                            return false;
                    }
                }
            }
        } catch (Exception e)
        {
            return false;
        }
        return true;
    }
}
