/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.property;

import org.opendof.tools.repository.interfaces.opendof.saxParser.InterfaceElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection.FieldData;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Property Element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class PropertyElement extends OpenDofSaxHandler
{
	/** The Property Element tag */
    public static final String PropertyElement = "property";
    public static final String ReadAttr = "read";
    public static final String WriteAttr = "write";

    public volatile int itemId;
    public volatile int reference;
    public volatile MetadataElements metadata;
    private volatile boolean readable;
    private volatile boolean writeable;

    // working
    private final ChildComplete childComplete;

    /**
     * PropertyElement constructor
     * @param childComplete the child completion callback to my parent.  Must not be null.
     */
    public PropertyElement(ChildComplete childComplete)
    {
        this.childComplete = childComplete;
        itemId = -1;
        reference = -1;
    }

    public PropertyElement(int itemId, int reference, MetadataElements metadata, boolean readable, boolean writeable)
    {
        childComplete = null;
        this.itemId = itemId;
        this.reference = reference;
        this.metadata = metadata;
        this.readable = readable;
        this.writeable = writeable;
    }
    
    public int getItemId()
    {
        return itemId;
    }
    
    public int getReference()
    {
        return reference;
    }

    public MetadataElements getMetadata()
    {
        return metadata;
    }
    
    public boolean isReadable()
    {
        return readable;
    }
    
    public boolean isWriteable()
    {
        return writeable;
    }
    
    /**
     * Validate that this Property element meets OpenDOF requirements
     * @param typedefs the typedefs structure to validate typeIds with.  Must not be null
     * @throws SAXException if the validation fails.
     */
    public void validate(TypedefsElement typedefs) throws SAXException
    {
        if (itemId < 0 || itemId > 32767)
            throw new SAXException("PropertyData.itemId is less than 0 or greater than 32767");
        if (reference == -1)
            throw new SAXException("typeref was not set");
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (localName.equals(PropertyElement))
        {
            // The stack here is "interface,properties"
            if (attributes.getLength() != 4)
                throw new SAXException(currentPath() + "."+localName + " expected 4 attibutes");
            int index = attributes.getIndex("", InterfaceElement.ItemIdAttr);
            if (index < 0)
                throw new SAXException(currentPath() + "."+localName + " could not find property: " + InterfaceElement.ItemIdAttr);
            itemId = Integer.parseInt(attributes.getValue(index));
            
            index = attributes.getIndex("", FieldData.TypeRefAttr);
            if (index < 0)
                throw new SAXException(currentPath() + "."+localName + " could not find property: " + FieldData.TypeRefAttr);
            reference = Integer.parseInt(attributes.getValue(index));
            
            index = attributes.getIndex("", ReadAttr);
            if (index < 0)
                throw new SAXException(currentPath() + "."+localName + " could not find property: " + ReadAttr);
            readable = Boolean.parseBoolean(attributes.getValue(index));
            
            index = attributes.getIndex("", WriteAttr);
            if (index < 0)
                throw new SAXException(currentPath() + "."+localName + " could not find property: " + WriteAttr);
            writeable = Boolean.parseBoolean(attributes.getValue(index));
            
            stack.push(PropertyElement);
            return;
        }
        if(MetadataElements.oneOfYours(localName))
        {
            if(metadata == null)
                metadata = new MetadataElements();
            metadata.startElement(uri, localName, qName, attributes);
            return;
        }
        // the only valid child is typeref
//        if(localName.equals(DofReference.TypeRefElement))
//        {
//            typeref = new FieldData(null);
//            typeref.startElement(uri, localName, qName, attributes);
//            return;
//        }
        //
        // The stack here is "interface,properties,property,..."
        // let the children determin errors
//        typeref.startElement(uri, localName, qName, attributes);
        throw new SAXException(currentPath() + "."+localName + " startElement unknown localName for this level: " + localName);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        String current = stack.peek();
        if(MetadataElements.oneOfYours(current))
        {
            metadata.characters(ch, start, length);
            return;
        }
        String value = new String(ch, start, length);
        throw new SAXException(currentPath() + " characters unknown current stack: " + value);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (localName.equals(PropertyElement))
        {
            // The stack here is "interface,properties,property"
            stack.pop();
            childComplete.setChildComplete(childComplete, this);
            return;
        }
        // The stack here is "interface,properties,property,..."
       if(MetadataElements.oneOfYours(localName))
        {
            metadata.endElement(uri, localName, qName);
            return;
        }
        throw new SAXException(currentPath() + "."+localName + " endElement unknown localName for this level: " + localName);
    }

    /**
     * Export this Property element as XML
     * @param sb  the StringBuilder to add the XML to.  Must not be null.
     * @param level the indent level for this element.
     * @return the StringBuilder handed in with this element's XML appended.
     */
    public StringBuilder export(StringBuilder sb, int level)
    {
        //@formatter:off
        element(sb, PropertyElement, 
                        new String[]{InterfaceElement.ItemIdAttr, FieldData.TypeRefAttr, ReadAttr, WriteAttr}, 
                        new String[]{""+itemId, ""+ reference, ""+readable, ""+writeable}, 
                        null, metadata == null, level);
        //@formatter:on
        return exportNames(sb, PropertyElement, metadata, true, level);
    }
    
    /**
     * Validate if published/immutable fields have been modified. 
     * @param pe The master/published element to be compared against.
     * @return true if an immutable field has changed, false otherwise.
     */
    public boolean validPublish(PropertyElement pe)
    {
        if(itemId != pe.itemId)
            return false;
        if(reference != pe.reference)
            return false;
        if(readable != pe.readable)
            return false;
        if(writeable != pe.writeable)
            return false;
        return true;
    }
}
