/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.typedef;

import java.lang.reflect.Constructor;

import org.opendof.tools.repository.interfaces.opendof.saxParser.InterfaceElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection.DofStructureElement;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.numerical.DofNumerical;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.reference.DofReference;
import org.slf4j.LoggerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * DofTypeBase Element SAX handler.  
 */
public class DofTypeBase extends OpenDofSaxHandler
{
    public static final String UnitElement = "unit";
    public static final String MyPath = InterfaceElement.InterfaceElement + "." + TypedefsElement.TypeDefsElement + ".";
    
    public final static String LengthAttr = "length";
    public final static String MinLengthAttr = "min-length";
    public static final String TypeIdAttr = "type-id";
    
    protected volatile String typeAsString;

	protected volatile int typeId;
    protected volatile MetadataElements metadata;

    //working
    public DofTypeBase currentType;
    protected ChildComplete childComplete;
    private String myTag;

    //@formatter:off
    public static final String MyTags[] = new String[]
    {
        "string",   "other",
        "boolean",  "other",
        "uint8",    "numerical",
        "uint16",   "numerical",
        "uint32",   "numerical",
        "uint64",   "numerical",
        "int8",     "numerical",
        "int16",    "numerical",
        "int32",    "numerical",
        "int64",    "numerical",
        "float32",  "numerical",
        "float64",  "numerical",
        "datetime", "other",
        "array",    "collection",
        "uuid",     "other",
        "blob",     "collection",
        "structure","collection",
        "reference","reference",
        "nullable", "reference",
        "iid",      "custom",
        "oid",      "custom",
        "enum",     "collection",
    };
    //@formatter:on

    /**
     * DofTypeBase constructor
     * @param childComplete the child completion callback to my parent.  Must not be null.
     */
    public DofTypeBase(ChildComplete childComplete)
    {
        this.childComplete = childComplete;
        typeId = -1;
    }

    public DofTypeBase(int typeId, MetadataElements metadata)
    {
        this.typeId = typeId;
        this.metadata = metadata;
    }
    
    public int getTypeId()
    {
        return typeId;
    }
    
    public String getTypeAsString()
	{
		return typeAsString;
	}

	public void setTypeAsString(String typeAsString)
	{
		this.typeAsString = typeAsString;
	}
    
    public MetadataElements getMetadata()
    {
        return metadata;
    }

    public static boolean oneOfYours(String element)
    {
        for (int i = 0; i < MyTags.length; i += 2)
        {
            if (MyTags[i].equals(element))
                return true;
        }
        return false;
    }

    /**
     * Validate that this DOFType element meets OpenDOF requirements
     * @param typedefs the typedefs structure to validate typeIds with.  Must not be null
     * @throws SAXException if the validation fails.
     */
    public void validate(TypedefsElement typedefs) throws SAXException
    {
        String location="interface.typedefs." + getClass().getSimpleName();
        if(typeId == -1)
            throw new SAXException("typeId was not set, " + location);
        if(metadata == null)
        {
            if(OpenDofSaxHandler.isPublish())
                throw new SAXException("interface.typedefs.<type>.metadata is null, " + location);
            LoggerFactory.getLogger(getClass()).warn("interface.typedefs.<type>.metadata is null, " + location);
        }
        else
            metadata.validate(true, false, false, location);
    }


    public DofTypeBase instantiate(String tag) throws SAXException
    {
        String clazzName = DofTypeBase.class.getName();
        StringBuilder base = new StringBuilder(clazzName.substring(0, clazzName.lastIndexOf('.') + 1));
        try
        {
            String upperFirstChar = tag.substring(0, 1).toUpperCase();
            StringBuilder postfix = new StringBuilder("Dof").append(upperFirstChar).append(tag.substring(1)).append("Element");
            for (int i = 0; i < MyTags.length; i += 2)
            {
                if (tag.equals(MyTags[i]))
                {
                    base.append(MyTags[i + 1]).append('.');
                    break;
                }
            }
            base.append(postfix);
            Class<?> clazz = Class.forName(base.toString());
            Constructor<?> constructor = clazz.getConstructor(ChildComplete.class);
            return (DofTypeBase) constructor.newInstance(childComplete);
        } catch (Exception e)
        {
            throw new SAXException("could not instantiate DofType: " + base.toString());
        }
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (oneOfYours(localName))
        {
            // first call in from TypedefsElement will always be the specialized DOFType for this base class and will hit here
            // the stack here will be: "interface,typedefs"
            stack.push(localName);
            myTag = localName;
            currentType = instantiate(localName);   // currentType will be the specialized DOFType class for the tag
            currentType.setAttributes(attributes);  // every DOFType will at least have a typeid from this level, capture it here
            currentType.setTypeAsString(localName); // for convenience getter, .getTypeAsString(); this is our last time to know this
            return;
        }
        // subsequent calls hit here with a stack of "interface,typedefs,<type>"
        // the following sub class instances have grouped/shared behaviors
        // checking instance of is sure fire and easy to see what is happening at this level
        if (currentType instanceof DofNumerical)
        {
            // the stack here will be: "interface,typedefs,<type>"
            currentType.startElement(uri, localName, qName, attributes);
            return;
        }
        if (currentType instanceof DofReference)
        {
            // the stack here will be: "interface,typedefs,<type>"
            currentType.startElement(uri, localName, qName, attributes);
            return;
        }
        
        if (currentType instanceof DofStructureElement)
        {
            // the stack here will be: "interface,typedefs,<type>"
            currentType.startElement(uri, localName, qName, attributes);
            return;
        }
        
        // pick up the others which are not covered above
        if(currentPath().startsWith(MyPath + myTag))
        {
            // the stack here will be: "interface,typedefs,<type>,..."
            currentType.startElement(uri, localName, qName, attributes);
            return;
        }
        
        // the above instance checks will handle metadata at their own children levels
        // this one will only hit on the actual child.
        if (MetadataElements.oneOfYours(localName))
        {
            // the stack here will be: "interface,typedefs,<type>"
            if (metadata == null)
                metadata = new MetadataElements();
            metadata.startElement(uri, localName, qName, attributes);
            return;
        }
        throw new SAXException(currentPath() + "."+localName + " DofType startElement unknown local name for this level: " + localName);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        // the following sub class instances have grouped/shared behaviors
        // checking instance of is sure fire and easy to see what is happening at this level
        if(currentType instanceof DofNumerical)
        {
            // the stack here will be: "interface,typedefs,<type>,<metadata>"
            // or: "interface,typedefs,<type>,unit"
            currentType.characters(ch, start, length);
            return;
        }
        if(currentType instanceof DofReference)
        {
            // the stack here will be: "interface,typedefs,<type>,typeref,<metadata>"
            currentType.characters(ch, start, length);
            return;
        }
        if(currentType instanceof DofStructureElement)
        {
            // the stack here will be: "interface,typedefs,structure,typeref,<metadata>"
            currentType.characters(ch, start, length);
            return;
        }
        // pick up the others which are not covered above
        if(currentPath().startsWith(MyPath + myTag) && currentType != null)
        {
            // the stack here will be: "interface,typedefs,<type>,..."
            currentType.characters(ch, start, length);
            return;
        }
        // the above instance checks will handle metadata at their own children levels
        // this one if only for the top level specialized DOFType objects.
        // the stack here will be: "interface,typedefs,<type>,<metadata>"
        String current = stack.peek();
        if (MetadataElements.oneOfYours(current))
        {
            metadata.characters(ch, start, length);
            return;
        }
        String value = new String(ch, start, length);
        throw new SAXException(currentPath() + " characters unknown current stack: " + value);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        // the following sub class instances have grouped/shared behaviors
        // checking instance of is sure fire and easy to see what is happening at this level
        if(currentType instanceof DofNumerical)
        {
            // the stack here will be: "interface,typedefs,<type>,<metadata>"
            // or: "interface,typedefs,<type>,unit"
            // or: "interface,typedefs,<type>
            currentType.endElement(uri, localName, qName);
            return;
        }
        if(currentType instanceof DofReference)
        {
            // the stack here will be: "interface,typedefs,<type>,typeref,<metadata>"
            // or: "interface,typedefs,<type>,typeref"
            // or: "interface,typedefs,<type>
            currentType.endElement(uri, localName, qName);
            return;
        }
        if(currentType instanceof DofStructureElement)
        {
            // the stack here will be: "interface,typedefs,structure,typeref,<metadata>"
            // or: "interface,typedefs,structure,typeref"
            // or: "interface,typedefs,structure
            currentType.endElement(uri, localName, qName);
            return;
        }
        if(currentPath().startsWith(MyPath + myTag) && currentType != null)
        {
            // the stack here will be: "interface,typedefs,<type>,..."
            currentType.endElement(uri, localName, qName);
            return;
        }
        // the above instance checks will handle metadata at their own children levels
        // this one if only for the top level specialized DOFType objects.
        String current = stack.peek();
        if (MetadataElements.oneOfYours(localName))
        {
            // the stack here will be: "interface,typedefs,<type>,<metadata>"
            metadata.endElement(uri, localName, qName);
            return;
        }
        
        // the only thing left is to close up the specialized sub class instances.
        // the stack here will be: "interface,typedefs,<type>"
        for (int i = 0; i < MyTags.length; i += 2)
        {
            if (current.equals(MyTags[i]))
            {
                stack.pop();
                childComplete.setChildComplete(childComplete, this);
                return;
            }
        }
        throw new SAXException(currentPath() + "."+localName + " DofType endElement unknown current stack localName: " + localName);
    }

    protected int setAttributes(Attributes attributes) throws SAXException
    {
        // this method is overridden by all specialized DOFTypes which have more than the one typeid attribute
        int length = attributes.getLength();
        int index = attributes.getIndex(TypeIdAttr);
        if (index != 0)
            throw new SAXException(currentPath() + " expected attribute: " + TypeIdAttr);
        typeId = Integer.parseInt(attributes.getValue(index));
        return length;
    }

    /**
     * Export this DOFType element as XML
     * @param sb  the StringBuilder to add the XML to.  Must not be null.
     * @param level the indent level for this element.
     * @return the StringBuilder handed in with this element's XML appended.
     */
    public StringBuilder export(StringBuilder sb, int level)
    {
        throw new RuntimeException("needs override");
    }
    
    @Override
    public String toString()
    {
        return getClass().getSimpleName() + " typeId: " + typeId;
    }
    
    public boolean validPublish(DofTypeBase doftype)
    {
        if(typeId != doftype.typeId)
            return false;
        return true;
    }
}
