/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.typedef;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler.ChildComplete;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Typedefs Element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class TypedefsElement extends OpenDofSaxHandler implements ChildComplete
{
	public static final String TypeDefsElement = "typedefs";

	// guarded by self
    private final List<DofTypeBase> typedefs;

    //working variables follow
    private DofTypeBase currentDofType;

    /**
     * Default Typedefs element constructor 
     */
    public TypedefsElement()
    {
        typedefs = new ArrayList<DofTypeBase>();
    }

    public List<DofTypeBase> getTypeDefs()
    {
        synchronized (typedefs)
        {
            return new ArrayList<DofTypeBase>(typedefs);
        }
    }
    
    public void addTypeDef(DofTypeBase typedef)
    {
        synchronized (typedefs)
        {
            typedefs.add(typedef);
        }
    }
    
    public void validate() throws SAXException
    {
    	
    	//TODO: removed this test
//        if (typedefs.size() < 1)
//            throw new SAXException("interface.typedefs is empty");
        Hashtable<Integer, Integer> uniqueMap = new Hashtable<Integer, Integer>();
        synchronized (typedefs)
        {
            for (DofTypeBase dtb : typedefs)
            {
                dtb.validate(this);
                Integer typeid = dtb.typeId;
                if(typeid == -1)
                    throw new SAXException("typeId was not set: " + dtb.getClass().getSimpleName());
                if (uniqueMap.get(typeid) != null)
                    throw new SAXException("TypeId: " + typeid + " not unique: " + dtb.getClass().getSimpleName());
                uniqueMap.put(typeid, typeid);
            }
        }
    }

    public void validateTyperef(int typeref) throws SAXException
    {
        synchronized (typedefs)
        {
            for (DofTypeBase dtb : typedefs)
            {
                if (typeref == dtb.typeId)
                    return;
            }
            throw new SAXException("unknown typeref: " + typeref);
        }
    }

    public static boolean oneOfYours(String element)
    {
        return TypeDefsElement.equals(element);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (localName.equals(TypeDefsElement))
        {
            // the stack here will be: "interface"
            stack.push(localName);
            if (attributes.getLength() != 0)
                throw new SAXException(currentPath() + "."+localName + " did not expect attributes");
            return;
        }
        if (stack.peek().equals(TypeDefsElement))
        {
            if (DofTypeBase.oneOfYours(localName))
            {
                // the stack here will be: "interface,typedefs"
                if (currentDofType == null)
                    currentDofType = new DofTypeBase(this);
                currentDofType.startElement(uri, localName, qName, attributes);
                return;
            }
            throw new SAXException(currentPath() + "."+localName + " startElement unknown localName for this level " + localName);
        }
        // if its invalid let the leafs throw the exception
        // the stack here will be: "interface,typedefs,..."
        currentDofType.startElement(uri, localName, qName, attributes);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        if (isMyTag(TypeDefsElement))
        {
            String value = new String(ch, start, length);
            throw new SAXException(currentPath() + " characters unexpected: " + value);
        }
        // the stack here will be: "interface,typedefs,..."
        // let children deal with any bad elements.
        currentDofType.characters(ch, start, length);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (localName.equals(TypeDefsElement))
        {
            stack.pop();
            return;
        }
        // the stack here will be: "interface,typedefs,..."
        // let children deal with any bad elements.
        currentDofType.endElement(uri, localName, qName);
    }

    @Override
    public void setChildComplete(Object listener, Object obj) throws SAXException
    {
        if(obj == null)
            throw new SAXException(currentPath() + " child can not be null");
        synchronized (typedefs)
        {
            typedefs.add(((DofTypeBase) obj));
        }
        currentDofType = null;
    }

    /**
     * Export this Typedefs element as XML
     * @param sb  the StringBuilder to add the XML to.  Must not be null.
     * @param level the indent level for this element.
     * @return the StringBuilder handed in with this element's XML appended.
     */
    public StringBuilder export(StringBuilder sb, int level)
    {
        element(sb, TypeDefsElement, CType.Open, level);
        ++level;
        synchronized (typedefs)
        {
            for (DofTypeBase type : typedefs)
                type.export(sb, level);
        }
        --level;
        element(sb, TypeDefsElement, CType.Close, level);
        return sb;
    }
    
    public boolean validPublish(TypedefsElement tde)
    {
        try
        {
            if(tde.typedefs.size() != typedefs.size())
                return false;
            synchronized (typedefs)
            {
                synchronized(tde.typedefs)
                {
                    for(int i=0; i < typedefs.size(); i++)
                    {
                        DofTypeBase base  = typedefs.get(i);
                        if(!base.validPublish(tde.typedefs.get(i)))
                            return false;
                    }
                }
            }
        }catch(Exception e)
        {
            return false;
        }
        return true;
    }
}
