/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection;

import java.util.ArrayList;
import java.util.List;

import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler.ChildComplete;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.DofTypeBase;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * DofEnum Element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class DofEnumElement extends DofTypeBase implements ChildComplete
{
    private static String EnumElement = "enum";

    public final List<FieldData> fields;

    //working
    private FieldData currentField;

    /**
     * DofEnumElement constructor
     * @param childComplete the child completion callback to my parent.  Must not be null.
     */
    public DofEnumElement(ChildComplete childComplete)
    {
        super(childComplete);
        fields = new ArrayList<FieldData>();
    }

    public DofEnumElement(int itemId, MetadataElements metadata)
    {
        super(itemId, metadata);
        fields = new ArrayList<FieldData>();
    }

    public List<FieldData> getFields()
    {
        synchronized (fields)
        {
            return new ArrayList<FieldData>(fields);
        }
    }

    public void addField(FieldData field)
    {
        synchronized (fields)
        {
            fields.add(field);
        }
    }

    @Override
    public void validate(TypedefsElement typedefs) throws SAXException
    {
        String location = "interface.typedefs." + getClass().getSimpleName();
        super.validate(typedefs);
        synchronized (fields)
        {
            if (fields.size() == 0)
                throw new SAXException("fields is empty, " + location);
            for (FieldData field : fields)
                field.validate(typedefs, location);
        }
    }

    public static boolean oneOfYours(String element)
    {
        if (element.equals(EnumElement))
            return true;
        return FieldData.oneOfYours(element);
    }

    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {

        if (localName.equals(FieldData.ConstantElement))
        {
            // the stack here will be "interface,typedefs,structure"
            currentField = new FieldData(this, false, true);
            currentField.startElement(uri, localName, qName, attributes);
            return;
        }
        String current = stack.peek();
        if (FieldData.ConstantElement.equals(current))
        {
            // the stack here will be "interface,typedefs,structure,typeref"
            currentField.startElement(uri, localName, qName, attributes);
            return;
        }
        // the stack here will be "interface,typedefs,structure"
        if (MetadataElements.oneOfYours(localName))
        {
            // the stack here will be "interface,typedefs,structure,typeref" or "interface,typedefs,structure,typreref,<metadata>"
            if (metadata == null)
                metadata = new MetadataElements();
            metadata.startElement(uri, localName, qName, attributes);
        } else
            throw new SAXException(currentPath() + "."+localName + " startElement unknown localName for this level: " + localName);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        String current = stack.peek();
        if (amIParent(EnumElement) && MetadataElements.oneOfYours(current))
        {
            // the stack here will be "interface,typedefs,structure,<metadata>"
            metadata.characters(ch, start, length);
            return;
        }
        // the stack here will be "interface,typedefs,structure,typreref,<metadata>"
        currentField.characters(ch, start, length);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (localName.equals(EnumElement))
        {
            // the stack here will be "interface,typedefs,structure"
            stack.pop();
            childComplete.setChildComplete(childComplete, this);
            return;
        }
        if (amIParent(EnumElement) && MetadataElements.oneOfYours(localName))
        {
            // the stack here will be "interface,typedefs,structure,<metadata>"
            metadata.endElement(uri, localName, qName);
            return;
        }
        // it better be metadata for the field ... let field throw if not
        // the stack here will be "interface,typedefs,structure,typreref,<metadata>"
        currentField.endElement(uri, localName, qName);
    }

    @Override
    public StringBuilder export(StringBuilder sb, int level)
    {
        element(sb, EnumElement, TypeIdAttr, "" + typeId, null, metadata == null, level);
        ++level;
        synchronized (fields)
        {
            for (FieldData field : fields)
                field.export(sb, FieldData.FieldElement, level);
        }
        --level;
        if (metadata != null)
            metadata.export(sb, level);
        element(sb, EnumElement, CType.Close, level);
        return sb;
    }

    @Override
    public void setChildComplete(Object listener, Object value) throws SAXException
    {
        synchronized (fields)
        {
            fields.add((FieldData) value);
        }
    }

    @Override
    public boolean validPublish(DofTypeBase doftype)
    {
        if (!super.validPublish(doftype))
            return false;

        try
        {
            synchronized (fields)
            {
                synchronized (((DofEnumElement) doftype).fields)
                {
                    for (int i = 0; i < fields.size(); i++)
                    {
                        FieldData field = fields.get(i);
                        if (!field.validPublish(((DofEnumElement) doftype).fields.get(i)))
                            return false;
                    }
                }
            }
        } catch (Exception e)
        {
            return false;
        }
        return true;
    }
}
