/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.collection;

import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * Field data element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class FieldData extends OpenDofSaxHandler
{
	/** The Input Element tag */
    public static final String InputElement = "input";
	/** The Output Element tag */
    public static final String OutputElement = "output";
	/** The Input Element tag */
    public static final String FieldElement = "field";
    public static final String ConstantElement = "constant";
    
    public static final String TypeRefAttr = "type-ref";
    public static final String OrdinalAttr = "ordinal";

    public volatile MetadataElements metadata;
    public volatile int reference;
    public volatile long ordinal;
    
    // working
    private final ChildComplete childComplete;
    private final String attrType;
    public final boolean inputs;
    
    /**
     * FieldData constructor
     * @param childComplete the child completion callback to my parent.  Must not be null.
     */
    public FieldData(ChildComplete childComplete)
    {
        this(childComplete, false, false);
    }
    
    public FieldData(ChildComplete childComplete, boolean inputs, boolean ordinal)
    {
        this.childComplete = childComplete;
        this.inputs = inputs;
        if(ordinal)
            attrType = OrdinalAttr;
        else
            attrType = TypeRefAttr;
    }
    
    public FieldData(int reference, MetadataElements metadata)
    {
        this(reference, -1, false, metadata);
    }
    
    public FieldData(int reference, long ordinal, boolean ordinalType, MetadataElements metadata)
    {
        this.reference = reference;
        this.ordinal = ordinal;
        this.metadata = metadata;
        this.childComplete = null;
        this.inputs = false;
        if(ordinalType)
            attrType = OrdinalAttr;
        else
            attrType = TypeRefAttr;
    }
    
    
    public int getReference()
    {
        return reference;
    }
    
    public MetadataElements getMetadata()
    {
        return metadata;
    }
    
    public long getOrdinal()
    {
        return ordinal;
    }
    
    /**
     * Validate that this Field meets OpenDOF requirements
     * @param typedefs the typedefs structure to validate typeIds with.  Must not be null
     * @param location string identifying the containing element of this field.
     * @throws SAXException if the validation fails.
     */
    public void validate(TypedefsElement typedefs, String location) throws SAXException
    {
      location += ".typeref";
      if(metadata != null)
          metadata.validate(false, false, false, location);
        if(reference < 0 || reference > 32767)
            throw new SAXException("FieldData.typeref is less than 0 or greater than 32767, " + location);
        typedefs.validateTyperef(reference);
    }
    
    public static boolean oneOfYours(String element)
    {
        if(element.equals(ConstantElement))
            return true;
        if(element.equals(InputElement))
            return true;
        if(element.equals(OutputElement))
            return true;
        return element.equals(FieldElement);
    }
    
    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        if (oneOfYours(localName))
        {
            stack.push(localName);
            int size = attributes.getLength();
            if (size == 0)
                throw new SAXException(currentPath() + "."+localName + " " + InputElement + ", " + OutputElement + " and " + FieldElement + " requires  the " + attrType + " attribute");
            if (size > 1)
                throw new SAXException(currentPath() + "."+localName + " " + InputElement + ", " + OutputElement + " or " + FieldElement + " has more attributes than expected");
            int index = attributes.getIndex(attrType);
            if(attrType.equals(TypeRefAttr))
                reference = Integer.parseInt(attributes.getValue(index));
            else
                ordinal = Long.parseLong(attributes.getValue(index));
            return;
        }
        
        if(oneOfYours(stack.peek()))
        {
            if(MetadataElements.oneOfYours(localName))
            {
                if(metadata == null)
                    metadata = new MetadataElements();
                metadata.startElement(uri, localName, qName, attributes);
                return;
            }
        }
        throw new SAXException(currentPath() + "."+localName + " startElement unknown localName for this level: " + localName);
    }
    
    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        String current = stack.peek();
        if(amIParent(InputElement) || amIParent(OutputElement) || amIParent(FieldElement) || amIParent(ConstantElement))
        {
            if(MetadataElements.oneOfYours(current))
            {
                metadata.characters(ch, start, length);
                return;
            }
        }
        String value = new String(ch, start, length);
        throw new SAXException(currentPath() + " characters unknown current stack: " + value);
    }
    
    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if(oneOfYours(localName))
        {
            stack.pop();
            if(childComplete != null)
                childComplete.setChildComplete(childComplete, this);
            return;
        }
        String current = stack.peek();
        if(amIParent(InputElement) || amIParent(OutputElement) || amIParent(FieldElement) || amIParent(ConstantElement))
        {
            if(MetadataElements.oneOfYours(current))
            {
                metadata.endElement(uri, localName, qName);
                return;
            }
        }
        throw new SAXException(currentPath() + "."+localName + " endElement unknown current stack localName: " + localName);
    }
    
    public StringBuilder export(StringBuilder sb, String element, int level)
    {
        element(sb, element, attrType, ""+reference, null, metadata == null, level);
        return exportNames(sb, element, metadata, true, level);
    }
    
    public boolean validPublish(FieldData field)
    {
        if(attrType.equals(TypeRefAttr))
        {
            if(reference != field.reference)
                return false;
        }else
        {
            if(ordinal != field.ordinal)
                return false;
        }
        return true;
    }
}
