/*
**  Copyright (c) 2010-2015, Panasonic Corporation.
**
**  Permission to use, copy, modify, and/or distribute this software for any
**  purpose with or without fee is hereby granted, provided that the above
**  copyright notice and this permission notice appear in all copies.
**
**  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
**  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
**  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
**  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
**  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
**  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/
package org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.numerical;

import org.opendof.tools.repository.interfaces.opendof.saxParser.MetadataElements;
import org.opendof.tools.repository.interfaces.opendof.saxParser.OpenDofSaxHandler;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.DofTypeBase;
import org.opendof.tools.repository.interfaces.opendof.saxParser.typedef.TypedefsElement;
import org.slf4j.LoggerFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

/**
 * DofNumerical element SAX handler.  
 */
@SuppressWarnings("javadoc")
public class DofNumerical extends DofTypeBase
{
    public static final String UnitElement = "unit";

    public final static String MinAttr = "min";
    public final static String MaxAttr = "max";

    protected volatile Number min;
    protected volatile Number max;
    protected volatile String unit;

    //working
    private StringBuilder unitChars;
    
    /**
     * DofNumerical constructor
     * @param childComplete the child completion callback to my parent.  Must not be null.
     */
    public DofNumerical(ChildComplete childComplete)
    {
        super(childComplete);
    }

    public DofNumerical(int itemId, MetadataElements metadata, Number min, Number max, String unit)
    {
        super(itemId, metadata);
        this.min = min;
        this.max = max;
        this.unit = unit;
    }
    
    public Number getMin()
    {
        return min;
    }
    
    public Number getMax()
    {
        return max;
    }
    
    public String getUnit()
    {
        return unit;
    }
    
    public static boolean oneOfYours(String element)
    {
        return UnitElement.equals(element);
    }
    
    @Override
    public void validate(TypedefsElement typedefs) throws SAXException
    {
        super.validate(typedefs);
    }
    
    @Override
    public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
    {
        // for numericals, only metadata and unit are valid as children
        if (MetadataElements.oneOfYours(localName))
        {
            // the stack here will be "interface,typedefs,<type>"
            if (metadata == null)
                metadata = new MetadataElements();
            metadata.startElement(uri, localName, qName, attributes);
            return;
        }
        if (localName.equals(UnitElement))
        {
            // the stack here will be "interface,typedefs,<type>"
            stack.push(localName);
            unitChars = new StringBuilder();
            return;
        }
        throw new SAXException(currentPath() + "."+localName + " DofNumerical startElement unknown local name for this level " + localName);
    }

    @Override
    public void characters(char ch[], int start, int length) throws SAXException
    {
        // for numericals, only metadata and unit are valid
        String current = stack.peek();
        if (MetadataElements.oneOfYours(current))
        {
            // the stack here will be "interface,typedefs,<type>,<metadata>" 
            metadata.characters(ch, start, length);
            return;
        }
        if (DofNumerical.oneOfYours(current))
        {
            // the stack here will be "interface,typedefs,<type>,unit" 
            unitChars.append(new String(ch, start, length));
            return;
        }
        String value = new String(ch, start, length);
        throw new SAXException(currentPath() + " characters unknown current stack: " + value);
    }

    @Override
    public void endElement(String uri, String localName, String qName) throws SAXException
    {
        if (MetadataElements.oneOfYours(localName))
        {
            // the stack here will be "interface,typedefs,<type>,<metadata>"  
            metadata.endElement(uri, localName, qName);
            return;
        }
        if (UnitElement.equals(localName))
        {
            // the stack here will be "interface,typedefs,<type>,unit"  
            stack.pop();
            unit = unitChars.toString();
            return;
        }
        // the stack here will be "interface,typedefs,<type>" let the super clean it up
        super.endElement(uri, localName, qName);
    }

    
    public StringBuilder exportNumerical(StringBuilder sb, String tag, String typeId, Object min, Object max, String unit, MetadataElements names, int level)
    {
        if (min == null)
            element(sb, tag, TypeIdAttr, typeId, null, names == null, level);
        else
            element(sb, tag, new String[] { TypeIdAttr, MinAttr, MaxAttr }, new String[] { typeId, min.toString(), max.toString() }, null, names == null, level);
        exportNames(sb, tag, names, unit == null, level);
        if (unit != null)
        {
            ++level;
            element(sb, UnitElement, (String) null, (String) null, unit, true, level);
            --level;
            element(sb, tag, CType.Close, level);
        }
        return sb;
    }
    
    @Override
    public boolean validPublish(DofTypeBase doftype)
    {
        if(!super.validPublish(doftype))
            return false;
        if(!checkEquals(min, ((DofNumerical)doftype).min))
            return false;
        if(!checkEquals(max, ((DofNumerical)doftype).max))
            return false;
        if(!checkEquals(unit, ((DofNumerical)doftype).unit))
            return false;
        return true;
    }
}
